unit COMReports;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  DsgnIntf, SEditProp, Dialogs, SSmrtRep, DDEReports, Excel97, OleServer,
  DDEClient, ActiveX, ComObj, RThread, CommonType;

type

{TExcelCOMReport}

  TExcelCOMReport = class(TReportAbstract)
    private

      FExcel: _Application;
      FCurrentWorkBook: _Workbook;
      FCurrentWorkSheet: _Worksheet;
      FReportWorkBook: _Workbook;
      FReportSheet: _Worksheet;
      FServerData: PServerData;
      FMaket: String;
      FOutBook: String;
      FCurrRow: Integer; //LongWord;
      FCurrCol: Integer; //LongWord;
      FSectionsV: array [1..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;
      FSectionsH: array [1..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;
      FCurrentSection: String;
      FCurrentSectionNum: Integer;
      FInputSheet: String;
      FOutputSheet: String;
      FCharByNum: Char;
      FCharByName: Char;
      FTypeDiraction: TTtypeDiraction;
      FLastDynSection:    Integer;
      FLastDynSectionH:    Integer;
      lWM_REP_SECTION_PREPARE: boolean;
      FRemoteMachineName: String;
      FConnectKind:       TConnectKind;

      function ActivateComObject: Boolean;
      function OpenInputBlank: Boolean;

      procedure StartReportV;
      procedure PrepareSectionV(SectionName: String; SectionNum: LongWord);
      procedure OutSectionV(SectionName: String; bOut: Boolean; SectionNum: LongWord);
      procedure CloseSectionsV;

      procedure StartReportH;
      procedure PrepareSectionH(SectionName: String; SectionNum: LongWord);
      procedure OutSectionH(SectionName: String; bOut: Boolean; SectionNum: LongWord);
      procedure CloseSectionsH;

      procedure StartBiDirectionReport;
      function GetConnectKind: TConnectKind;
      procedure SetConnectKind(const Value: TConnectKind);

    protected

      function GetCurrentSection: String; override;
      function GetCurrentSectionNum: Integer; override;
      function GetTypeDiraction: TTtypeDiraction; override;

      property  ServerData: PServerData read FServerData write FServerData;

    public

      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      procedure ReportResume; override;
      procedure ReportExecute; override;
      procedure ReportHorizontalyResume; override;
      procedure ReportHorizontalyExecute; override;
      procedure BiDirectionResume; override;
      procedure BiDirectionExecute; override;
      function PrepareSection(const SectionName: ShortString): Integer; override;
      function PrepareHorizontalySection(const SectionName: ShortString): Integer; override;

      procedure OutSection(SectionName: ShortString); override;
      procedure OutHorizontalySection(SectionName: ShortString); override;
      procedure OutSectionByNum(SectionNum: Integer); override;
      procedure OutHorizontalySectionByNum(SectionNum: Integer); override;

      property Excel: _Application read FExcel;
      property ReportSheet: _Worksheet read FReportSheet;

    published

      property RemoteMachineName: String read FRemoteMachineName write FRemoteMachineName;
      property ConnectKind: TConnectKind read GetConnectKind write SetConnectKind;
      property InputBlank: String read FMaket write FMaket;
      property InputSheet: String read FInputSheet write FInputSheet;
      property OutputBlank: String read FOutBook write FOutBook;
      property OutputSheet: String read FOutputSheet write FOutputSheet;
      property CharByNum: Char read FCharByNum write FCharByNum default #35; {#0001}
      property CharByName: Char read FCharByName write FCharByName default #126; {~name~}

      property OnStartReport;
      property OnEndReport;
      property OnReportError;
      property OnMoveTop;
      property OnMoveNext;
      property OnSectionBegin;
      property OnSectionEnd;
      property OnSectionComplit;
      property OnSectionCheck;
      property OnDataRequest;
      property OnRequestByName;
      property OnRequestByNum;
      property OnSectionPrepare;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TExcelCOMReport]);
  RegisterPropertyEditor(TypeInfo(String), TExcelCOMReport, 'InputBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TExcelCOMReport, 'OutputBlank', TEdSaveBlank);
end;

{ TExcelCOMReport }

function TExcelCOMReport.ActivateComObject: Boolean;
var
  HR: HResult;
  ErrorStr: string;
  iTmp: IUnknown;
begin
  try
    FExcel := nil;
    iTmp := nil;
    case ConnectKind of
      ckNewInstance:
        FExcel := CreateComObject(FServerData^.ClassId) as _Application;
      ckRunningInstance:
        begin
          HR := GetActiveObject(FServerData^.ClassId, nil, iTmp);
          FExcel := iTmp as _Application;
          iTmp := nil;
          if not Succeeded(HR) then
          begin
            ErrorStr := Format('sNoRunningObject %d %s', [ClassIDToProgID(FServerData^.ClassId),
                                                  GuidToString(FServerData^.ClassId)]);
            raise EOleSysError.Create(ErrorStr, HR, 0);
          end;
        end;
      ckRunningOrNew:
        if not Succeeded(GetActiveObject(FServerData^.ClassId, nil, iTmp)) then
          FExcel := CreateComObject(FServerData^.ClassId) as _Application
        else
          FExcel := iTmp as _Application;
      ckRemote:
        {Highly inefficient: requires at least two round trips - GetClassObject + QI}
        FExcel := CreateRemoteComObject(RemoteMachineName, FServerData^.ClassID) as _Application;
    end;
  finally
    iTmp := nil;
    Result := true;
    if FExcel = nil then Result := false;
  end;
end;

procedure TExcelCOMReport.BiDirectionExecute;
begin
  StartBiDirectionReport;
end;

procedure TExcelCOMReport.BiDirectionResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartBiDirectionReport;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TExcelCOMReport.CloseSectionsH;
var
  i, j, k: Integer;
  pp: pTExcelCellContent;
begin
	for i := 1 to MAX_SECTION_NUM do
	  if FSectionsH[i].SectionUsed then
    begin
		  FSectionsH[i].SectionUsed := FALSE;
      //StrDispose(PChar(FSectionsH[i].Widths));
      k := FSectionsH[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTExcelCellContent(FSectionsH[i].Contents.Items[j]);
        //StrDispose(pp^.ContentColRow);
        pp^.ContentColRow := Unassigned;
        StrDispose(PChar(pp));
      end;
//      FSectionsH[i].iRange := nil;
      FSectionsH[i].Contents.Destroy;
    end;
end;

procedure TExcelCOMReport.CloseSectionsV;
var
  i, j, k: Integer;
  pp: pTExcelCellContent;
begin
	for i := 1 to MAX_SECTION_NUM do
	  if FSectionsV[i].SectionUsed then
    begin
		  FSectionsV[i].SectionUsed := FALSE;
      k := FSectionsV[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTExcelCellContent(FSectionsV[i].Contents.Items[j]);
        //StrDispose(pp^.ContentColRow);
        pp^.ContentColRow := Unassigned;
        StrDispose(PChar(pp));
      end;
      //FSectionsV[i].iRange := nil;
      FSectionsV[i].Contents.Destroy;
    end;
end;

constructor TExcelCOMReport.Create(aOwner: TComponent);
const
  CServerData: TServerData = (
    ClassID:   '{00024500-0000-0000-C000-000000000046}';
    IntfIID:   '{000208D5-0000-0000-C000-000000000046}';
    EventIID:  '{00024413-0000-0000-C000-000000000046}';
    LicenseKey: nil;
    Version: 500);
begin
  inherited Create(aOwner);
  ServerData := @CServerData;
  FRemoteMachineName := '';
  FConnectKind := ckRunningOrNew;
  FCharByNum := #35;
  FCharByName := #126;
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  lWM_REP_SECTION_PREPARE := false;
end;

destructor TExcelCOMReport.Destroy;
begin
  CloseSectionsH;
  CloseSectionsV;
  inherited Destroy;
end;

function TExcelCOMReport.GetConnectKind: TConnectKind;
begin
  // Should the setting of a RemoteMachine name override the Connection Kind ??
  if RemoteMachineName <> '' then
    Result := ckRemote
  else
    Result := FConnectKind;
end;

function TExcelCOMReport.GetCurrentSection: String;
begin
  Result := FCurrentSection;
end;

function TExcelCOMReport.GetCurrentSectionNum: Integer;
begin
  Result := FCurrentSectionNum;
end;

function TExcelCOMReport.GetTypeDiraction: TTtypeDiraction;
begin
  Result := FTypeDiraction;
end;

function TExcelCOMReport.OpenInputBlank: Boolean;
var
  i, j: Integer;
  bFound: Boolean;
  CDir: String;
begin
  bFound := false;
  FCurrentWorkBook := nil;
  j := 0;
  for i := 1 to Excel.Workbooks.Count do
    if Excel.Workbooks.Item[i].Name = ExtractFileName(InputBlank) then
    begin
      j := i;
      bFound := true;
      break;
    end;
  if bFound then
    FCurrentWorkBook := Excel.Workbooks.Item[j]
  else
    try
      Excel.DisplayAlerts[0] := false;
      FCurrentWorkBook := Excel.Workbooks.Open(InputBlank, true, false, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0);
    except
      try
        Excel.DisplayAlerts[0] := false;
        CDir := GetCurrentDir;
        if CDir[Length(CDir)] <> '\' then CDir := CDir + '\';
        FCurrentWorkBook := Excel.Workbooks.Open(CDir + InputBlank, true, false, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0);
      except
        FCurrentWorkBook := nil;
      end;
    end;
  Result := (FCurrentWorkBook <> nil);
end;

procedure TExcelCOMReport.OutHorizontalySection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin

  if SectionName = 'H_REPORTHEADER'         then
  begin
    OutSectionH('H_REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPHEADER'          then
  begin
    OutSectionH('H_GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPHEADER'       then
  begin
    OutSectionH('H_SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPHEADER'    then
  begin
    OutSectionH('H_SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPHEADER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'H_BODY'                 then
  begin
    OutSectionH('H_BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPFOOTER'    then
  begin
    OutSectionH('H_SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPFOOTER'       then
  begin
    OutSectionH('H_SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPFOOTER'          then
  begin
    OutSectionH('H_GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'H_REPORTFOOTER'         then
  begin
    OutSectionH('H_REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsH[i].SectionName then
    begin
      OutSectionH(SectionName, true, i);
      Break;
    end;
lEnd:
end;

procedure TExcelCOMReport.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  OutSectionH(FSectionsH[SectionNum].SectionName, true, SectionNum);
end;

procedure TExcelCOMReport.OutSection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin
  if SectionName = 'REPORTHEADER'         then
  begin
    OutSectionV('REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'GROUPHEADER'          then
  begin
    OutSectionV('GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPHEADER'       then
  begin
    OutSectionV('SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPHEADER'    then
  begin
    OutSectionV('SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPHEADER'  then
  begin
    OutSectionV('SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'BODY'                 then
  begin
    OutSectionV('BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionV('SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPFOOTER'    then
  begin
    OutSectionV('SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPFOOTER'       then
  begin
    OutSectionV('SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'GROUPFOOTER'          then
  begin
    OutSectionV('GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'REPORTFOOTER'         then
  begin
    OutSectionV('REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsV[i].SectionName then
    begin
      OutSectionV(SectionName, true, i);
      Break;
    end;
lEnd:
end;

procedure TExcelCOMReport.OutSectionByNum(SectionNum: Integer);
begin
  OutSectionV(FSectionsV[SectionNum].SectionName, true, SectionNum);
end;

procedure TExcelCOMReport.OutSectionH(SectionName: String; bOut: Boolean;
  SectionNum: LongWord);
var
  dwi: Integer;
  //i: Integer;
  pp: pTExcelCellContent;
  ContentColRow: Variant;
  ContentColRowStr: String;
  bCheck: Boolean;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
  H: Boolean;
begin
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
	if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
	if FSectionsH[SectionNum].SectionUsed then
  begin
    if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
		if bOut then
    begin
      //(FSectionsV[SectionNum].iRange as Range).Copy(FReportSheet.Cells.Item[1, FCurrCol]);
      FCurrentWorkSheet.Range[SectionName, SectionName].Copy(FReportSheet.Cells.Item[1, FCurrCol]);
      //for i := 0 to FSectionsH[SectionNum].ColsCount-1 do
      //  (FReportSheet.Range[NULL, FCurrCol + i] as Range).ColumnWidth := Integer(pInteger(PChar(FSectionsH[SectionNum].Widths) + i*SizeOf(Integer))^);
			dwi := 0;
			while dwi < FSectionsH[SectionNum].Contents.Count do
      begin
        pp := pTExcelCellContent(FSectionsH[SectionNum].Contents.Items[dwi]);
        ContentColRow := pp^.ContentColRow;
        bCheck := false;
        if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, ContentColRow, bCheck, pp^.SectionRow, FCurrCol + pp^.SectionCol - 1, pp^.SectionRow, pp^.SectionCol);
        if bCheck then
          FReportSheet.Cells.Item[pp^.SectionRow, FCurrCol + pp^.SectionCol - 1] := ContentColRow
        else begin
          //
          if Assigned(OnRequestByNum) then
          begin
            H := false;
            pSt := 0;
            ContentColRowStr := VarToStr(ContentColRow);
            while true do
            begin
              pInt := LScan(ContentColRowStr, FCharByNum, pSt);
              if pInt = 0 then Break else H := true;
              pSt := pInt;
              if  ( ContentColRowStr[pInt + 1] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 2] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 3] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 4] in ['0'..'9'] ) then
              begin
                iP := StrToInt(Copy(ContentColRow, pInt + 1, 4));
                sStr := '';
                OnRequestByNum(self, SectionName, iP, sStr);
                ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt + 5, Length(ContentColRow));
              end;
            end;
            if H then FReportSheet.Cells.Item[pp^.SectionRow, FCurrCol + pp^.SectionCol - 1] := ContentColRow;
          end;
          if Assigned(OnRequestByName) then
          begin
            H := false;
            pSt := 0;
            while true do
            begin
              pInt := LScan(ContentColRow, FCharByName, pSt);
              if pInt = 0 then Break;
              pInt1 := LScan(ContentColRow, FCharByName, pInt);
              if pInt1 = 0 then Break else h := true;
              pSt := pInt1;
              sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
              sStr := '';
              OnRequestByName(self, SectionName, sField, sStr);
              ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
            end;
            if H then FReportSheet.Cells.Item[pp^.SectionRow, FCurrCol + pp^.SectionCol - 1] := ContentColRow;
          end;
          //
        end;
				Inc(dwi);
			end;
			FCurrCol := FCurrCol + FSectionsH[SectionNum].ColsCount;
      if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
		end;
	end;
  if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
end;

procedure TExcelCOMReport.OutSectionV(SectionName: String; bOut: Boolean;
  SectionNum: LongWord);
var
  dwi: Integer;
  pp: pTExcelCellContent;
  ContentColRow: Variant;
  ContentColRowStr: String;
  bCheck: Boolean;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
  H: Boolean;
begin
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
	if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
	if FSectionsV[SectionNum].SectionUsed then
  begin
    if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
		if bOut then
    begin
      //(FSectionsV[SectionNum].iRange as Range).Copy(FReportSheet.Cells.Item[FCurrRow, 1]);
      FCurrentWorkSheet.Range[SectionName, SectionName].Copy(FReportSheet.Cells.Item[FCurrRow, 1]);
			dwi := 0;
			while dwi < FSectionsV[SectionNum].Contents.Count do
      begin
        pp := pTExcelCellContent(FSectionsV[SectionNum].Contents.Items[dwi]);
        ContentColRow := pp^.ContentColRow;
        bCheck := false;
        if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, ContentColRow, bCheck, FCurrRow + pp^.SectionRow - 1, pp^.SectionCol, pp^.SectionRow, pp^.SectionCol);
        if bCheck then
          FReportSheet.Cells.Item[FCurrRow + pp^.SectionRow - 1, pp^.SectionCol] := ContentColRow
        else begin
          //
          if Assigned(OnRequestByNum) then
          begin
            H := false;
            pSt := 0;
            ContentColRowStr := VarToStr(ContentColRow);
            while true do
            begin
              pInt := LScan(ContentColRowStr, FCharByNum, pSt);
              if pInt = 0 then Break else H := true;
              pSt := pInt;
              if  ( ContentColRowStr[pInt + 1] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 2] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 3] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 4] in ['0'..'9'] ) then
              begin
                iP := StrToInt(Copy(ContentColRow, pInt + 1, 4));
                sStr := '';
                OnRequestByNum(self, SectionName, iP, sStr);
                ContentColRowStr := Copy(ContentColRowStr, 1, pInt - 1) + sStr + Copy(ContentColRowStr, pInt + 5, Length(ContentColRowStr));
              end;
            end;
            if H then FReportSheet.Cells.Item[FCurrRow + pp^.SectionRow - 1, pp^.SectionCol] := ContentColRowStr;
          end;
          if Assigned(OnRequestByName) then
          begin
            H := false;
            pSt := 0;
            while true do
            begin
              pInt := LScan(ContentColRow, FCharByName, pSt);
              if pInt = 0 then Break;
              pInt1 := LScan(ContentColRow, FCharByName, pInt);
              if pInt1 = 0 then Break else H := true;
              pSt := pInt1;
              sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
              sStr := '';
              OnRequestByName(self, SectionName, sField, sStr);
              ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
            end;
            if H then FReportSheet.Cells.Item[FCurrRow + pp^.SectionRow - 1, pp^.SectionCol] := ContentColRow;
          end;
          //
        end;
				Inc(dwi);
			end;
			FCurrRow := FCurrRow + FSectionsV[SectionNum].RowsCount;
      if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
		end;
	end;
  if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
end;

function TExcelCOMReport.PrepareHorizontalySection(
  const SectionName: ShortString): Integer;
begin
  if FLastDynSectionH > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSectionH;
  Inc(FLastDynSectionH);
  PrepareSectionH(SectionName, Result);
end;

function TExcelCOMReport.PrepareSection(
  const SectionName: ShortString): Integer;
begin
  if FLastDynSection > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSection;
  Inc(FLastDynSection);
  PrepareSectionV(SectionName, Result);
end;

procedure TExcelCOMReport.PrepareSectionH(SectionName: String; SectionNum: LongWord);
var
  ContentColRow: Variant;
  pp: pTExcelCellContent;
  oRR: Range;
  CC: OleVariant;
  i, j, k, x1, x2, x3, y1, y2, y3: LongWord;
  TopLeft, BottomRight: TExcelCell;
begin
	FSectionsH[SectionNum].SectionName := SectionName;
	FSectionsH[SectionNum].SectionNum := SectionNum;
	FSectionsH[SectionNum].SectionUsed := FALSE;
  try
    oRR := FCurrentWorkSheet.Range[SectionName, SectionName];
    TExcelDDEClient.ParseSelEntry(oRR.Address[NULL, NULL, XlReferenceStyle(xlR1C1), NULL, NULL], TopLeft, BottomRight);
    x3 := oRR.SpecialCells(xlCellTypeLastCell, NULL).Row;
    y3 := oRR.SpecialCells(xlCellTypeLastCell, NULL).Column;
    if (TopLeft.Col = 0) then TopLeft.Col := 1;
    if (TopLeft.Row = 0) then TopLeft.Row := 1;
    if (BottomRight.Col > y3) or (BottomRight.Col = 0) then BottomRight.Col := y3;
    if (BottomRight.Row > x3) or (BottomRight.Row = 0) then BottomRight.Row := x3;
    oRR := FCurrentWorkSheet.Range[ToA1(TopLeft), ToA1(BottomRight)];
  except
    oRR := nil;
  end;
	if oRR <> nil then
  begin
		FSectionsH[SectionNum].SectionUsed := TRUE;
    FSectionsH[SectionNum].Contents := TList.Create;
    //FSectionsH[SectionNum].iRange := oRR;
    //
    CC := oRR.Value;
    if VarType(CC) >= varArray then
    begin
      k := VarArrayDimCount(CC);
      if k = 1 then
      begin
        FSectionsH[SectionNum].ColsCount := 1;
        x1 := VarArrayLowBound(CC, 1);
        x2 := VarArrayHighBound(CC,1);
        for i := x1 to x2 do
        begin
          ContentColRow := Unassigned;
          ContentColRow := CC[i];
          if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
          begin
            pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
            FillChar(pp^, SizeOf(TExcelCellContent), 0);
            //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
            //pp^.ContentLen := Length(ContentColRow);
            //StrPCopy(pp^.ContentColRow, ContentColRow);
            pp^.ContentColRow := Unassigned;
            pp^.ContentColRow := ContentColRow;
            pp^.SectionRow := 1;
            pp^.SectionCol := i;
            FSectionsH[SectionNum].Contents.Add(pp);
          end;
        end;
      end else if k = 2 then
      begin
        x1 := VarArrayLowBound(CC, 1);
        x2 := VarArrayHighBound(CC,1);
        y1 := VarArrayLowBound(CC, 2);
        y2 := VarArrayHighBound(CC,2);
        FSectionsH[SectionNum].ColsCount := y2 - y1 + 1;
        for i := x1 to x2 do
          for j := y1 to y2 do
          begin
            ContentColRow := Unassigned;
            ContentColRow := CC[i, j];
            if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
            begin
              pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
              FillChar(pp^, SizeOf(TExcelCellContent), 0);
              //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
              //pp^.ContentLen := Length(ContentColRow);
              //StrPCopy(pp^.ContentColRow, ContentColRow);
              pp^.ContentColRow := Unassigned;
              pp^.ContentColRow := ContentColRow;
              pp^.SectionRow := i;
              pp^.SectionCol := j;
              FSectionsH[SectionNum].Contents.Add(pp);
            end;
          end;
      end else begin
      end;
    end else begin
      FSectionsH[SectionNum].ColsCount := 1;
      ContentColRow := Unassigned;
      ContentColRow := CC;
      if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
      begin
        pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
        FillChar(pp^, SizeOf(TExcelCellContent), 0);
        //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
        //pp^.ContentLen := Length(ContentColRow);
        //StrPCopy(pp^.ContentColRow, ContentColRow);
        pp^.ContentColRow := Unassigned;
        pp^.ContentColRow := ContentColRow;
        pp^.SectionRow := 1;
        pp^.SectionCol := 1;
        FSectionsH[SectionNum].Contents.Add(pp);
      end;
    end;
    //
    //FSectionsH[SectionNum].Widths := pInteger(StrAlloc(SizeOf(Integer)*FSectionsH[SectionNum].ColsCount));
    //for k := 0 to FSectionsH[SectionNum].ColsCount-1 do
    //  (pInteger(PChar(FSectionsH[SectionNum].Widths) + k*SizeOf(Integer)))^ := (FReportSheet.Range[NULL, FCurrCol + k] as Range).ColumnWidth;
    oRR := nil;
	end;
end;

procedure TExcelCOMReport.PrepareSectionV(SectionName: String; SectionNum: LongWord);
var
  ContentColRow: Variant;
  pp: pTExcelCellContent;
  oRR: Range;
  CC: OleVariant;
  i, j, k, x1, x2, x3, y1, y2, y3: LongWord;
  TopLeft, BottomRight: TExcelCell;
begin
	FSectionsV[SectionNum].SectionName := SectionName;
	FSectionsV[SectionNum].SectionNum := SectionNum;
	FSectionsV[SectionNum].SectionUsed := FALSE;
  try
    oRR := FCurrentWorkSheet.Range[SectionName, SectionName];
    TExcelDDEClient.ParseSelEntry(oRR.Address[NULL, NULL, XlReferenceStyle(xlR1C1), NULL, NULL], TopLeft, BottomRight);
    x3 := oRR.SpecialCells(xlCellTypeLastCell, NULL).Row;
    y3 := oRR.SpecialCells(xlCellTypeLastCell, NULL).Column;
    if (TopLeft.Col = 0) then TopLeft.Col := 1;
    if (TopLeft.Row = 0) then TopLeft.Row := 1;
    if (BottomRight.Col > y3) or (BottomRight.Col = 0) then BottomRight.Col := y3;
    if (BottomRight.Row > x3) or (BottomRight.Row = 0) then BottomRight.Row := x3;
    oRR := FCurrentWorkSheet.Range[ToA1(TopLeft), ToA1(BottomRight)];
  except
    oRR := nil;
  end;
	if oRR <> nil then
  begin
		FSectionsV[SectionNum].SectionUsed := TRUE;
    FSectionsV[SectionNum].Contents := TList.Create;
    //FSectionsV[SectionNum].iRange := oRR;
    //
    CC := oRR.Value;
    if VarType(CC) >= varArray then
    begin
      k := VarArrayDimCount(CC);
      if k = 1 then
      begin
        FSectionsV[SectionNum].RowsCount := 1;
        x1 := VarArrayLowBound(CC, 1);
        x2 := VarArrayHighBound(CC,1);
        for i := x1 to x2 do
        begin
          ContentColRow := Unassigned;
          ContentColRow := CC[i];
          if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
          begin
            pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
            FillChar(pp^, SizeOf(TExcelCellContent), 0);
            //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
            //pp^.ContentLen := Length(ContentColRow);
            //StrPCopy(pp^.ContentColRow, ContentColRow);
            pp^.ContentColRow := Unassigned;
            pp^.ContentColRow := ContentColRow;
            pp^.SectionRow := 1;
            pp^.SectionCol := i;
            FSectionsV[SectionNum].Contents.Add(pp);
          end;
        end;
      end else if k = 2 then
      begin
        x1 := VarArrayLowBound(CC, 1);
        x2 := VarArrayHighBound(CC,1);
        y1 := VarArrayLowBound(CC, 2);
        y2 := VarArrayHighBound(CC,2);
        FSectionsV[SectionNum].RowsCount := x2 - x1 + 1;
        for i := x1 to x2 do
          for j := y1 to y2 do
          begin
            ContentColRow := Unassigned;
            ContentColRow := CC[i, j];
            if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
            begin
              pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
              FillChar(pp^, SizeOf(TExcelCellContent), 0);
              //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
              //pp^.ContentLen := Length(ContentColRow);
              //StrPCopy(pp^.ContentColRow, ContentColRow);
              pp^.ContentColRow := Unassigned;
              pp^.ContentColRow := ContentColRow;
              pp^.SectionRow := i;
              pp^.SectionCol := j;
              FSectionsV[SectionNum].Contents.Add(pp);
            end;
          end;
      end else begin
      end;
    end else begin
      FSectionsV[SectionNum].RowsCount := 1;
      ContentColRow := Unassigned;
      ContentColRow := CC;
      if not (VarIsEmpty(ContentColRow) or VarIsNull(ContentColRow)) then
      begin
        pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
        FillChar(pp^, SizeOf(TExcelCellContent), 0);
        //pp^.ContentColRow := StrAlloc(Length(ContentColRow)+1);
        //pp^.ContentLen := Length(ContentColRow);
        //StrPCopy(pp^.ContentColRow, ContentColRow);
        pp^.ContentColRow := Unassigned;
        pp^.ContentColRow := ContentColRow;
        pp^.SectionRow := 1;
        pp^.SectionCol := 1;
        FSectionsV[SectionNum].Contents.Add(pp);
      end;
    end;
	end;
end;

procedure TExcelCOMReport.ReportExecute;
begin
  StartReportV;
end;

procedure TExcelCOMReport.ReportHorizontalyExecute;
begin
  StartReportH;  
end;

procedure TExcelCOMReport.ReportHorizontalyResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartReportH;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TExcelCOMReport.ReportResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartReportV;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TExcelCOMReport.SetConnectKind(const Value: TConnectKind);
begin
  // Should we validate that we have a RemoteMachineName for ckRemote ??
  FConnectKind := Value;
end;

procedure TExcelCOMReport.StartBiDirectionReport;
var
  bCheck: Boolean;
  SS: _Worksheet;
  FHReportSheet: _Worksheet;
begin
  try
    FLastDynSection := 13;
    FLastDynSectionH := 13;
    FTypeDiraction := tdHorizontaly;
    if Assigned(OnStartReport) then OnStartReport(self);
    if ActivateComObject then
    begin
      Excel.DisplayAlerts[0] := false;
      OpenInputBlank;
      if FCurrentWorkBook <> nil then
      begin
        try
          FCurrCol := 1;
          FCurrentWorkSheet := nil;
          FCurrentWorkSheet := (FCurrentWorkBook.Sheets.Item[InputSheet] as _Worksheet);
          FReportWorkBook := nil;
          FReportWorkBook := Excel.Workbooks.Add(NULL, 0);
          SS := nil;
          SS := (FReportWorkBook.Sheets.Item[1] as _Worksheet);
          FCurrentWorkSheet.Copy(NULL, SS, 0);
          FReportSheet := nil;
          FReportSheet := (SS.Next as _Worksheet);
          SS := nil;
          if OutputSheet <> '' then FReportSheet.Name := OutputSheet;
          while FReportSheet.Shapes.Count > 0 do
            FReportSheet.Shapes.Item(FReportSheet.Shapes.Count).Delete;
          FReportSheet.UsedRange[0].Clear;
          FReportSheet.UsedRange[0].ColumnWidth := FReportSheet.StandardWidth[0];
          PrepareSectionH('H_REPORTHEADER', 1);
          PrepareSectionH('H_GROUPHEADER', 2);
          PrepareSectionH('H_SUBGROUPHEADER', 3);
          PrepareSectionH('H_SUBSUBGROUPHEADER', 4);
          PrepareSectionH('H_SUPERSUBGROUPHEADER', 5);
          PrepareSectionH('H_BODY', 6);
          PrepareSectionH('H_SUPERSUBGROUPFOOTER', 7);
          PrepareSectionH('H_SUBSUBGROUPFOOTER', 8);
          PrepareSectionH('H_SUBGROUPFOOTER', 9);
          PrepareSectionH('H_GROUPFOOTER', 10);
          PrepareSectionH('H_REPORTFOOTER', 11);
          lWM_REP_SECTION_PREPARE := true;
          if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
          lWM_REP_SECTION_PREPARE := false;
          OutSectionH('H_REPORTHEADER', TRUE, 1);
          bCheck := True;
          if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
          while not bCheck do
          begin
            OutSectionH('H_GROUPHEADER', FALSE, 2);
            OutSectionH('H_SUBGROUPHEADER', FALSE, 3);
            OutSectionH('H_SUBSUBGROUPHEADER', FALSE, 4);
            OutSectionH('H_SUPERSUBGROUPHEADER', FALSE, 5);
            OutSectionH('H_BODY', TRUE, 6);
            OutSectionH('H_SUPERSUBGROUPFOOTER', FALSE, 7);
            OutSectionH('H_SUBSUBGROUPFOOTER', FALSE, 8);
            OutSectionH('H_SUBGROUPFOOTER', FALSE, 9);
            OutSectionH('H_GROUPFOOTER', FALSE, 10);
            bCheck := True;
            if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
          end;
          OutSectionH('H_REPORTFOOTER', TRUE, 11);
        finally
          while FReportWorkBook.Sheets.Count > 1 do
            if (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Name <> FReportSheet.Name then
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Delete(0)
            else
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count - 1] as _Worksheet).Delete(0);

          FHReportSheet := FReportSheet;
          //
          FTypeDiraction := tdVerticaly;
          if Assigned(OnStartReport) then OnStartReport(self);
          try
            FCurrRow := 1;
            FCurrentWorkBook := nil;
            FCurrentWorkBook := FHReportSheet.Parent as _Workbook;
            FCurrentWorkSheet := nil;
            FCurrentWorkSheet := FHReportSheet;
            FReportWorkBook := nil;
            FReportWorkBook := Excel.Workbooks.Add(NULL, 0);
            SS := nil;
            SS := (FReportWorkBook.Sheets.Item[1] as _Worksheet);
            FCurrentWorkSheet.Copy(NULL, SS, 0);
            FReportSheet := nil;
            FReportSheet := (SS.Next as _Worksheet);
            SS := nil;
            if OutputSheet <> '' then FReportSheet.Name := OutputSheet;
            while FReportSheet.Shapes.Count > 0 do
              FReportSheet.Shapes.Item(FReportSheet.Shapes.Count).Delete;
            FReportSheet.UsedRange[0].Clear;
            FReportSheet.UsedRange[0].RowHeight := FReportSheet.StandardHeight[0];
            PrepareSectionV('REPORTHEADER', 1);
            PrepareSectionV('GROUPHEADER', 2);
            PrepareSectionV('SUBGROUPHEADER', 3);
            PrepareSectionV('SUBSUBGROUPHEADER', 4);
            PrepareSectionV('SUPERSUBGROUPHEADER', 5);
            PrepareSectionV('BODY', 6);
            PrepareSectionV('SUPERSUBGROUPFOOTER', 7);
            PrepareSectionV('SUBSUBGROUPFOOTER', 8);
            PrepareSectionV('SUBGROUPFOOTER', 9);
            PrepareSectionV('GROUPFOOTER', 10);
            PrepareSectionV('REPORTFOOTER', 11);
            lWM_REP_SECTION_PREPARE := true;
            if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
            lWM_REP_SECTION_PREPARE := false;
            OutSectionV('REPORTHEADER', TRUE, 1);
            bCheck := True;
            if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
            while not bCheck do
            begin
              OutSectionV('GROUPHEADER', FALSE, 2);
              OutSectionV('SUBGROUPHEADER', FALSE, 3);
              OutSectionV('SUBSUBGROUPHEADER', FALSE, 4);
              OutSectionV('SUPERSUBGROUPHEADER', FALSE, 5);
              OutSectionV('BODY', TRUE, 6);
              OutSectionV('SUPERSUBGROUPFOOTER', FALSE, 7);
              OutSectionV('SUBSUBGROUPFOOTER', FALSE, 8);
              OutSectionV('SUBGROUPFOOTER', FALSE, 9);
              OutSectionV('GROUPFOOTER', FALSE, 10);
              bCheck := True;
              if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
            end;
            OutSectionV('REPORTFOOTER', TRUE, 11);
          finally
            while FReportWorkBook.Sheets.Count > 1 do
              if (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Name <> FReportSheet.Name then
                (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Delete(0)
              else
                (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count - 1] as _Worksheet).Delete(0);
            Excel.DisplayAlerts[0] := false;
            try
              FReportSheet.SaveAs(OutputBlank,
                                  NULL, NULL,
                                  NULL, NULL,
                                  NULL, NULL,
                                  NULL, NULL, 0);
              (FReportSheet.Parent as _Workbook).Close(false, NULL, NULL, 0);
            except
              OutputBlank := FReportWorkBook.Name;
            end;
            CloseSectionsV;
          end;
          //FTypeDiraction := tdNotDefined;
          //
          (FHReportSheet.Parent as _Workbook).Close(false, NULL, NULL, 0);
          CloseSectionsH;
        end;
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
    if Assigned(OnEndReport) then OnEndReport(self);
  finally
    SS := nil;
    FReportSheet := nil;
    FCurrentWorkSheet := nil;
    FReportWorkBook := nil;
    FCurrentWorkBook := nil;
    FExcel := nil;
    FTypeDiraction := tdNotDefined;
  end;
end;

procedure TExcelCOMReport.StartReportH;
var
  bCheck: Boolean;
  SS: _Worksheet;
begin
  try
    FLastDynSection := 13;
    FLastDynSectionH := 13;
    FTypeDiraction := tdHorizontaly;
    if Assigned(OnStartReport) then OnStartReport(self);
    if ActivateComObject then
    begin
      Excel.DisplayAlerts[0] := false;
      OpenInputBlank;
      if FCurrentWorkBook <> nil then
      begin
        try
          FCurrCol := 1;
          FCurrentWorkSheet := nil;
          FCurrentWorkSheet := (FCurrentWorkBook.Sheets.Item[InputSheet] as _Worksheet);
          FReportWorkBook := nil;
          FReportWorkBook := Excel.Workbooks.Add(NULL, 0);
          SS := nil;
          SS := (FReportWorkBook.Sheets.Item[1] as _Worksheet);
          FCurrentWorkSheet.Copy(NULL, SS, 0);
          FReportSheet := nil;
          FReportSheet := (SS.Next as _Worksheet);
          SS := nil;
          if OutputSheet <> '' then FReportSheet.Name := OutputSheet;
          while FReportSheet.Shapes.Count > 0 do
            FReportSheet.Shapes.Item(FReportSheet.Shapes.Count).Delete;
          FReportSheet.UsedRange[0].Clear;
          FReportSheet.UsedRange[0].ColumnWidth := FReportSheet.StandardWidth[0];
          PrepareSectionH('H_REPORTHEADER', 1);
          PrepareSectionH('H_GROUPHEADER', 2);
          PrepareSectionH('H_SUBGROUPHEADER', 3);
          PrepareSectionH('H_SUBSUBGROUPHEADER', 4);
          PrepareSectionH('H_SUPERSUBGROUPHEADER', 5);
          PrepareSectionH('H_BODY', 6);
          PrepareSectionH('H_SUPERSUBGROUPFOOTER', 7);
          PrepareSectionH('H_SUBSUBGROUPFOOTER', 8);
          PrepareSectionH('H_SUBGROUPFOOTER', 9);
          PrepareSectionH('H_GROUPFOOTER', 10);
          PrepareSectionH('H_REPORTFOOTER', 11);
          lWM_REP_SECTION_PREPARE := true;
          if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
          lWM_REP_SECTION_PREPARE := false;
          OutSectionH('H_REPORTHEADER', TRUE, 1);
          bCheck := True;
          if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
          while not bCheck do
          begin
            OutSectionH('H_GROUPHEADER', FALSE, 2);
            OutSectionH('H_SUBGROUPHEADER', FALSE, 3);
            OutSectionH('H_SUBSUBGROUPHEADER', FALSE, 4);
            OutSectionH('H_SUPERSUBGROUPHEADER', FALSE, 5);
            OutSectionH('H_BODY', TRUE, 6);
            OutSectionH('H_SUPERSUBGROUPFOOTER', FALSE, 7);
            OutSectionH('H_SUBSUBGROUPFOOTER', FALSE, 8);
            OutSectionH('H_SUBGROUPFOOTER', FALSE, 9);
            OutSectionH('H_GROUPFOOTER', FALSE, 10);
            bCheck := True;
            if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
          end;
          OutSectionH('H_REPORTFOOTER', TRUE, 11);
        finally
          while FReportWorkBook.Sheets.Count > 1 do
            if (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Name <> FReportSheet.Name then
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Delete(0)
            else
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count - 1] as _Worksheet).Delete(0);
          try
            Excel.DisplayAlerts[0] := false;
            FReportSheet.SaveAs(OutputBlank,
                                NULL, NULL,
                                NULL, NULL,
                                NULL, NULL,
                                NULL, NULL, 0);
          except
            OutputBlank := FReportWorkBook.Name;
          end;
          (FReportSheet.Parent as _Workbook).Close(false, NULL, NULL, 0);
          CloseSectionsH;
        end;
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
    if Assigned(OnEndReport) then OnEndReport(self);
  finally
    SS := nil;
    FReportSheet := nil;
    FCurrentWorkSheet := nil;
    FReportWorkBook := nil;
    FCurrentWorkBook := nil;
    FExcel := nil;
    FTypeDiraction := tdNotDefined;
  end;
end;

procedure TExcelCOMReport.StartReportV;
var
  bCheck: Boolean;
  SS: _Worksheet;
begin
  try
    FLastDynSection := 13;
    FLastDynSectionH := 13;
    FTypeDiraction := tdVerticaly;
    if Assigned(OnStartReport) then OnStartReport(self);
    if ActivateComObject then
    begin
      Excel.DisplayAlerts[0] := false;
      OpenInputBlank;
      if FCurrentWorkBook <> nil then
      begin
        try
          FCurrRow := 1;
          FCurrentWorkSheet := nil;
          FCurrentWorkSheet := (FCurrentWorkBook.Sheets.Item[InputSheet] as _Worksheet);
          FReportWorkBook := nil;
          FReportWorkBook := Excel.Workbooks.Add(NULL, 0);
          SS := nil;
          SS := (FReportWorkBook.Sheets.Item[1] as _Worksheet);
          FCurrentWorkSheet.Copy(NULL, SS, 0);
          FReportSheet := nil;
          FReportSheet := (SS.Next as _Worksheet);
          SS := nil;
          if OutputSheet <> '' then FReportSheet.Name := OutputSheet;
          while FReportSheet.Shapes.Count > 0 do
            FReportSheet.Shapes.Item(FReportSheet.Shapes.Count).Delete;
          FReportSheet.UsedRange[0].Clear;
          FReportSheet.UsedRange[0].RowHeight := FReportSheet.StandardHeight[0];
          PrepareSectionV('REPORTHEADER', 1);
          PrepareSectionV('GROUPHEADER', 2);
          PrepareSectionV('SUBGROUPHEADER', 3);
          PrepareSectionV('SUBSUBGROUPHEADER', 4);
          PrepareSectionV('SUPERSUBGROUPHEADER', 5);
          PrepareSectionV('BODY', 6);
          PrepareSectionV('SUPERSUBGROUPFOOTER', 7);
          PrepareSectionV('SUBSUBGROUPFOOTER', 8);
          PrepareSectionV('SUBGROUPFOOTER', 9);
          PrepareSectionV('GROUPFOOTER', 10);
          PrepareSectionV('REPORTFOOTER', 11);
          lWM_REP_SECTION_PREPARE := true;
          if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
          lWM_REP_SECTION_PREPARE := false;
          OutSectionV('REPORTHEADER', TRUE, 1);
          bCheck := True;
          if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
          while not bCheck do
          begin
            OutSectionV('GROUPHEADER', FALSE, 2);
            OutSectionV('SUBGROUPHEADER', FALSE, 3);
            OutSectionV('SUBSUBGROUPHEADER', FALSE, 4);
            OutSectionV('SUPERSUBGROUPHEADER', FALSE, 5);
            OutSectionV('BODY', TRUE, 6);
            OutSectionV('SUPERSUBGROUPFOOTER', FALSE, 7);
            OutSectionV('SUBSUBGROUPFOOTER', FALSE, 8);
            OutSectionV('SUBGROUPFOOTER', FALSE, 9);
            OutSectionV('GROUPFOOTER', FALSE, 10);
            bCheck := True;
            if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
          end;
          OutSectionV('REPORTFOOTER', TRUE, 11);
        finally
          while FReportWorkBook.Sheets.Count > 1 do
            if (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Name <> FReportSheet.Name then
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count] as _Worksheet).Delete(0)
            else
              (FReportWorkBook.Sheets.Item[FReportWorkBook.Sheets.Count - 1] as _Worksheet).Delete(0);
          Excel.DisplayAlerts[0] := false;
          try
            FReportSheet.SaveAs(OutputBlank,
                                NULL, NULL,
                                NULL, NULL,
                                NULL, NULL,
                                NULL, NULL, 0);
            (FReportSheet.Parent as _Workbook).Close(false, NULL, NULL, 0);
          except
            OutputBlank := FReportWorkBook.Name;
          end;
          CloseSectionsV;
        end;
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
    if Assigned(OnEndReport) then OnEndReport(self);
  finally
    SS := nil;
    FReportSheet := nil;
    FCurrentWorkSheet := nil;
    FReportWorkBook := nil;
    FCurrentWorkBook := nil;
    FExcel := nil;
    FTypeDiraction := tdNotDefined;
  end;
end;

end.
