unit DDEClient;

interface

uses
  Windows, SysUtils, Classes, Graphics, Forms, Controls, DDEML, StdCtrls, Dialogs;

const
  MaxDdeDataRequestBuffer = 1024;     //1K

type

  TDDEClient = class;

  TItemDataChange = procedure(sItem: String) of object;
  TItemDataChangeHot = procedure(sData, sItem: String) of object;
  TDDEErrorEvent = procedure(oError: Exception) of object;

  TDdeRequestData = packed record
	  pData: PChar;
	  pLen: LongWord;
	  pRet: Boolean;
  end;

  TAdvizeItem = record
    sItem: ShortString;
    hszItem: HSZ;
  end;
  pTAdvizeItem = ^TAdvizeItem;

  TExcelCell = record
    Col: LongWord;
    Row: LongWord;
  end;
  pTExcelCell = ^TExcelCell;

  TClearSheetEvent = procedure(var TopCell, BottomCell: TExcelCell) of object;

  TColWidthType = (cwtNotUsed, cwtHide, cwtUnHide, cwtUnHideBestFit);

  TExcelPattern = 0..18;
  TExcelColor = 0..56;

  TExcelHorizontalAlignment = ( haNotChanged,
                                haHORIZ_ALIGN_GENERAL,
                                haHORIZ_ALIGN_LEFT,
                                haHORIZ_ALIGN_CENTER,
                                haHORIZ_ALIGN_RIGHT,
                                haHORIZ_ALIGN_FILL,
                                haHORIZ_ALIGN_JUSTIFY,
                                haHORIZ_ALIGN_DENTER_ACROSS_SELECTION);

  TExcelWrap = (  wFalse,
                  wTrue,
                  wNotChanged );

  TExcelVerticalAlignment = ( vaNotChanged,
                              vaVERT_ALIGN_TOP,
                              vaVERT_ALIGN_CENTER,
                              vaVERT_ALIGN_BOTTOM,
                              vaVERT_ALIGN_JUSTIFY );

  TExcelOrientation = ( eoORIENTATION_HORIZONTAL,
                        eoORIENTATION_VERTICAL,
                        eoORIENTATION_UPWARD,
                        eoORIENTATION_DOWNWARD,
                        eoORIENTATION_AUTOMATIC,
                        eoNotChanged );
  TExcelLineStyle = (
                      lsNO_BORDER,
                      lsTHIN_LINE,
                      lsMEDIUM_LINE,
                      lsDASHED_LINE,
                      lsDOTTED_LINE,
                      lsTHICK_LINE,
                      lsDOUBLE_LINE,
                      lsHAIRLINE,
                      lsNotChanged  );


  TExcelUnderLine = (
                      ulNotChanged,
                      ulUNDERLINE_NONE,
                      ulUNDERLINE_SINGLE,
                      ulUNDERLINE_DOUBLE,
                      ulUNDERLINE_SINGLE_ACCOUNTING,
                      ulUNDERLINE_DOUBLE_ACCOUNTING );

  TExcelBackGround = (
                        bgNotChanged,
                        bgBACKGROUND_AUTOMATIC,
                        bgBACKGROUND_TRANSPERENT,
                        bgBACKGROUND_OPAQUE );

  TWorkBookInsertType = ( itNotChanged,
                          itWorksheet,
                          itChart,
                          itMacroSheet40,
                          itInternationalMacroSheet40,
                          itReserved,
                          itVisualBasicModule,
                          itDialog  );

{ TDdeMgrClient }

  TDdeMgrClient = class(TComponent)
  private
    FDdeInstId: LongInt;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    property DdeInstId: LongInt read FDdeInstId write FDdeInstId;

  end;

{ TDDEClient }

  TDDEClient = class(TComponent)
  private
    { Private declarations }
    FActive: Boolean;
    FDdeMgrClient: TDdeMgrClient;
    FWaitStat: Boolean;
    FAppName: String;
    FWaitLimit: LongWord;
    FTimeOutTransaction: LongWord;
    FErrorProcessed: Boolean;
    FCnvInfo: TConvInfo;
    RData: TDdeRequestData;
  	FHszApp: HSZ;
  	FHszTopic: HSZ;
    FDdeFmt: LongWord;
    FOnItemDataChange: TItemDataChange;
    FOnItemDataChangeHot: TItemDataChangeHot;
    FOnChange: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnOpen: TNotifyEvent;
    FOnError: TDDEErrorEvent;
    FRequestDataSize: LongWord;
    procedure SetRequestDataSize(const Value: LongWord);
    procedure SetDdeMgrClient(const Value: TDdeMgrClient);

  protected
    { Protected declarations }
    FStreamedActive: Boolean;
  	FDdeService: String;
  	FDdeTopic: String;
    FConv: HConv;
    FAdviseItem: TList;
    function ErrorCheck(): Exception;
    procedure XactComplete; virtual;
    procedure DataChange(hData: HDDEData; hsz2: HSZ); virtual;
    procedure SrvrDisconnect(); virtual;
    function ChangeLink(const App, Topic, Item: string): Boolean;
    function CreateDdeConv(FHszApp: HSZ; FHszTopic: HSZ): Boolean;
    procedure TerminateTransaction(Sender: TObject; sCmd: String); virtual;
    procedure Loaded; override;
    procedure SetActive(const Value: Boolean); virtual;
    procedure SetDdeTopic(const Value: String); virtual;
    function IsLoadedServiceApp: Boolean; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

  public
    { Public declarations }
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    function Exec(const Cmd: String): Boolean;
    function Exec1(const Cmd: String): Boolean;
    function Request(const Item: String): String;
    function Poke(const Item, Data: String): Boolean;

    procedure MgrNotify;

    procedure Close; virtual;
    procedure Open; virtual;
    procedure CloseLink;
    function SetLink(const Service, Topic: String): Boolean;
    function OpenLink: Boolean;
    function RequestData(const Item: String): PChar;
    function PokeData(const Item, Data: string): Boolean;
    function ExecuteMacro(const Cmd: String; waitFlg: Boolean): Boolean;
    function StartAdvise(const Item: string): Boolean;
    function StopAdvise(const Item: string): Boolean;

    property ValidRequestData: Boolean read RData.pRet;

  published
    { Published declarations }
    property Active: Boolean read FActive write SetActive;
    property DdeMgrClient: TDdeMgrClient read FDdeMgrClient write SetDdeMgrClient;
    property ServiceApplication: String read FAppName write FAppName;
    property WaitStat: Boolean read FWaitStat;
    property WaitLimit: LongWord read FWaitLimit write FWaitLimit;
    property TimeOutTransaction: LongWord read FTimeOutTransaction write FTimeOutTransaction;
    property ErrorProcessed: Boolean read FErrorProcessed write FErrorProcessed;
    property DDEService: String read FDdeService write FDdeService;
    property DDETopic: String read FDdeTopic write SetDdeTopic;
    property DdeFmt: LongWord read FDdeFmt;
    property RequestDataSize: LongWord read FRequestDataSize write SetRequestDataSize;

    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnItemDataChange: TItemDataChange read FOnItemDataChange write FOnItemDataChange;
    property OnItemDataChangeHot: TItemDataChangeHot read FOnItemDataChangeHot write FOnItemDataChangeHot;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnError: TDDEErrorEvent read FOnError write FOnError;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;

  end;

{ TExcelDDEClient }

  TExcelDDEClient = class(TDDEClient)
  private

		FIgnoreErrors: Boolean;
    FStreamedIgnoreErrors: Boolean;
		FRowChar:	String;
		FColChar:	String;
		FFileExt: String;
		FExecutable: String;
		FStripCRLF:	Boolean;

    FProtocolsList: TStringList;
    FEditEnvItemsList: TStringList;
    FTopicsList: TStringList;
    FFormatsList: TStringList;

    FSheetsOnCurrentBook: TStringList;

    procedure SetIgnoreErrors(const Value: Boolean);
    procedure LocateExcel; virtual;
    function GetSheetNumber: Integer;
    procedure SetSheetNumber(const Value: Integer);
    function GetSheetsCount: Integer;
    class function CountExcelInstance: LongInt;

  protected

    procedure Loaded; override;
    procedure TerminateTransaction(Sender: TObject; sCmd: String); override;
    function GetStrings(var List: TStringList; const Topic: string): TStringList;
    procedure SetActive(const Value: Boolean); override;
    function GetSelection: String;
    function GetSelectionAddress: String;
    function SwitchTopic(const NewTopic: String): Boolean;
    procedure SetDdeTopic(const Value: String); override;
    function GetCurrentWorkBook: String;
    procedure SetCurrentWorkBook(const Value: String);
    function GetCurrentSheet: String;
    procedure SetCurrentSheet(const Value: String);
    function IsLoadedServiceApp: Boolean; override;

  public

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    class function ParseSelEntry(const Sel: string; var TopLeft, BottomRight: TExcelCell): string;

    procedure Connect;
    procedure Disconnect;

    function SwitchToSystemTopic: Boolean;
    function SwitchTopicToCurrentSheet: Boolean;
    function GetFormats: TStringList;
    function GetTopics: TStringList;
    function GetProtocols: TStringList;
    function GetEditEnvItems: TStringList;

    function GetSheetsOnCurrentBook: TStringList;

    function CloseExcel: Boolean;
    function OpenWorkBook(const FileName: String): Boolean;
    function NewWorkBook: Boolean;
    function SetCellNumberFormat(const sFormat: String): Boolean;
    function CloseCurrentBook: Boolean;
    function CloseCurrentBookWithSave: Boolean;
    function Save: Boolean;
    function SaveAs(const FileName: String): Boolean;
    function AppMaximize: Boolean;
    function AppActivate: Boolean;
    function SetColWidth(Width: Real; sRef: String = ''; Standard: Boolean = False; Type_num: TColWidthType = cwtNotUsed; StandardNum: Real = -1): Boolean;
    function SetRowHeight(Height: Real; sRef: String = ''; Standard: Boolean = False; Type_num: TColWidthType = cwtNotUsed; StandardNum: Real = -1): Boolean;
    function SetColWidthToStandard(var TopCell, BottomCell: TExcelCell): Boolean;
    function SetRowHeightToStandard(var TopCell, BottomCell: TExcelCell): Boolean;
    function GetCellWidth: Integer;
    function GetLastCell(var LastCell: TExcelCell): Boolean;
    function Hide: Boolean;
    function UnHude: Boolean;
    function Move(deltaRow, deltaCol: LongWord): Boolean;
    function RunMacro(const sMacroFileName, MacroName: String): Boolean; overload;
    function RunMacro(const MacroName: String): Boolean; overload;
    function NewSheet: Boolean; overload;
    function NewSheet(const Type_Num: TWorkBookInsertType): Boolean; overload;
    function NewSheet(const Tamplate: String): Boolean; overload;
    function CopySheet(SourseSheet: String = ''; DestBook: String = ''; const Position: Integer = -1): Boolean;
    function MoveSheet(SourseSheet: String = ''; DestBook: String = ''; const Position: Integer = -1): Boolean;
    function DeleteSheet(SourseSheet: String = ''): Boolean;
    function ClearSheet(SourseSheet: String = ''; pProc: TClearSheetEvent = nil): Boolean;
    function EchoOn: Boolean;
    function EchoOff: Boolean;
    function DisableInput: Boolean;
    function EnableInput: Boolean;
    function RenameSheet(const OldName, NewName: string): Boolean;

    function Patterns(const apattern: TExcelPattern = 0;	const afore: TExcelColor = 0;	const aback: TExcelColor = 0;	const newui: Boolean = True): Boolean;
    function Alignment( horiz_align: TExcelHorizontalAlignment = haNotChanged;
                        wrap: TExcelWrap = wNotChanged;
                        vert_align: TExcelVerticalAlignment = vaNotChanged;
                        orientation: TExcelOrientation = eoNotChanged): Boolean;
    function Border(  outline: TExcelLineStyle = lsNotChanged;
                      left: TExcelLineStyle = lsNotChanged;
                      right: TExcelLineStyle = lsNotChanged;
                      top: TExcelLineStyle = lsNotChanged;
                      bottom: TExcelLineStyle = lsNotChanged;
                      shade: TExcelWrap = wNotChanged;
                      outline_color: TExcelColor = 0;
                      left_color: TExcelColor = 0;
                      right_color: TExcelColor = 0;
                      top_color: TExcelColor = 0;
                      bottom_color: TExcelColor = 0): Boolean;
    function FontProperties(
                              font: String = '';
                              font_style: String = '';
                              size: Integer = -1;
                              strikethrough: TExcelWrap = wNotChanged;
                              superscript: TExcelWrap = wNotChanged;
                              subscript: TExcelWrap = wNotChanged;
                              outline: TExcelWrap = wNotChanged;
                              shadow: TExcelWrap = wNotChanged;
                              underline: TExcelUnderLine = ulNotChanged;
                              color: TExcelColor = 0;
                              normal: TExcelWrap = wNotChanged;
                              background: TExcelBackGround = bgNotChanged;
                              start_char: Integer = -1;
                              char_count: Integer = -1
                            ): Boolean;

    function GetData: String; overload;
    function GetData(const Item: String): String; overload;
    function GetData(const Row, Col: LongWord): String; overload;
    function GetData(var TopCell, BottomCell: TExcelCell): String; overload;

    function PutData(const sData: String): Boolean; overload;
    function PutData(const sData: Variant): Boolean; overload;
    function PutData(const Item: String; const sData: String): Boolean; overload;
    function PutData(const Row, Col: LongWord; const sData: String): Boolean; overload;
    function PutData(const Row, Col: LongWord; const sData: Variant): Boolean; overload;
    function PutData(var TopCell, BottomCell: TExcelCell; const sData: String): Boolean; overload;

    function Select(const Item: String): Boolean; overload;
    function Select(const Row, Col: LongWord): Boolean; overload;
    function Select(const FromRow, FromCol, aRowCount, aColCount: LongWord): Boolean; overload;
    function Select(var TopCell, BottomCell: TExcelCell): Boolean; overload;

    function GetRectSelection(var TopCell, BottomCell: TExcelCell): String;
    function GetTopRowSelection: LongWord;
    function GetTopColSelection: LongWord;
    function GetBottomRowSelection: LongWord;
    function GetBottomColSelection: LongWord;

    property EditEnvItems: TStringList read GetEditEnvItems;
    property Formats: TStringList read GetFormats;
    property Protocols: TStringList read GetProtocols;
    property Topics: TStringList read GetTopics;

    property SheetsOnCurrentBook: TStringList read GetSheetsOnCurrentBook;

    property Selection: String read GetSelection;
    property TopRowSelection: LongWord    read  GetTopRowSelection;
    property TopColSelection: LongWord    read  GetTopColSelection;
    property BottomRowSelection: LongWord read  GetBottomRowSelection;
    property BottomColSelection: LongWord read  GetBottomColSelection;
    property CurrentWorkBook: String read GetCurrentWorkBook write SetCurrentWorkBook;
    property CurrentSheet: String read GetCurrentSheet write SetCurrentSheet;
    property CurrentSheetNumber: Integer read GetSheetNumber write SetSheetNumber;
    property SheetsCount: Integer read GetSheetsCount;
    property SelectionAddress: String read GetSelectionAddress;

  published

    property RowChar: String read FRowChar write FRowChar;
    property ColChar: String read FColChar write FColChar;
    property IgnoreErrors: Boolean read FIgnoreErrors write SetIgnoreErrors;
    property FileExt: String read FFileExt write FFileExt;
    property Executable: String read FExecutable write FExecutable;
    property StripCRLF: Boolean read FStripCRLF write FStripCRLF;

  end;

procedure Register;
function DdeRCallBack(CallType, Fmt: LongWord; Conv: HConv; hsz1, hsz2: HSZ; Data: HDDEData; Data1, Data2: LongWord): HDDEData; stdcall;
function Seconds(): Real;
function iif(l: Boolean; s1, s2: String): String; overload;
function RScan(const S: string; Chr: Char): Integer;
function LScan(const S: string; Chr: Char; StartPos: Integer = 0): Integer;
function StripQuotation(const AString: string): string;
procedure TABStringToStringList(TABString: PChar; List: TStringList);
function EnumInstanceWindows(AHandle: THandle; PInstanceCount: LPARAM): Boolean; stdcall;
function ToA1(const sCell: TExcelCell): String;

{$R VKReports.res}

implementation

procedure Register;
begin
  RegisterComponents('VK DDE', [TDdeMgrClient, TDDEClient, TExcelDDEClient]);
end;

function ToA1(const sCell: TExcelCell): String;
var
  q:  Byte;
begin
  q := Trunc(sCell.Col/26);
  if q <> 0 then
    Result := Chr(q + 64) + Chr(sCell.Col - q * 26 + 64) + Trim(IntToStr(sCell.Row))
  else
    Result := Chr(sCell.Col + 64) + Trim(IntToStr(sCell.Row));
end;

function Seconds(): Real;
begin
  Result := (DateTimeToTimeStamp(Time()).Time)/1000;
end;

function DdeRCallBack(CallType, Fmt: LongWord; Conv: HConv; hsz1, hsz2: HSZ; Data: HDDEData; Data1, Data2: LongWord): HDDEData; stdcall;
var
	ci:	TConvInfo;
	ddeCli: TDdeClient;
begin
	Result := 0;
	if Conv <> 0 then
  begin
		ci.cb := SizeOf(TConvInfo);
		if DdeQueryConvInfo(Conv, QID_SYNC, @ci) = 0 then Exit;
		ddeCli := TDdeClient(ci.hUser);
		if ddeCli <> nil then
    begin
			case CallType of
				XTYP_XACT_COMPLETE: ddeCli.XactComplete();
			  XTYP_ADVDATA: ddeCli.DataChange(Data, hsz2);
				XTYP_DISCONNECT: ddeCli.SrvrDisconnect();
			end;
		end;
	end;
end;

function iif(l: Boolean; s1, s2: String): String;
begin
  if l then
    result := s1
  else
    result := s2;
end;

function RScan(const S: string; Chr: Char): Integer;
begin
  Result := Length(s);
  while Result > 0 do
  begin
    if S[Result] = Chr then
      Break;
    Dec(Result);
  end;
end;

function LScan(const S: string; Chr: Char; StartPos: Integer): Integer;
begin
  Result := StartPos;
  while Result < Length(s) do
  begin
    if S[Result + 1] = Chr then
    begin
      Inc(Result);
      Exit;
    end;
    Inc(Result);
  end;
  Result := 0;
end;

function StripQuotation(const AString: string): string;
begin
  if (Length(AString) > 0) and
     (AString[Length(AString)] = #39) and (AString[1] = #39) then { test for ' }
    Result := Copy(AString, 2, Length(AString)-2)
  else
    Result := AString;
end;

{ procedure relies on List <> nil }
procedure TABStringToStringList(TABString: PChar; List: TStringList);
var
  StartPos: PChar;
  TabPos: PChar;
begin
  if TABString = nil then Exit;

  StartPos := TABString;
  TabPos := StrScan(StartPos, #9);
  while TabPos <> nil do
  begin
    TabPos[0] := #0; { replace #9 with #0 }
    List.Add(StrPas(StartPos)); { add format to list }
    StartPos := TabPos+1; { position to after TAB}
    TabPos := StrScan(StartPos, #9);
  end;

  if StrLen(StartPos) > 0 then
    List.Add(StrPas(StartPos));
end;

function EnumInstanceWindows(AHandle: THandle; PInstanceCount: LPARAM): Boolean; stdcall;
type
  PLongint = ^Longint;
var
  ClassName: array[0..8] of Char;
begin
  if GetClassName(AHandle, ClassName, SizeOf(ClassName)) = 0 then
    raise Exception.Create('WinApi Error');
  if StrComp(ClassName, 'XLMAIN') = 0 then
    Inc(PLongint(PInstanceCount)^);
  Result := (PLongint(PInstanceCount)^ < 2);
end;

{ TDDEClient }

function TDDEClient.ChangeLink(const App, Topic, Item: string): Boolean;
begin
	CloseLink();
	FDDEService := App;
	FDDETopic := Topic;
	Result := OpenLink();
end;

procedure TDDEClient.Close;
begin
  CloseLink();
  if Assigned(OnClose) then OnClose(self);
end;

procedure TDDEClient.CloseLink;
var
	OldConv: HConv;
	sSec: LongWord;
	oErr: Exception;
begin
  sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	 then //Exit If wait more then FWaitLimit
    begin
  		oErr := Exception.Create('DDE Client Error - ' + IntToStr(1002) + ' ' + 'End Wait limit while CloseLink');
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
  while (FAdviseItem.Count > 0) do
    StopAdvise(pTAdvizeItem(FAdviseItem.Items[FAdviseItem.Count - 1])^.sItem);
	if FConv <> 0 then
  begin
		OldConv := FConv;
		FConv := 0;
		DdeSetUserHandle(OldConv, QID_SYNC, 0);
		ErrorCheck();
		DdeDisconnect(OldConv);
		ErrorCheck();
	end;
	if FHszApp <> 0 then
  begin
    if Assigned(ddeMgrClient) then
		  DdeFreeStringHandle(ddeMgrClient.DdeInstId, FHszApp);
		ErrorCheck();
		FHszApp := 0;
	end;
	if FHszTopic <> 0 then
  begin
    if Assigned(ddeMgrClient) then
		  DdeFreeStringHandle(ddeMgrClient.DdeInstId, FHszTopic);
		ErrorCheck();
		FHszTopic := 0;
	end;
  FActive := false;
end;

constructor TDDEClient.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FActive := false;
  FStreamedActive := false;
	FWaitStat := false;
	FDdeService := '';
	FDdeTopic := '';
	FHszApp := 0;
	FHszTopic := 0;
	FConv := 0;
	FAppName := '';
  FRequestDataSize := MaxDdeDataRequestBuffer + 1;
	RData.pData := StrAlloc(FRequestDataSize);
	FWaitLimit :=  8;		//8 times to try execute DDE trancaction.
	FTimeOutTransaction  := 1000;		//In miliseconds or TIMEOUT_ASYNC ( DWORD(-1) )  for asyncronius transaction
	FErrorProcessed := false;
  FAdviseItem := TList.Create;
end;

function TDDEClient.CreateDdeConv(FHszApp, FHszTopic: HSZ): Boolean;
var
  Context: TConvContext;
begin
  FillChar(Context, SizeOf(Context), 0);
  with Context do
  begin
    cb := SizeOf(TConvConText);
    iCodePage := CP_WINANSI;
  end;
  if Assigned(ddeMgrClient) then
    FConv := DdeConnect(ddeMgrClient.DdeInstId, FHszApp, FHszTopic, @Context);
  Result := FConv <> 0;
  if Result then
  begin
    FCnvInfo.cb := sizeof(TConvInfo);
    DdeQueryConvInfo(FConv, QID_SYNC, @FCnvInfo);
    ErrorCheck();
    DdeSetUserHandle(FConv, QID_SYNC, LongInt(Self));
    ErrorCheck();
  end;
end;

procedure TDDEClient.DataChange(hData: HDDEData; hsz2: HSZ);
var
  i: Integer;
  sItem: String;
  Len: Integer;
  pData: Pointer;
begin
  if Assigned(OnChange) then OnChange(self);
  for i := 0 to (FAdviseItem.Count - 1) do
  begin
    if pTAdvizeItem(FAdviseItem.Items[i])^.HszItem = hsz2 then
    begin
      sItem := pTAdvizeItem(FAdviseItem.Items[i])^.sItem;
      if Assigned(OnItemDataChange) then OnItemDataChange(sItem);
      if hData <> 0 then
      begin
        pData := DdeAccessData(hData, @Len);
        if pData <> nil then
        try
          ErrorCheck();
          if Assigned(OnItemDataChangeHot) then OnItemDataChangeHot(String(PChar(pData)), sItem);
        finally
          DdeUnaccessData(hData);
          ErrorCheck();
        end;
      end;
      Break;
    end;
  end;
end;

destructor TDDEClient.Destroy;
begin
  CloseLink;
  while (FAdviseItem.Count > 0) do
    StopAdvise(pTAdvizeItem(FAdviseItem.Items[FAdviseItem.Count - 1])^.sItem);
  FAdviseItem.Destroy;
  StrDispose(RData.pData);
  inherited Destroy;
end;

function TDDEClient.ErrorCheck: Exception;
var
	dwErr: LongWord;
	errStr: String;
begin
  Result := nil;
  //Bug in Windows - not resets the error code to DMLERR_NO_ERROR.
  if Assigned(ddeMgrClient) then
	  dwErr := DdeGetLastError(ddeMgrClient.DdeInstId)
  else
    dwErr :=  DMLERR_DLL_NOT_INITIALIZED;
  //
	case dwErr of
		DMLERR_ADVACKTIMEOUT: errStr := 'DMLERR_ADVACKTIMEOUT';
		DMLERR_BUSY: errStr := 'DMLERR_BUSY';
		DMLERR_DATAACKTIMEOUT: errStr := 'DMLERR_DATAACKTIMEOUT';
		DMLERR_DLL_NOT_INITIALIZED: errStr := 'DMLERR_DLL_NOT_INITIALIZED';
		DMLERR_DLL_USAGE: errStr := 'DMLERR_DLL_USAGE';
		DMLERR_EXECACKTIMEOUT: errStr := 'DMLERR_EXECACKTIMEOUT';
		DMLERR_INVALIDPARAMETER: errStr := 'DMLERR_INVALIDPARAMETER';
		DMLERR_LOW_MEMORY: errStr := 'DMLERR_LOW_MEMORY';
		DMLERR_MEMORY_ERROR: errStr := 'DMLERR_MEMORY_ERROR';
		DMLERR_NO_CONV_ESTABLISHED: errStr := 'DMLERR_NO_CONV_ESTABLISHED';
		DMLERR_NOTPROCESSED: errStr := 'DMLERR_NOTPROCESSED';
		DMLERR_POKEACKTIMEOUT: errStr := 'DMLERR_POKEACKTIMEOUT';
		DMLERR_POSTMSG_FAILED: errStr := 'DMLERR_POSTMSG_FAILED';
		DMLERR_REENTRANCY: errStr := 'DMLERR_REENTRANCY';
		DMLERR_SERVER_DIED: errStr := 'DMLERR_SERVER_DIED';
		DMLERR_SYS_ERROR: errStr := 'DMLERR_SYS_ERROR';
		DMLERR_UNADVACKTIMEOUT: errStr := 'DMLERR_UNADVACKTIMEOUT';
		DMLERR_UNFOUND_QUEUE_ID: errStr := 'DMLERR_UNFOUND_QUEUE_ID';
	end;
	if dwErr <> DMLERR_NO_ERROR then
  begin
		Result := Exception.Create('DDE Client Error - ' + IntToStr(dwErr) + ' ' + errStr);
		if FErrorProcessed then raise Result else if Assigned(OnError) then OnError(Result);
	end;
end;

function TDDEClient.ExecuteMacro(const Cmd: String; waitFlg: Boolean): Boolean;
var
	sSec: LongWord;
	oErr: Exception;
  ddeRslt: LongInt;
  hszCmd: HDDEData;
  hdata: HDDEData;
//  i: Integer;
begin
	sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	then //Exit If  wait more then SELF:FWaitLimit
    begin
			oErr := Exception.Create('DDE Client Error 1005 End Wait limit while execute macro: ' + Cmd);
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
	Result := FALSE;
	if (FConv = 0) or FWaitStat then Exit;
  if not Assigned(ddeMgrClient) then Exit;
	hszCmd := DdeCreateDataHandle(ddeMgrClient.DdeInstId, PChar(Cmd), Length(Cmd) + 1, 0, 0, FDdeFmt, 0);
	ErrorCheck();
	if hszCmd = 0 then Exit;
	if waitFlg then FWaitStat := TRUE;
	hData := DdeClientTransaction(Pointer(hszCmd), LongWord(-1), FConv, 0, FDdeFmt, XTYP_EXECUTE, FTimeOutTransaction, @ddeRslt);
	//if hData = 0 then
  //begin
	//	for i := 1 to 3 do			// Try 3 times
  //  begin
	//		Application.ProcessMessages;
	//		hData := DdeClientTransaction(Pointer(hszCmd), LongWord(-1), FConv, 0, FDdeFmt, XTYP_EXECUTE, FTimeOutTransaction, @ddeRslt);
	//		if hData <> 0 then Break;
	//	end;
	//end;
	ErrorCheck();
	if FTimeOutTransaction <> TIMEOUT_ASYNC then
  begin
		/////////
		// For not asinchronic transaction
		FWaitStat := FALSE;
		TerminateTransaction(self, Cmd);
		//////
	end;
	if hData = 0 then
		FWaitStat := FALSE
	else
		Result := TRUE;
end;

function TDDEClient.Exec(const Cmd: String): Boolean;
var
  sSec: LongWord;
	oErr: Exception;
begin
	sSec := 0;
  Result := ExecuteMacro(Cmd,  True);
	while not Result do
  begin
		if not Active then break;
		Application.ProcessMessages;
		if sSec > FWaitLimit  then  //Exit If  wait more then FWaitLimit
    begin
			oErr := Exception.Create('Excel DDE Client Error ' + '1000 ' + 'End Wait limit while execute macro: ' + Cmd);
      if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
			Break;
		end;
    Result := ExecuteMacro(Cmd,  True);
    Inc(sSec);
	end;
end;

function TDDEClient.Exec1(const Cmd: String): Boolean;
begin
  Result := ExecuteMacro(Cmd,  FALSE);
end;

function TDDEClient.IsLoadedServiceApp: Boolean;
begin
  Result := false;
end;

procedure TDDEClient.Loaded;
begin
  inherited Loaded;
  if FStreamedActive then Active := true;
end;

procedure TDDEClient.MgrNotify;
begin
  Active := false;
  FDdeMgrClient := nil;
end;

procedure TDDEClient.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent =  ddeMgrClient then
    begin
      Active := False;
      ddeMgrClient := nil;
    end;
end;

procedure TDDEClient.Open;
begin
  if Assigned(OnOpen) then OnOpen(self);
end;

function TDDEClient.OpenLink: Boolean;
var
  CharVal: array[0..255] of Char;
  Res: Boolean;
	sSec: LongWord;
	oErr: Exception;
begin
	sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	then //Exit If  wait more then SELF:FWaitLimit
    begin
			oErr := Exception.Create('DDE Client Error ' + '1001 ' + 'End Wait limit while OpenLink');
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
  Result := False;
  FActive := False;
  if FConv <> 0 then Exit;
  if (Length(DdeService) = 0) and (Length(DdeTopic) = 0) then Exit;
  if FHszApp = 0 then
  begin
    StrPCopy(CharVal, DdeService);
    if Assigned(ddeMgrClient) then
      FHszApp := DdeCreateStringHandle(ddeMgrClient.DdeInstId, CharVal, CP_WINANSI);
    ErrorCheck();
  end;
  if FHszTopic = 0 then
  begin
    StrPCopy(CharVal, DdeTopic);
    if Assigned(ddeMgrClient) then
      FHszTopic := DdeCreateStringHandle(ddeMgrClient.DdeInstId, CharVal, CP_WINANSI);
    ErrorCheck();
  end;
  Res := CreateDdeConv(FHszApp, FHszTopic);
  if Not Res then
  begin
    if Not((Length(DdeService) = 0) and
      (Length(ServiceApplication) = 0)) then
    begin
      if Length(ServiceApplication) <> 0 then
        StrPCopy(CharVal, ServiceApplication)
      else
        StrPCopy(CharVal, DdeService + ' ' + DdeTopic);
      if not IsLoadedServiceApp then
      begin
        if WinExec(CharVal, SW_SHOWMINNOACTIVE) >= 32 then
          Res := CreateDdeConv(FHszApp, FHszTopic);
      end else Res := CreateDdeConv(FHszApp, FHszTopic);
      if Not Res then ErrorCheck();
    end;
  end;
  if Not Res then Exit;
  if FCnvInfo.wFmt <> 0 then FDdeFmt := FCnvInfo.wFmt
  else FDdeFmt := CF_TEXT;
  Open;
  Result := True;
  FActive := True;
end;

function TDDEClient.PokeData(const Item, Data: string): Boolean;
var
  hszDat: HDDEData;
  hdata: HDDEData;
  hszItem: HSZ;
  sSec: LongWord;
	oErr: Exception;
  ddeRslt: LongInt;
begin
	sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	then //Exit If  wait more then FWaitLimit
    begin
			oErr := Exception.Create('DDE Client Error 1004 End Wait limit while PokeData ' + Data + ' to Item ' + Item);
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
	Result := FALSE;
	IF (FConv = 0) or (FWaitStat) then Exit;
  if not Assigned(ddeMgrClient) then Exit;
	hszItem := DdeCreateStringHandle(ddeMgrClient.DdeInstId, PChar(Item), CP_WINANSI);
	ErrorCheck();
	IF hszItem = 0 then Exit;
	hszDat := DdeCreateDataHandle (ddeMgrClient.DdeInstId, PChar(Data), Length(Data) + 1, 0, hszItem, FDdeFmt, 0);
	ErrorCheck();
	IF hszDat <> 0 then
  begin
		hdata := DdeClientTransaction(Pointer(hszDat), LongWord(-1), FConv, hszItem, FDdeFmt, XTYP_POKE, FTimeOutTransaction, @ddeRslt);
		ErrorCheck();
		Result := ( hdata <> 0);
	end;
	DdeFreeStringHandle(ddeMgrClient.DdeInstId, hszItem);
	ErrorCheck();
end;

function TDDEClient.Poke(const Item, Data: String): Boolean;
var
  sSec: LongWord;
  oErr: Exception;
begin
	sSec := 0;
  Result := PokeData(Item, Data);
	while not Result do
  begin
		if not Active then Exit;
		Application.ProcessMessages;
		if sSec > FWaitLimit  then//Exit If  wait more then SELF:FWaitLimit
    begin
			oErr := Exception.Create('Excel DDE Client Error 1003 End Wait limit while Poke ' + Data + ' to Item ' + Item);
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  Exit;
		end;
    Result := PokeData(Item, Data);
    Inc(sSec);
	end;
end;

function TDDEClient.RequestData(const Item: String): PChar;
var
  hData: HDDEData;
  ddeRslt: LongInt;
  hItem: HSZ;
  pData: Pointer;
  Len: Integer;
  sSec: LongWord;
	oErr: Exception;
begin
	sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	then //Exit If  wait more then FWaitLimit
    begin
			oErr := Exception.Create('DDE Client Error 1006 End Wait limit on RequestDataBuffered from Item '+Item);
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
	RData.pLen := 0;
	RData.pRet := FALSE;
  Result := nil;
	if (FConv = 0) or (FWaitStat) then Exit;
  if not Assigned(ddeMgrClient) then Exit;
	hItem := DdeCreateStringHandle(ddeMgrClient.DdeInstId, PChar(Item), CP_WINANSI);
	ErrorCheck();
	if hItem <> 0 then
  begin
		hData := DdeClientTransaction(nil, 0, FConv, hItem, FDdeFmt, XTYP_REQUEST, FTimeOutTransaction, @ddeRslt);
		if hData <> 0 then
    try
      ErrorCheck();
      DdeFreeStringHandle(ddeMgrClient.DdeInstId, hItem);
      ErrorCheck();
      pData := DdeAccessData(hData, @Len);
      if pData <> nil then
      try
        ErrorCheck();
        Move(pData^, RData.pData^, Len);
        RData.pData[Len] := #0;
        RData.pLen := Len;
        Result := RData.pData;
        RData.pRet := TRUE;
      finally
        DdeUnaccessData(hData);
        ErrorCheck();
      end;
    finally
      DdeFreeDataHandle(hData);
      ErrorCheck();
    end;
	end;
end;

function TDDEClient.Request(const Item: String): String;
var
  sSec: LongWord;
	oErr: Exception;
begin
	Result := '';
	sSec := 0;
	while true do
  begin
		if not Active then Break;
		Result := String(RequestData(Item));
		if ValidRequestData then
			Break
		else
    begin
      Application.ProcessMessages;
      if sSec > FWaitLimit  then  //Exit If  wait more then FWaitLimit
      begin
        oErr := Exception.Create('Excel DDE Client Error ' + '1001 ' + 'End Wait limit while request data from Item: ' + Item);
        if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
        Break;
      end;
		end;
    Inc(sSec);
	end;
end;

procedure TDDEClient.SetActive(const Value: Boolean);
begin
  if ( csReading in ComponentState ) then FStreamedActive := Value
  else
  begin
    if Value then
    begin
      CloseLink;
      OpenLink;
    end else CloseLink;
  end;
end;

procedure TDDEClient.SetDdeMgrClient(const Value: TDdeMgrClient);
var
  OldActive: Boolean;
begin
  OldActive := Active;
  Active := false;
  FDdeMgrClient := Value;
  Active := OldActive;
end;

procedure TDDEClient.SetDdeTopic(const Value: String);
begin
  FDdeTopic := Value;
end;

function TDDEClient.SetLink(const Service, Topic: string): Boolean;
var
  sSec: LongWord;
	oErr: Exception;
begin
	sSec := 0;
	while FWaitStat do
  begin
		Application.ProcessMessages;
		if sSec > FWaitLimit	then //Exit If  wait more then FWaitLimit
    begin
			oErr := Exception.Create('DDE Client Error ' + '1000 ' + 'End Wait limit while SetLink');
  		if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
		  break;
		end;
    Inc(sSec);
	end;
  Result := ChangeLink(Service, Topic, '')
end;

procedure TDDEClient.SetRequestDataSize(const Value: LongWord);
begin
  FRequestDataSize := Value + 1;
  StrDispose(RData.pData);
	RData.pData := StrAlloc(FRequestDataSize);
end;

procedure TDDEClient.SrvrDisconnect;
begin
	XactComplete();
	if FConv <> 0 then Close();
end;

function TDDEClient.StartAdvise(const Item: string): Boolean;
var
  i: Integer;
  ddeRslt: LongInt;
begin
  Result := False;
  if not Assigned(ddeMgrClient) then Exit;
  i := FAdviseItem.Add(Pointer(StrAlloc(SizeOf(TAdvizeItem))));
  pTAdvizeItem(FAdviseItem.Items[i])^.sItem := Item;
  pTAdvizeItem(FAdviseItem.Items[i])^.hszItem := DdeCreateStringHandle(ddeMgrClient.DdeInstId, PChar(Item), CP_WINANSI);
  DdeClientTransaction(nil, LongWord(-1), FConv, pTAdvizeItem(FAdviseItem.Items[i])^.hszItem, FDdeFmt, XTYP_ADVSTART or XTYPF_NODATA, TIMEOUT_ASYNC, @ddeRslt);
  Result := True;
end;

function TDDEClient.StopAdvise(const Item: string): Boolean;
var
  i: Integer;
  ddeRslt: LongInt;
begin
  for i := 0 to (FAdviseItem.Count - 1) do
  begin
    if pTAdvizeItem(FAdviseItem.Items[i])^.sItem = Item then
    begin
      //
      if pTAdvizeItem(FAdviseItem.Items[i])^.HszItem <> 0 then
      begin
        DdeClientTransaction(nil, LongWord(-1), FConv, pTAdvizeItem(FAdviseItem.Items[i])^.HszItem, FDdeFmt, XTYP_ADVSTOP, TIMEOUT_ASYNC, @ddeRslt);
        if Assigned(ddeMgrClient) then
          DdeFreeStringHandle(ddeMgrClient.DdeInstId, pTAdvizeItem(FAdviseItem.Items[i])^.HszItem);
        pTAdvizeItem(FAdviseItem.Items[i])^.HszItem := 0;
      end;
      //
      StrDispose(PChar(FAdviseItem.Items[i]));
      FAdviseItem.Delete(i);
      Break;
    end;
  end;
  Result := True;
end;

procedure TDDEClient.TerminateTransaction(Sender: TObject; sCmd: String);
begin
end;

procedure TDDEClient.XactComplete;
begin
	FWaitStat := false;
end;

{ TDdeMgrClient }

constructor TDdeMgrClient.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
	FDdeInstId := 0;
	if DdeInitialize(FDdeInstId, DdeRCallBack, APPCMD_CLIENTONLY, 0) <>  DMLERR_NO_ERROR then
			raise Exception.Create('DDE Initialization Error');
end;

destructor TDdeMgrClient.Destroy;
begin
  DdeUninitialize(FDdeInstId);
  inherited Destroy;
end;

{ TExcelDDEClient }

function TExcelDDEClient.AppActivate: Boolean;
begin
  Result := Exec('[APP.ACTIVATE]');
end;

function TExcelDDEClient.AppMaximize: Boolean;
begin
  Result := Exec('[APP.MAXIMIZE]');
end;

function TExcelDDEClient.CloseCurrentBook: Boolean;
begin
  Result := Exec1('[CLOSE(0)]')
end;

function TExcelDDEClient.CloseCurrentBookWithSave: Boolean;
begin
  Result := Exec1('[CLOSE(1)]');
end;

function TExcelDDEClient.CloseExcel: Boolean;
begin
  Result := Exec('[QUIT]');
end;

procedure TExcelDDEClient.Connect;
var
  oErr: Exception;
begin
  if not Active then
  begin
    CloseLink;
    if OpenLink then Exit;
    LocateExcel;
    if not OpenLink then
      if not OpenLink then
      begin
  			oErr := Exception.Create('Excel not started!');
    		if ErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
      end;
  end;
end;

constructor TExcelDDEClient.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  Disconnect;
	FIgnoreErrors := false;
	FRowChar := 'R';
	FColChar := 'C';
	FFileExt := '.XLS';
	FExecutable := 'EXCEL.EXE';
  ServiceApplication := 'EXCEL';
	FStripCRLF := true;
  FDDEService := 'Excel';
  FDDETopic := 'System';
end;

destructor TExcelDDEClient.Destroy;
begin
  Disconnect;
  if Assigned(FProtocolsList) then FProtocolsList.Destroy;
  if Assigned(FEditEnvItemsList) then FEditEnvItemsList.Destroy;
  if Assigned(FTopicsList) then FTopicsList.Destroy;
  if Assigned(FFormatsList) then FFormatsList.Destroy;
  if Assigned(FSheetsOnCurrentBook) then FSheetsOnCurrentBook.Destroy;
  inherited Destroy;
end;

procedure TExcelDDEClient.Disconnect;
begin
  if Active then
    CloseLink;
end;

function TExcelDDEClient.GetBottomColSelection: LongWord;
var
  TopCell, BottomCell: TExcelCell;
begin
  GetRectSelection(TopCell, BottomCell);
  Result := BottomCell.Col;
end;

function TExcelDDEClient.GetBottomRowSelection: LongWord;
var
  TopCell, BottomCell: TExcelCell;
begin
  GetRectSelection(TopCell, BottomCell);
  Result := BottomCell.Row;
end;

function TExcelDDEClient.GetCurrentSheet: String;
var
  ExclPos: Integer;
begin
  if Active then
  begin
    Result := GetSelection;
    ExclPos := RScan(Result, '!');
    if ExclPos > 0 then
      Delete(Result, ExclPos, Length(Result))
    else
      Result := '';
    Result := StripQuotation(Result);
    ExclPos := RScan(Result, ']');
    if ExclPos > 0 then
      Delete(Result, 1, ExclPos);
  end else Result := '';
  Result := StripQuotation(Result);
end;

function TExcelDDEClient.GetCurrentWorkBook: String;
var
  ExclPos, ExclPos2: Integer;
begin
  if Active then
  begin
    Result := GetSelection;
    ExclPos := RScan(Result, ']');
    ExclPos2 := LScan(Result, '[');
    if ExclPos > 0 then
      Result := Copy(Result, ExclPos2 + 1, ExclPos - ExclPos2 - 1);
  end else Result := '';
end;

function TExcelDDEClient.GetData: String;
var
  CRLFPos: Integer;
begin
  Result := Request(SelectionAddress);
  if FStripCRLF and (Length(Result) > 1) then
  begin
    CRLFPos := Length(Result)-1;
    if (Result[CRLFPos] = #13) and (Result[succ(CRLFPos)] = #10) then
      Delete(Result, CRLFPos, 2);
  end;
end;

function TExcelDDEClient.GetData(const Item: String): String;
var
  CRLFPos: Integer;
begin
  Result := Request(Item);
  if FStripCRLF and (Length(Result) > 1) then
  begin
    CRLFPos := Length(Result)-1;
    if (Result[CRLFPos] = #13) and (Result[succ(CRLFPos)] = #10) then
      Delete(Result, CRLFPos, 2);
  end;
end;

function TExcelDDEClient.GetData(const Row, Col: LongWord): String;
var
  CRLFPos: Integer;
begin
  Result := Request(Format('%s%d%s%d', [RowChar, Row, ColChar, Col]));
  if FStripCRLF and (Length(Result) > 1) then
  begin
    CRLFPos := Length(Result)-1;
    if (Result[CRLFPos] = #13) and (Result[succ(CRLFPos)] = #10) then
      Delete(Result, CRLFPos, 2);
  end;
end;

function TExcelDDEClient.GetData(var TopCell, BottomCell: TExcelCell): String;
var
  CRLFPos: Integer;
begin
  Result := Request(Format('%s%d%s%d:%s%d%s%d', [RowChar, TopCell.Row, ColChar, TopCell.Col, RowChar, BottomCell.Row, ColChar, BottomCell.Col]));
  if FStripCRLF and (Length(Result) > 1) then
  begin
    CRLFPos := Length(Result)-1;
    if (Result[CRLFPos] = #13) and (Result[succ(CRLFPos)] = #10) then
      Delete(Result, CRLFPos, 2);
  end;
end;

function TExcelDDEClient.GetRectSelection(var TopCell,
  BottomCell: TExcelCell): String;
begin
  Result := ParseSelEntry(Selection, TopCell, BottomCell);
end;

function TExcelDDEClient.GetSelection: String;
var
  OldTopic: String;
begin
  OldTopic := DDETopic;
	SwitchToSystemTopic;
	Result := Request('Selection');
  if DDETopic <> OldTopic then
    DDETopic := OldTopic;
end;

function TExcelDDEClient.GetTopColSelection: LongWord;
var
  TopCell, BottomCell: TExcelCell;
begin
  GetRectSelection(TopCell, BottomCell);
  Result := TopCell.Col;
end;

function TExcelDDEClient.GetTopRowSelection: LongWord;
var
  TopCell, BottomCell: TExcelCell;
begin
  GetRectSelection(TopCell, BottomCell);
  Result := TopCell.Row;
end;

procedure TExcelDDEClient.Loaded;
begin
  inherited Loaded;
  if FStreamedIgnoreErrors then IgnoreErrors := true;
end;

procedure TExcelDDEClient.LocateExcel;
const
  BufferSize = 540;
var
  Buffer: PChar;
  StringPosition: PChar;
  ReturnedData: Longint;
  OldServiceApp: String;
begin
  OldServiceApp := ServiceApplication;
  Buffer := StrAlloc(BufferSize);
  try
    ReturnedData := BufferSize;
    StrPCopy(Buffer, FFileExt);
    RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData);
    if StrLen(Buffer) > 0 then
    begin
      StrCat(Buffer, '\shell\Open\command');
      ReturnedData := BufferSize;
      if RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData) = ERROR_SUCCESS then
      begin
        StringPosition := StrUpper(Buffer);
        while StrPos(StringPosition+1, PChar(FExecutable)) <> nil do
          StringPosition := StrPos(StringPosition+1, PChar(FExecutable));
        StringPosition[Length(FExecutable)] := #0;
        StringPosition := StrScan(Buffer, '"');
        if StringPosition <> nil then
        begin
          ReturnedData :=  Longint(StrLen(Buffer)) - (StringPosition-Buffer) - 1;
          StrMove(Buffer, Buffer + (StringPosition-Buffer) + 1, ReturnedData);
          Buffer[ReturnedData] := #0;
        end;
        ServiceApplication := StrPas(Buffer);
        if not FileExists(ServiceApplication) then ServiceApplication := OldServiceApp;
      end;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

function TExcelDDEClient.NewWorkBook: Boolean;
begin
  Result := Exec('[NEW(1)]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.OpenWorkBook(const FileName: String): Boolean;
begin
  Result := Exec('[OPEN("'+FileName+'")]');
end;

class function TExcelDDEClient.ParseSelEntry(const Sel: string; var TopLeft, BottomRight: TExcelCell): string;

  { parse cell part into number and remove it }
  function GetNumber(var AString: string): LongWord;
  var
    CharPos: Integer;
  begin
    CharPos := 2; { this is specialized code that knows that the first Char is non-numeric }
    while (CharPos <= Length(AString)) and (AString[CharPos] in ['0'..'9']) do
      Inc(CharPos);
    Dec(CharPos);
    { convert to number }
    Result := LongWord(StrToInt(Copy(AString, 2, CharPos-1)));
    { remove parsed part from string }
    Delete(AString, 1, CharPos);
  end;

var
  SeparatorPos: Integer;
  CellAddress: string;
  BottomInitialized: Boolean;
  oErr: Exception;
begin
  BottomInitialized := False;
  TopLeft.Col := 0;  BottomRight.Col := 0;
  TopLeft.Row := 0;  BottomRight.Row := 0;
  Result := Sel;
  SeparatorPos := RScan(Result, '!'); { find the separator }
  if SeparatorPos <> 0 then
  begin
    { copy the cell part into CellAddress}
    CellAddress := Copy(Result, succ(SeparatorPos), Length(Result));
    { and remove it from the existing string - return value }
    Delete(Result, SeparatorPos, Length(Result));
    Result := StripQuotation(Result);
  end else begin
    CellAddress := Sel;
    Result := StripQuotation(Result);
  end;
  { at this point we know the following:
      FRowChar and FColChar contain valid identifiers for an Excel row / column }
  { find the colon Char (R1C10:R20C15) -> array selection; or the semicolon
     -> multiple selection (R1C10;R20C40) ==> ";" + ":'

      R1C10;R20C20   [multiple selection]

      R1C10:R20C20   [array]

      R1C10

      R1             [single row]
      C1             [single column]

  }
  { find multiple selections which we are not parsing !}
  { it is safe to use "Pos" here, as a potential semicolon in the name part will
    have been removed by now }
  SeparatorPos := Pos(';', CellAddress);
  if SeparatorPos > 0 then
    { remove the parts we are not going to parse }
    Delete(CellAddress, SeparatorPos, Length(CellAddress));
  { there must be always at least ONE valid Char }
  if CellAddress[1] = 'R' then
    TopLeft.Row := GetNumber(CellAddress)
  else
  if CellAddress[1] = 'C' then
    TopLeft.Col := GetNumber(CellAddress);

  if Length(CellAddress) > 0 then { still something left ?}
    { assume that Excel does not return garbage }
    if CellAddress[1] = 'R' then
      TopLeft.Row := GetNumber(CellAddress)
    else
    if CellAddress[1] = 'C' then
      TopLeft.Col := GetNumber(CellAddress);
    //else
    //  begin
    //    oErr := Exception.Create('Excel DDE Client Error : Bad Cell format');
    //    raise oErr;
    //    //if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
    //  end;
  { at this point we have parsed the left part of an array selection, so
    that at most something like ":R1..." is left over }
  { test whether we do have an array indeed (otherwise we have a problem !) }
  if Length(CellAddress) > 0 then
    if CellAddress[1] <> ':' then
      begin
        oErr := Exception.Create('Excel DDE Client Error : Bad Cell format');
        raise oErr;
        //if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
      end
    else
    begin
      if Length(CellAddress) < 2 then
        begin
          oErr := Exception.Create('Excel DDE Client Error : Bad Cell format');
          raise oErr;
          //if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
        end;
      Delete(CellAddress, 1, 1); { remove colon }
      { what follows is effectively the same code as above, only
        with "BottomRight" instead of TopLeft
        Consider moving that into a local procedure to
        improve maintainability }
      { there must be always at least ONE valid Char }
      if CellAddress[1] = 'R' then
        BottomRight.Row := GetNumber(CellAddress)
      else
      if CellAddress[1] = 'C' then
        BottomRight.Col := GetNumber(CellAddress);
      if Length(CellAddress) > 0 then { still something left ?}
        { assume that Excel does not return garbage }
        if CellAddress[1] = 'R' then
          BottomRight.Row := GetNumber(CellAddress)
        else
        if CellAddress[1] = 'C' then
          BottomRight.Col := GetNumber(CellAddress)
        else
          begin
            oErr := Exception.Create('Excel DDE Client Error : Bad Cell format');
            raise oErr;
            //if FErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
          end;
      BottomInitialized := True;
    end;
  { we have transferred all the text into the TopLeft + BottomRight record;
    now handle the special cases }
  if not BottomInitialized then
    BottomRight := TopLeft;

end;

function TExcelDDEClient.PutData(const Row, Col: LongWord;
  const sData: String): Boolean;
begin
  Result := Poke(Format(RowChar+'%d'+ColChar+'%d', [Row, Col]), sData);
end;

function TExcelDDEClient.PutData(const Item, sData: String): Boolean;
begin
  Result := Poke(Item, sData);
end;

function TExcelDDEClient.PutData(const sData: String): Boolean;
begin
  Result := Poke(SelectionAddress, sData);
end;

function TExcelDDEClient.PutData(var TopCell, BottomCell: TExcelCell; const sData: String): Boolean;
begin
  Result := Poke(Format(RowChar+'%d'+ColChar+'%d:'+RowChar+'%d'+ColChar+'%d', [TopCell.Row, TopCell.Col, BottomCell.Row, BottomCell.Col]), sData);
end;

function TExcelDDEClient.Save: Boolean;
begin
  Result := Exec('[SAVE()]');
end;

function TExcelDDEClient.SaveAs(const FileName: String): Boolean;
begin
  Result := Exec('[SAVE.AS("'+Trim(FileName)+'")]')
end;

function TExcelDDEClient.Select(const Item: String): Boolean;
begin
  Result := Exec('[SELECT("'+Item+'")]');
end;

function TExcelDDEClient.Select(const Row, Col: LongWord): Boolean;
begin
  Result := False;
	if (Row <> 0) and (Col <> 0) then
		Result := Exec('[SELECT("'+RowChar+Trim(IntToStr(Row))+ColChar+Trim(IntToStr(Col))+'")]');
end;

function TExcelDDEClient.Select(const FromRow, FromCol, aRowCount,
  aColCount: LongWord): Boolean;
begin
  Result := False;
	if (FromRow <> 0) and (FromCol <> 0) and (ARowCount > 0) and (AColCount > 0) then
		Result := Exec('[SELECT("'+RowChar+Trim(IntToStr(FromRow))+ColChar+Trim(IntToStr(FromCol))+':'+RowChar+Trim(IntToStr(FromRow + ARowCount - 1))+ColChar+Trim(IntToStr(FromCol + AColCount - 1))+'")]')
end;

function TExcelDDEClient.Select(var TopCell,
  BottomCell: TExcelCell): Boolean;
begin
  Result := False;
	if (TopCell.Row <> 0) and (TopCell.Col <> 0) and (BottomCell.Row <> 0) and (BottomCell.Col <> 0) then
		Result := Exec('[SELECT("'+RowChar+Trim(IntToStr(TopCell.Row))+ColChar+Trim(IntToStr(TopCell.Col))+':'+RowChar+Trim(IntToStr(BottomCell.Row))+ColChar+Trim(IntToStr(BottomCell.Col))+'")]')
end;

procedure TExcelDDEClient.SetActive(const Value: Boolean);
begin
  if ( csReading in ComponentState ) then FStreamedActive := Value
  else
  begin
    if Value then
      Connect
    else Disconnect;
  end;
end;

function TExcelDDEClient.SetCellNumberFormat(
  const sFormat: String): Boolean;
begin
  Result := Exec('[FORMAT.NUMBER("'+sFormat+'")]');
end;

procedure TExcelDDEClient.SetCurrentSheet(const Value: String);
begin
  Exec('[WORKBOOK.SELECT("' + UpperCase(Trim(Value)) + '")]');
  SwitchTopicToCurrentSheet;
end;

procedure TExcelDDEClient.SetCurrentWorkBook(const Value: String);
begin
  Exec('[WORKBOOK.ACTIVATE("' + UpperCase(Trim(Value)) + '")]');
  SwitchTopicToCurrentSheet;
end;

procedure TExcelDDEClient.SetDDETopic(const Value: String);
begin
  SwitchTopic(Value);
end;

procedure TExcelDDEClient.SetIgnoreErrors(const Value: Boolean);
begin
  if ( csReading in ComponentState ) then FStreamedIgnoreErrors := Value
  else
  begin
    if Active then
    begin
      FIgnoreErrors := Value;
      Exec('[ERROR('+iif(not FIgnoreErrors, '1', '0')+')]');
    end else begin
      FIgnoreErrors := False;
    end;
  end;
end;

function TExcelDDEClient.SwitchTopic(const NewTopic: String): Boolean;
begin
  Result := True;
	if DdeTopic <> NewTopic then
		Result := SetLink('Excel', NewTopic);
end;

function TExcelDDEClient.SwitchTopicToCurrentSheet: Boolean;
begin
  Result := SwitchTopic(CurrentSheet);
end;

function TExcelDDEClient.SwitchToSystemTopic: Boolean;
begin
  Result := SwitchTopic('System');
end;

procedure TExcelDDEClient.TerminateTransaction(Sender: TObject; sCmd: String);
begin
	if (not Active) and ( (sCmd = '[CLOSE(0)]') or (sCmd = '[CLOSE(1)]') or (StrPos(PChar(sCmd), '[OPEN(') <> nil) ) then
		Active := True;
end;

function TExcelDDEClient.GetSelectionAddress: String;
var
  SeparatorPos: Integer;
begin
  Result := Selection;
  SeparatorPos := RScan(Result, '!');
  if SeparatorPos <> 0 then
  begin
    Result := Copy(Result, succ(SeparatorPos), Length(Result));
    Delete(Result, SeparatorPos, Length(Result));
    Result := StripQuotation(Result);
  end;
end;

function TExcelDDEClient.GetFormats: TStringList;
begin
  Result := GetStrings(FFormatsList, 'Formats');
end;

function TExcelDDEClient.GetTopics: TStringList;
begin
  Result := GetStrings(FTopicsList, 'Topics');
end;

function TExcelDDEClient.GetProtocols: TStringList;
begin
  Result := GetStrings(FProtocolsList, 'Protocols');
end;

function TExcelDDEClient.GetEditEnvItems: TStringList;
begin
  Result := GetStrings(FEditEnvItemsList, 'EditEnvItems');
end;

function TExcelDDEClient.GetStrings(var List: TStringList; const Topic: string): TStringList;
var
  Reply: String;
  OldTopic: String;
begin
  if not Assigned(List) then
    List := TStringList.Create
  else
    List.Clear;
  OldTopic := DDETopic;
  SwitchToSystemTopic;
  Reply := Request(Topic);
  TABStringToStringList(PChar(Reply), List);
  if DDETopic <> OldTopic then
    DDETopic := OldTopic;
  Result := List;
end;

function TExcelDDEClient.SetColWidth(Width: Real; sRef: String; Standard: Boolean; Type_num: TColWidthType; StandardNum: Real): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..5] of String;
	i, j: LongWord;
  l: Boolean;
  dc: Char;
begin
	//Set the width columns for current selection
  dc := DecimalSeparator;
  DecimalSeparator := '.';
  ParamArg := '';
  l := True;
  j := 5;
	ParamList[5] := iif(Trunc(StandardNum) = -1, '-1', Format('%f', [StandardNum]));
  if (ParamList[5] = '-1') and l then Dec(j) else l := false;
	ParamList[4] := iif(Type_num = cwtNotUsed, '', Trim(IntToStr(Ord(Type_num))));
  if (ParamList[4] = '') and l then Dec(j) else l := false;
	ParamList[3] := iif(Standard, 'TRUE', 'FALSE');
  if (ParamList[3] = '') and l then Dec(j) else l := false;
	ParamList[2] := iif(sRef <> '', '"'+sRef+'"', '');
  if (ParamList[2] = '') and l then Dec(j);
	ParamList[1] := iif(Trunc(Width) = -1, '-1', Format('%f', [Width]));
  if (ParamList[1] = '-1') then
  begin
    j := 3;
    ParamList[3] := 'TRUE';
  end;
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
  Result := Exec(Format('[COLUMN.WIDTH(%s)]', [ParamArg]));
  DecimalSeparator := dc;
end;

function TExcelDDEClient.SetRowHeight(Height: Real; sRef: String; Standard: Boolean; Type_num: TColWidthType; StandardNum: Real): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..5] of String;
	i, j: LongWord;
  l: Boolean;
  dc: Char;
begin
	//Set the height rows for current selection
  dc := DecimalSeparator;
  DecimalSeparator := '.';
  ParamArg := '';
  l := True;
  j := 5;
	ParamList[5] := iif(StandardNum = -1, '-1', trim(Floattostr(StandardNum)));
  if (ParamList[5] = '-1') and l then Dec(j) else l := false;
	ParamList[4] := iif(Type_num = cwtNotUsed, '', Trim(IntToStr(Ord(Type_num))));
  if (ParamList[4] = '') and l then Dec(j) else l := false;
	ParamList[3] := iif(Standard, 'TRUE', 'FALSE');
  if (ParamList[3] = '') and l then Dec(j) else l := false;
	ParamList[2] := iif(sRef <> '', '"'+sRef+'"', '');
  if (ParamList[2] = '') and l then Dec(j);
	ParamList[1] := iif(Trunc(Height) = -1, '-1', Format('%f', [Height]));
  if (ParamList[1] = '-1') then
  begin
    j := 3;
    ParamList[3] := 'TRUE';
  end;
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
  Result := Exec(Format('[ROW.HEIGHT(%s)]', [ParamArg]));
  DecimalSeparator := dc;
end;

function TExcelDDEClient.GetLastCell(var LastCell: TExcelCell): Boolean;
var
  OldSelection: String;
  TopCell, BottomCell: TExcelCell;
begin
	OldSelection := SelectionAddress;
	Result := Exec('[SELECT.LAST.CELL()]');
  GetRectSelection(TopCell, BottomCell);
  LastCell.Row := TopCell.Row;
  LastCell.Col := TopCell.Col;
  Select(OldSelection);
end;

function TExcelDDEClient.Hide: Boolean;
begin
  Result := Exec('[HIDE()]');
end;

function TExcelDDEClient.UnHude: Boolean;
begin
  Result := Exec('[UNHIDE()]');
end;

function TExcelDDEClient.Move(deltaRow, deltaCol: LongWord): Boolean;
begin
	Result := Exec(Format('[SELECT("%s[%d]%s[%d]")]', [RowChar, ColChar, deltaRow, deltaCol]));
end;

function TExcelDDEClient.RunMacro(const sMacroFileName,
  MacroName: String): Boolean;
begin
  Result := Exec(Format('[RUN("%s!%s";FALSE)]', [sMacroFileName, MacroName]));
end;

function TExcelDDEClient.NewSheet: Boolean;
begin
  Result := Exec('[WORKBOOK.INSERT()]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.EchoOff: Boolean;
begin
  Result := Exec('[ECHO(FALSE)]');
end;

function TExcelDDEClient.EchoOn: Boolean;
begin
  Result := Exec('[ECHO(TRUE)]');
end;

function TExcelDDEClient.DisableInput: Boolean;
begin
  Result := Exec('[DISABLE.INPUT(TRUE)]');
end;

function TExcelDDEClient.EnableInput: Boolean;
begin
  Result := Exec('[DISABLE.INPUT(FALSE)]');
end;

function TExcelDDEClient.RenameSheet(const OldName, NewName: string): Boolean;
begin
  { Excel 8 bug: this DDE command will NOT work }
  Result := Exec( Format('[WORKBOOK.NAME("%s","%s")]',
              [StripQuotation(OldName), StripQuotation(NewName)] ));
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.Patterns(const apattern: TExcelPattern; const afore, aback: TExcelColor; const newui: Boolean): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..4] of String;
	i, j: LongWord;
begin
	ParamList[1] := Trim(IntToStr(apattern));
	ParamList[2] := Trim(IntToStr(afore));
	ParamList[3] := Trim(IntToStr(aback));
	ParamList[4] := iif(newui, 'TRUE', 'FALSE');
	j := 4;
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
	Result := Exec('[PATTERNS('+ParamArg+')]');
end;

function TExcelDDEClient.Alignment(horiz_align: TExcelHorizontalAlignment;
  wrap: TExcelWrap; vert_align: TExcelVerticalAlignment;
  orientation: TExcelOrientation): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..4] of String;
	i, j: LongWord;
  l: Boolean;
begin
  ParamArg := '';
  l := True;
  j := 4;
	ParamList[4] := iif(orientation = eoNotChanged, '', Trim(IntToStr(Ord(orientation))));
  if (ParamList[4] = '') and l then Dec(j) else l := false;
	ParamList[3] := iif(vert_align = vaNotChanged, '', Trim(IntToStr(Ord(vert_align))));
  if (ParamList[3] = '') and l then Dec(j) else l := false;
	ParamList[2] := iif(wrap = wNotChanged, '', iif(wrap = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[2] = '') and l then Dec(j) else l := false;
	ParamList[1] := iif(horiz_align = haNotChanged, '', Trim(IntToStr(Ord(horiz_align))));
  if (ParamList[1] = '') and l then Dec(j);
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
	Result := Exec('[ALIGNMENT('+ParamArg+')]');
end;

function TExcelDDEClient.GetCellWidth: Integer; {sorry! only integer}
var
  LastCell: TExcelCell;
  OldSelection: String;
begin
	OldSelection := SelectionAddress;
  GetLastCell(LastCell);
  //Exec(Format('[FORMULA("=GET.CELL(17, ""R2"")", "%s%d%s%d")]', [RowChar, LastCell.Row+1, ColChar, LastCell.Col+1]));
  Exec(Format('[FORMULA("=CELL(""width"")", "%s%d%s%d")]', [RowChar, LastCell.Row+1, ColChar, LastCell.Col+1]));
  try
    Result := StrToInt(GetData(LastCell.Row+1, LastCell.Col+1));
  except
    Result := -1;
  end;
  select(LastCell.Row+1, LastCell.Col+1);
  Exec('[CLEAR(1)]');
  select(OldSelection);
end;

function TExcelDDEClient.RunMacro(const MacroName: String): Boolean;
begin
  Result := Exec(Format('[RUN("%s";FALSE)]', [MacroName]));
end;

function TExcelDDEClient.GetSheetsOnCurrentBook: TStringList;
var
  i: Integer;
  ss, ss1: String;
begin
  ss := CurrentWorkBook;
  GetStrings(FSheetsOnCurrentBook, 'Topics');
  i := 0;
  while i <= FSheetsOnCurrentBook.Count - 1 do
  begin
    if Pos(ss, FSheetsOnCurrentBook.Strings[i]) = 0 then
      FSheetsOnCurrentBook.Delete(i)
    else
      begin
        ss1 := FSheetsOnCurrentBook.Strings[i];
        FSheetsOnCurrentBook.Strings[i] := Copy(ss1, Length(ss) + 3, Length(ss1));
        Inc(i);
      end;
  end;
  Result := FSheetsOnCurrentBook;
end;

function TExcelDDEClient.GetSheetNumber: Integer;
var
  i: Integer;
  ss: String;
begin
  ss := CurrentSheet;
  GetSheetsOnCurrentBook;
  i := 0;
  Result := 0;
  while i <= FSheetsOnCurrentBook.Count - 1 do
  begin
    if FSheetsOnCurrentBook.Strings[i] = ss then
    begin
      Result := i;
      Exit;
    end;
    Inc(i);
  end;
end;

procedure TExcelDDEClient.SetSheetNumber(const Value: Integer);
begin
  GetSheetsOnCurrentBook;
  CurrentSheet := FSheetsOnCurrentBook.Strings[Value];
end;

function TExcelDDEClient.GetSheetsCount: Integer;
begin
  GetSheetsOnCurrentBook;
  Result := FSheetsOnCurrentBook.Count;
end;

function TExcelDDEClient.Border(outline, left, right, top,
  bottom: TExcelLineStyle; shade: TExcelWrap; outline_color, left_color,
  right_color, top_color, bottom_color: TExcelColor): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..11] of String;
	i, j: LongWord;
  l: Boolean;
begin
  ParamArg := '';
  l := True;
  j := 11;
	ParamList[11] := iif(bottom_color = 0, '', Trim(IntToStr(bottom_color)));
  if (ParamList[11] = '') and l then Dec(j) else l := false;
	ParamList[10] := iif(top_color = 0, '', Trim(IntToStr(top_color)));
  if (ParamList[10] = '') and l then Dec(j) else l := false;
	ParamList[9] := iif(right_color = 0, '', Trim(IntToStr(right_color)));
  if (ParamList[9] = '') and l then Dec(j) else l := false;
	ParamList[8] := iif(left_color = 0, '', Trim(IntToStr(left_color)));
  if (ParamList[8] = '') and l then Dec(j) else l := false;
	ParamList[7] := iif(outline_color = 0, '', Trim(IntToStr(outline_color)));
  if (ParamList[7] = '') and l then Dec(j) else l := false;
	ParamList[6] := iif(shade = wNotChanged, '', iif(shade = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[6] = '') and l then Dec(j) else l := false;
	ParamList[5] := iif(bottom = lsNotChanged, '', Trim(IntToStr(Ord(bottom))));
  if (ParamList[5] = '') and l then Dec(j) else l := false;
	ParamList[4] := iif(top = lsNotChanged, '', Trim(IntToStr(Ord(top))));
  if (ParamList[4] = '') and l then Dec(j) else l := false;
	ParamList[3] := iif(right = lsNotChanged, '', Trim(IntToStr(Ord(right))));
  if (ParamList[3] = '') and l then Dec(j) else l := false;
	ParamList[2] := iif(left = lsNotChanged, '', Trim(IntToStr(Ord(left))));
  if (ParamList[2] = '') and l then Dec(j) else l := false;
	ParamList[1] := iif(outline = lsNotChanged, '', Trim(IntToStr(Ord(outline))));
  if (ParamList[1] = '') and l then Dec(j);
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
	Result := Exec('[BORDER('+ParamArg+')]');
end;

function TExcelDDEClient.FontProperties(font, font_style: String;
  size: Integer; strikethrough, superscript, subscript, outline,
  shadow: TExcelWrap; underline: TExcelUnderLine; color: TExcelColor;
  normal: TExcelWrap; background: TExcelBackGround; start_char,
  char_count: Integer): Boolean;
var
	ParamArg: String;
	ParamList: Array [1..14] of String;
	i, j: LongWord;
  l: Boolean;
begin
  ParamArg := '';
  l := True;
  j := 14;
	ParamList[14] := iif(char_count = -1, '', Trim(IntToStr(char_count)));
  if (ParamList[14] = '') and l then Dec(j) else l := false;
	ParamList[13] := iif(start_char = -1, '', Trim(IntToStr(start_char)));
  if (ParamList[13] = '') and l then Dec(j) else l := false;
	ParamList[12] := iif(background = bgNotChanged, '', Trim(IntToStr(Ord(background))));
  if (ParamList[12] = '') and l then Dec(j) else l := false;
	ParamList[11] := iif(normal = wNotChanged, '', iif(normal = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[11] = '') and l then Dec(j) else l := false;
	ParamList[10] := iif(color = 0, '', Trim(IntToStr(color)));
  if (ParamList[10] = '') and l then Dec(j) else l := false;
	ParamList[9] := iif(underline = ulNotChanged, '', Trim(IntToStr(Ord(underline))));
  if (ParamList[9] = '') and l then Dec(j) else l := false;
	ParamList[8] := iif(shadow = wNotChanged, '', iif(shadow = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[8] = '') and l then Dec(j) else l := false;
	ParamList[7] := iif(outline = wNotChanged, '', iif(outline = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[7] = '') and l then Dec(j) else l := false;
	ParamList[6] := iif(subscript = wNotChanged, '', iif(subscript = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[6] = '') and l then Dec(j) else l := false;
	ParamList[5] := iif(superscript = wNotChanged, '', iif(superscript = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[5] = '') and l then Dec(j) else l := false;
	ParamList[4] := iif(strikethrough = wNotChanged, '', iif(strikethrough = wTrue, 'TRUE', 'FALSE'));
  if (ParamList[4] = '') and l then Dec(j) else l := false;
	ParamList[3] := iif(size = -1, '', Trim(IntToStr(size)));
  if (ParamList[3] = '') and l then Dec(j) else l := false;
	ParamList[2] := font_style;
  if (ParamList[2] = '') and l then Dec(j) else
  begin
    ParamList[2] := '"'+ParamList[2]+'"';
    l := false;
  end;
	ParamList[1] := font;
  if (ParamList[1] = '') and l then Dec(j) else ParamList[1] := '"'+ParamList[1]+'"';
	for i := 1 to j do
  begin
		if i = 1 then
			ParamArg := ParamList[1]
		else
			ParamArg := ParamArg + ', ' + ParamList[i];
	end;
	Result := Exec('[FONT.PROPERTIES('+ParamArg+')]');
end;

function TExcelDDEClient.NewSheet(const Tamplate: String): Boolean;
begin
  Result := Exec('[WORKBOOK.INSERT("'+UpperCase(Trim(Tamplate))+'")]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.NewSheet(const Type_Num: TWorkBookInsertType): Boolean;
begin
  Result := Exec('[WORKBOOK.INSERT('+Trim(IntToStr(Ord(Type_Num)))+')]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.CopySheet(SourseSheet, DestBook: String; const Position: Integer): Boolean;
var
  ArgList: String;
begin
  if SourseSheet = '' then SourseSheet := CurrentSheet;
  if DestBook = '' then ArgList := '"'+SourseSheet+'"'
  else
    if Position = -1 then
      ArgList := '"'+SourseSheet + '", "' + DestBook + '"'
    else
      ArgList := '"'+SourseSheet + '", "' + DestBook + '", ' + Trim(IntToStr(Position));
  Result := Exec('[WORKBOOK.COPY('+ArgList+')]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.MoveSheet(SourseSheet, DestBook: String; const Position: Integer): Boolean;
var
  ArgList: String;
begin
  if SourseSheet = '' then SourseSheet := CurrentSheet;
  if DestBook = '' then ArgList := '"'+SourseSheet+'"'
  else
    if Position = -1 then
      ArgList := '"'+SourseSheet + '", "' + DestBook + '"'
    else
      ArgList := '"'+SourseSheet + '", "' + DestBook + '", ' + Trim(IntToStr(Position));
  Result := Exec('[WORKBOOK.MOVE('+ArgList+')]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.DeleteSheet(SourseSheet: String): Boolean;
var
  ArgList: String;
begin
  if SourseSheet = '' then ArgList := CurrentSheet else ArgList := SourseSheet;
  Result := Exec('[WORKBOOK.DELETE("'+ArgList+'")]');
  SwitchTopicToCurrentSheet;
end;

function TExcelDDEClient.ClearSheet(SourseSheet: String; pProc: TClearSheetEvent): Boolean;
var
  ArgList: String;
  TopCell, BottomCell: TExcelCell;
begin
  if SourseSheet = '' then ArgList := CurrentSheet else ArgList := SourseSheet;
  GetLastCell(BottomCell);
  TopCell.Col := 1;
  TopCell.Row := 1;
  Select(TopCell, BottomCell);
  Result := Exec('[CLEAR(1)]');
  if Result then
    if Exec('[SELECT.SPECIAL(13)]') then
      Result := Exec('[CLEAR(1)]');
  if Result then
    if Assigned(pProc) then pProc(TopCell, BottomCell);
end;

function TExcelDDEClient.SetColWidthToStandard(var TopCell,
  BottomCell: TExcelCell): Boolean;
begin
  Result := SetColWidth(10,  Format('%s%d:%s%d', [ColChar, TopCell.Col, ColChar, BottomCell.Col]), True);
end;

function TExcelDDEClient.SetRowHeightToStandard(var TopCell,
  BottomCell: TExcelCell): Boolean;
begin
  Result := SetRowHeight(10,  Format('%s%d:%s%d', [RowChar, TopCell.Row, RowChar, BottomCell.Row]), True);
end;

function TExcelDDEClient.IsLoadedServiceApp: Boolean;
begin
  Result := ( CountExcelInstance > 0 );
end;

class function TExcelDDEClient.CountExcelInstance: LongInt;
var
  ExcelCounter: Longint;
begin
  ExcelCounter := 0;
  EnumWindows(@EnumInstanceWindows, LPARAM(@ExcelCounter));
  Result := ExcelCounter;
end;

function TExcelDDEClient.PutData(const sData: Variant): Boolean;
begin
  Result := Poke(SelectionAddress, VarToStr(sData));
end;

function TExcelDDEClient.PutData(const Row, Col: LongWord;
  const sData: Variant): Boolean;
begin
  Result := Poke(Format(RowChar+'%d'+ColChar+'%d', [Row, Col]), VarToStr(sData));
end;

end.
