unit DDEReports;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, DDEML, StdCtrls,
  DDEClient, SEditProp, Dialogs, SSmrtRep, Dsgnintf, CommonType;

const
  WM_REP_NOT_EXCEL_CONNECTED =  WM_USER + 500;
  WM_REP_OPEN_FILE_NOT_EXISTS = WM_USER + 501;

type

  TExcelCellContent = record
    ContentColRow: Variant;
    ContentLen: Integer;
    SectionRow: Integer;
    SectionCol: Integer;
  end;
  pTExcelCellContent = ^TExcelCellContent;

  pInteger = ^Integer;

  TExcelDdeReportSectionInfo = record
	  SectionName: ShortString;
    Contents: TList;
	  SectionNum: Integer;
	  SectionUsed: Boolean;
	  RowsCount: Integer;
    ColsCount: Integer;
    Widths: pInteger;
    iRange: iUnknown;
  end;
  pTExcelDdeReportSectionInfo = ^TExcelDdeReportSectionInfo;

  TOnReportError = procedure(Sender: TObject; mError, nError: LongWord; sError: String) of object;
  TOnCheck = procedure(Sender: TObject; var bCheck: Boolean) of object;
  TOnSectionCheck = procedure(Sender: TObject; SectionName: String; SectionNum: LongWord; var bCheck: Boolean) of object;
  TOnBeginEndSectionEvent = procedure(Sender: TObject; SectionName: String; SectionNum: LongWord) of object;
  TOnDataRequest = procedure(Sender: TObject; SectionName: String; SectionNum: LongWord; var sData: Variant; var bCheck: Boolean; OutR, OutC, InR, InC: LongWord) of object;
  TOnSectionDataRequestByNum = procedure(Sender: TObject; sSectionName: String; var iP: Longint; var sStr: String) of object;
  TOnSectionDataRequestByName = procedure(Sender: TObject; sSectionName: String; var sFieldName: String; var sStr: String) of object;

{TReportAbstract}

  TReportAbstract = class(TComponent)
    private

      FOnStartReport: TNotifyEvent;
      FOnEndReport: TNotifyEvent;
      FOnReportError: TOnReportError;
      FOnMoveTop: TOnCheck;
      FOnMoveNext: TOnCheck;
      FOnSectionBegin: TOnBeginEndSectionEvent;
      FOnSectionEnd: TOnBeginEndSectionEvent;
      FOnSectionComplit: TOnBeginEndSectionEvent;
      FOnSectionCheck: TOnSectionCheck;
      FOnDataRequest: TOnDataRequest;
      FOnRequestByName: TOnSectionDataRequestByName;
      FOnRequestByNum: TOnSectionDataRequestByNum;
      FOnSectionPrepare: TNotifyEvent;

    protected

      function GetCurrentSection: String; virtual; abstract;
      function GetCurrentSectionNum: Integer; virtual; abstract;
      function GetTypeDiraction: TTtypeDiraction; virtual; abstract;

    public

      procedure ReportResume; virtual; abstract;
      procedure ReportExecute; virtual; abstract;
      procedure ReportHorizontalyResume; virtual; abstract;
      procedure ReportHorizontalyExecute; virtual; abstract;
      procedure BiDirectionResume; virtual; abstract;
      procedure BiDirectionExecute; virtual; abstract;
      function PrepareSection(const SectionName: ShortString): Integer; virtual; abstract;
      function PrepareHorizontalySection(const SectionName: ShortString): Integer; virtual; abstract;

      procedure OutSection(SectionName: ShortString); virtual; abstract;
      procedure OutHorizontalySection(SectionName: ShortString); virtual; abstract;
      procedure OutSectionByNum(SectionNum: Integer); virtual; abstract;
      procedure OutHorizontalySectionByNum(SectionNum: Integer); virtual; abstract;

      property CurrentSection: String read GetCurrentSection;
      property CurrentSectionNum: Integer read GetCurrentSectionNum;

      property TypeDiraction: TTtypeDiraction read GetTypeDiraction;

      property OnStartReport: TNotifyEvent read FOnStartReport write FOnStartReport;
      property OnEndReport: TNotifyEvent read FOnEndReport write FOnEndReport;
      property OnReportError: TOnReportError read FOnReportError write FOnReportError;
      property OnMoveTop: TOnCheck read FOnMoveTop write FOnMoveTop;
      property OnMoveNext: TOnCheck read FOnMoveNext write FOnMoveNext;
      property OnSectionBegin: TOnBeginEndSectionEvent read FOnSectionBegin write FOnSectionBegin;
      property OnSectionEnd: TOnBeginEndSectionEvent read FOnSectionEnd write FOnSectionEnd;
      property OnSectionComplit: TOnBeginEndSectionEvent read FOnSectionComplit write FOnSectionComplit;
      property OnSectionCheck: TOnSectionCheck read FOnSectionCheck write FOnSectionCheck;
      property OnDataRequest: TOnDataRequest read FOnDataRequest write FOnDataRequest;
      property OnRequestByName: TOnSectionDataRequestByName read FOnRequestByName write FOnRequestByName;
      property OnRequestByNum: TOnSectionDataRequestByNum read FOnRequestByNum write FOnRequestByNum;
      property OnSectionPrepare: TNotifyEvent read FOnSectionPrepare write FOnSectionPrepare;

  end;

{TExcelDDEReport}

  TExcelDDEReport = class(TReportAbstract)
    private

      FExcel: TExcelDDEClient;
      FMaket: String;
      FOutBook: String;
      FCurrRow: Integer;
      FCurrCol: Integer;
      FSectionsV: array [1..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;
      FSectionsH: array [1..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;
      FCurrentSection: String;
      FCurrentSectionNum: Integer;
      FCurrentWorkBook: String;
      FTmpFile: String;
      FInputSheet: String;
      FOutputSheet: String;
      FCharByNum: Char;
      FCharByName: Char;
      FTypeDiraction: TTtypeDiraction;
      FLastDynSection:    Integer;
      FLastDynSectionH:    Integer;
      lWM_REP_SECTION_PREPARE: boolean;

      procedure StartReportV;
      procedure PrepareSectionV(SectionName: String; SectionNum: LongWord);
      procedure OutSectionV(SectionName: String; bOut: Boolean; SectionNum: LongWord);
      procedure CloseSectionsV;

      procedure StartReportH;
      procedure PrepareSectionH(SectionName: String; SectionNum: LongWord);
      procedure OutSectionH(SectionName: String; bOut: Boolean; SectionNum: LongWord);
      procedure CloseSectionsH;

      procedure StartBiDirectionReport;
      procedure SetDdeMgrClient(const Value: TDdeMgrClient);
      function GetDdeMgrClient: TDdeMgrClient;
      function GetInputSheet: String;

    protected

      function GetCurrentSection: String; override;
      function GetCurrentSectionNum: Integer; override;
      function GetTypeDiraction: TTtypeDiraction; override;

      procedure SetStandardRowHeight(var t, b: TExcelCell);
      procedure SetStandardColWidth(var t, b: TExcelCell);

    public

      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      procedure ReportResume; override;
      procedure ReportExecute; override;
      procedure ReportHorizontalyResume; override;
      procedure ReportHorizontalyExecute; override;
      procedure BiDirectionResume; override;
      procedure BiDirectionExecute; override;
      function PrepareSection(const SectionName: ShortString): Integer; override;
      function PrepareHorizontalySection(const SectionName: ShortString): Integer; override;

      procedure OutSection(SectionName: ShortString); override;
      procedure OutHorizontalySection(SectionName: ShortString); override;
      procedure OutSectionByNum(SectionNum: Integer); override;
      procedure OutHorizontalySectionByNum(SectionNum: Integer); override;

      property Excel: TExcelDDEClient read FExcel;

    published

      property DdeMgrClient: TDdeMgrClient read GetDdeMgrClient write SetDdeMgrClient;
      property InputBlank: String read FMaket write FMaket;
      property InputSheet: String read GetInputSheet write FInputSheet;
      property OutputBlank: String read FOutBook write FOutBook;
      property OutputSheet: String read FOutputSheet write FOutputSheet;
      property CharByNum: Char read FCharByNum write FCharByNum default #35; {#0001}
      property CharByName: Char read FCharByName write FCharByName default #126; {~name~}

      property OnStartReport;
      property OnEndReport;
      property OnReportError;
      property OnMoveTop;
      property OnMoveNext;
      property OnSectionBegin;
      property OnSectionEnd;
      property OnSectionComplit;
      property OnSectionCheck;
      property OnDataRequest;
      property OnRequestByName;
      property OnRequestByNum;
      property OnSectionPrepare;

  end;

procedure Register;
function GetTmpFile(): String;
function GetFileName(sFile: String): String;
function GetContent(s: String): Variant;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TExcelDDEReport]);
  RegisterPropertyEditor(TypeInfo(String), TExcelDDEReport, 'InputBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TExcelDDEReport, 'OutputBlank', TEdSaveBlank);
end;

function GetContent(s: String): Variant;
var
  i: Integer;
  f: double;
  d: TDateTime;
begin
  try
    i := StrToInt(s);
    Result := i;
    Exit;
  except
  end;
  try
    f := StrToFloat(s);
    Result := f;
    Exit;
  except
  end;
  try
    d := StrToDateTime(s);
    Result := d;
    Exit;
  except
  end;
  Result := s;
end;

function GetTmpFile(): String;
var
  cpBufferPath: pChar;
  cpBufferTmpFile: pChar;
  i: Integer;
begin

  cpBufferPath := StrAlloc(MAX_PATH + 1);
  FillChar(cpBufferPath^, MAX_PATH + 1, 0);

  cpBufferTmpFile := StrAlloc(MAX_PATH + 1);
  FillChar(cpBufferTmpFile^, MAX_PATH + 1, 0);

  GetTempPath(MAX_PATH, cpBufferPath);

  for i := 0 to MAX_PATH - 1 do
  begin
    if Ord(cpBufferPath[i]) = 0 then break;
    if Ord(cpBufferPath[i]) <= 32 then
    begin
      cpBufferPath[i] := '\';
      cpBufferPath[i + 1] := #0;
      break;
    end;
  end;

  GetTempFileName(cpBufferPath, PChar(''), 0, cpBufferTmpFile);

  Result := Copy(cpBufferTmpFile, 1, Length(cpBufferTmpFile));

  StrDispose(cpBufferPath);
  StrDispose(cpBufferTmpFile);

end;

function GetFileName(sFile: String): String;
var
  i: LongWord;
begin
  Result := sFile;
  i := RScan(sFile, '\');
  if i <> 0 then Result := Copy(sFile, i+1, Length(sFile));
end;

{ TExcelDDEReport }

procedure TExcelDDEReport.CloseSectionsV;
var
  i, j, k: Integer;
  pp: pTExcelCellContent;
begin
	for i := 1 to MAX_SECTION_NUM do
	  if FSectionsV[i].SectionUsed then
    begin
		  FSectionsV[i].SectionUsed := FALSE;
      k := FSectionsV[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTExcelCellContent(FSectionsV[i].Contents.Items[j]);
        pp^.ContentColRow := Unassigned;
        StrDispose(PChar(pp));
      end;
      FSectionsV[i].Contents.Destroy;
    end;
end;

constructor TExcelDDEReport.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FExcel := TExcelDDEClient.Create(self);
  FExcel.RequestDataSize := 4096;
  FCharByNum := #35;
  FCharByName := #126;
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  lWM_REP_SECTION_PREPARE := false;
end;

destructor TExcelDDEReport.Destroy;
begin
  FExcel.Destroy;
  CloseSectionsH;
  CloseSectionsV;
  inherited Destroy;
end;

procedure TExcelDDEReport.OutSectionV(SectionName: String; bOut: Boolean; SectionNum: LongWord);
var
  dwi: Integer;
  pp: pTExcelCellContent;
  ContentColRow: Variant;
  ContentColRowStr: String;
  bCheck: Boolean;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
  H: Boolean;
begin
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
	if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
	if FSectionsV[SectionNum].SectionUsed then
  begin
    if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
		if bOut then
    begin
      Excel.Exec(Format('[COPY("''[%s]%s''!%s", "''[%s]%s''!%s%d%s1")]', [FCurrentWorkBook, InputSheet, SectionName, FCurrentWorkBook, OutputSheet, Excel.RowChar, FCurrRow, Excel.ColChar]));
			dwi := 0;
			while dwi < FSectionsV[SectionNum].Contents.Count do
      begin
        pp := pTExcelCellContent(FSectionsV[SectionNum].Contents.Items[dwi]);
        ContentColRow := pp^.ContentColRow;
        bCheck := false;
        if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, ContentColRow, bCheck, FCurrRow + pp^.SectionRow - 1, pp^.SectionCol, pp^.SectionRow, pp^.SectionCol);
        if bCheck then
          Excel.PutData(FCurrRow + pp^.SectionRow - 1, pp^.SectionCol, ContentColRow)
        else begin
          //
          if Assigned(OnRequestByNum) then
          begin
            H := false;
            pSt := 0;
            ContentColRowStr := VarToStr(ContentColRow);
            while true do
            begin
              pInt := LScan(ContentColRowStr, FCharByNum, pSt);
              if pInt = 0 then Break else H := true;
              pSt := pInt;
              if  ( ContentColRowStr[pInt + 1] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 2] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 3] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 4] in ['0'..'9'] ) then
              begin
                iP := StrToInt(Copy(ContentColRowStr, pInt + 1, 4));
                sStr := '';
                OnRequestByNum(self, SectionName, iP, sStr);
                ContentColRowStr := Copy(ContentColRowStr, 1, pInt - 1) + sStr + Copy(ContentColRowStr, pInt + 5, Length(ContentColRowStr));
              end;
            end;
            if H then Excel.PutData(FCurrRow + pp^.SectionRow - 1, pp^.SectionCol, ContentColRowStr);
          end;
          if Assigned(OnRequestByName) then
          begin
            H := false;
            pSt := 0;
            while true do
            begin
              pInt := LScan(ContentColRow, FCharByName, pSt);
              if pInt = 0 then Break;
              pInt1 := LScan(ContentColRow, FCharByName, pInt);
              if pInt1 = 0 then Break else H := true;
              pSt := pInt1;
              sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
              sStr := '';
              OnRequestByName(self, SectionName, sField, sStr);
              ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
            end;
            if H then Excel.PutData(FCurrRow + pp^.SectionRow - 1, pp^.SectionCol, ContentColRow);
          end;
          //
        end;
				Inc(dwi);
			end;
			FCurrRow := FCurrRow + FSectionsV[SectionNum].RowsCount;
      if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
		end;
	end;
  if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
end;

procedure TExcelDDEReport.PrepareSectionV(SectionName: String; SectionNum: LongWord);
var
	TopCell: TExcelCell;
	BottomCell: TExcelCell;
  SectionRow, SectionCol: LongWord;
  ContentColRow: String;
  CurrentContetns: String;
  CurrentChar: Char;
  i, j: Integer;
  pp: pTExcelCellContent;
begin
	FSectionsV[SectionNum].SectionName := SectionName;
	FSectionsV[SectionNum].SectionNum := SectionNum;
	FSectionsV[SectionNum].SectionUsed := FALSE;
  Excel.Select(100, 100);
	Excel.Select(SectionName);
	Excel.GetRectSelection(TopCell, BottomCell);
	if not (( TopCell.Row = 100 ) and ( BottomCell.Row = 100 ) and ( TopCell.Col  = 100 ) and ( BottomCell.Col = 100 )) then
  begin
		FSectionsV[SectionNum].SectionUsed := TRUE;
		FSectionsV[SectionNum].RowsCount := BottomCell.Row - TopCell.Row + 1;
    FSectionsV[SectionNum].Contents := TList.Create;
    //
      SectionRow := 1;
      SectionCol := 1;
      ContentColRow := '';
      CurrentContetns := Excel.GetData(SectionName);
      j := Length(CurrentContetns);
      for i := 1 to j do
      begin
        CurrentChar := CurrentContetns[i];
        if  CurrentChar = #9 then
        begin
          //
          if Length(ContentColRow) > 0 then
          begin
            pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
            pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
            pp^.ContentLen := Length(ContentColRow);
            //StrPCopy(pp^.ContentColRow, ContentColRow);
            pp^.SectionRow := SectionRow;
            pp^.SectionCol := SectionCol;
            FSectionsV[SectionNum].Contents.Add(pp);
          end;
          //
          Inc(SectionCol);
          ContentColRow := '';
        end;
        if  CurrentChar = #$D then
        begin
          //
          if Length(ContentColRow) > 0 then
          begin
            pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
            pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
            pp^.ContentLen := Length(ContentColRow);
            //StrPCopy(pp^.ContentColRow, ContentColRow);
            pp^.SectionRow := SectionRow;
            pp^.SectionCol := SectionCol;
            FSectionsV[SectionNum].Contents.Add(pp);
          end;
          //
          SectionCol := 1;
          Inc(SectionRow);
          ContentColRow := '';
        end;
        if  CurrentChar = #$A then ContentColRow := '';
        if not (CurrentChar in [#9, #$A, #$D]) then ContentColRow := ContentColRow + CurrentChar;
      end;
      if ContentColRow <> '' then
      begin
        //
        if Length(ContentColRow) > 0 then
        begin
          pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
          pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
          pp^.ContentLen := Length(ContentColRow);
          //StrPCopy(pp^.ContentColRow, ContentColRow);
          pp^.SectionRow := SectionRow;
          pp^.SectionCol := SectionCol;
          FSectionsV[SectionNum].Contents.Add(pp);
        end;
        //
      end;
    //
	end;
	Excel.Select('R1C1');
end;

procedure TExcelDDEReport.ReportExecute;
begin
  StartReportV;
end;

procedure TExcelDDEReport.ReportResume;
begin
  StartReportV;
end;

procedure TExcelDDEReport.StartReportV;
var
  OldIgnoreErrors: Boolean;
  bCheck: Boolean;
begin
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  FTypeDiraction := tdVerticaly;
  if Assigned(OnStartReport) then OnStartReport(self);
  FExcel.DDEService := 'Excel';
  FExcel.DDETopic := 'System';
  FExcel.Active := true;
  if Excel.Active then
  begin
    OldIgnoreErrors := Excel.IgnoreErrors;
    Excel.IgnoreErrors := true;
    if FileExists(InputBlank) then
    begin
      FTmpFile := GetTmpFile;
      if CopyFile(pChar(InputBlank), pChar(FTmpFile), False) then
      begin
        Excel.OpenWorkBook(FTmpFile);
        FCurrentWorkBook := Excel.CurrentWorkBook;
        if UpperCase(Trim(FCurrentWorkBook)) = UpperCase(Trim(GetFileName(FTmpFile))) then
        begin
          try
            FCurrRow := 1;
            Excel.CurrentSheet := InputSheet;
            Excel.CopySheet;
            Excel.ClearSheet('', SetStandardRowHeight);
            Excel.Select(1, 1);
            OutputSheet := Excel.CurrentSheet;
            Excel.CurrentSheet := InputSheet;
            if not Excel.Active then Exit;
            PrepareSectionV('REPORTHEADER', 1);
            PrepareSectionV('GROUPHEADER', 2);
            PrepareSectionV('SUBGROUPHEADER', 3);
            PrepareSectionV('SUBSUBGROUPHEADER', 4);
            PrepareSectionV('SUPERSUBGROUPHEADER', 5);
            PrepareSectionV('BODY', 6);
            PrepareSectionV('SUPERSUBGROUPFOOTER', 7);
            PrepareSectionV('SUBSUBGROUPFOOTER', 8);
            PrepareSectionV('SUBGROUPFOOTER', 9);
            PrepareSectionV('GROUPFOOTER', 10);
            PrepareSectionV('REPORTFOOTER', 11);
            if not Excel.Active then Exit;
            lWM_REP_SECTION_PREPARE := true;
            if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
            lWM_REP_SECTION_PREPARE := false;
            Excel.CurrentSheet := OutputSheet;
            OutSectionV('REPORTHEADER', TRUE, 1);
            bCheck := True;
            if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
            while not bCheck do
            begin
              OutSectionV('GROUPHEADER', FALSE, 2);
              OutSectionV('SUBGROUPHEADER', FALSE, 3);
              OutSectionV('SUBSUBGROUPHEADER', FALSE, 4);
              OutSectionV('SUPERSUBGROUPHEADER', FALSE, 5);
              OutSectionV('BODY', TRUE, 6);
              OutSectionV('SUPERSUBGROUPFOOTER', FALSE, 7);
              OutSectionV('SUBSUBGROUPFOOTER', FALSE, 8);
              OutSectionV('SUBGROUPFOOTER', FALSE, 9);
              OutSectionV('GROUPFOOTER', FALSE, 10);
              if not Excel.Active then Exit;
              bCheck := True;
              if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
            end;
            OutSectionV('REPORTFOOTER', TRUE, 11);
          finally
            if Excel.Active then
            begin
              Excel.CurrentSheet := InputSheet;
              Excel.DeleteSheet;
              Excel.CurrentSheet := OutputSheet;
              Excel.SaveAs(OutputBlank);
              Excel.select(1, 1);
            end;
            CloseSectionsV;
            Excel.IgnoreErrors := OldIgnoreErrors;
          end;
        end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_FILE_NOT_EXISTS, 0, 'Open file not found');
  end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
  if Assigned(OnEndReport) then OnEndReport(self);
  FExcel.Active := false;
  DeleteFile(FTmpFile);
  FTypeDiraction := tdNotDefined;
end;

procedure TExcelDDEReport.SetStandardRowHeight(var t, b: TExcelCell);
begin
  Excel.SetRowHeightToStandard(t, b);
end;

procedure TExcelDDEReport.ReportHorizontalyExecute;
begin
  StartReportH;
end;

procedure TExcelDDEReport.ReportHorizontalyResume;
begin
  StartReportH;
end;

procedure TExcelDDEReport.StartReportH;
var
  OldIgnoreErrors: Boolean;
  bCheck: Boolean;
begin
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  FTypeDiraction := tdHorizontaly;
  if Assigned(OnStartReport) then OnStartReport(self);
  FExcel.DDEService := 'Excel';
  FExcel.DDETopic := 'System';
  FExcel.Active := true;
  if Excel.Active then
  begin
    OldIgnoreErrors := Excel.IgnoreErrors;
    Excel.IgnoreErrors := true;
    if FileExists(InputBlank) then
    begin
      FTmpFile := GetTmpFile;
      if CopyFile(pChar(InputBlank), pChar(FTmpFile), False) then
      begin
        Excel.OpenWorkBook(FTmpFile);
        FCurrentWorkBook := Excel.CurrentWorkBook;
        if UpperCase(Trim(FCurrentWorkBook)) = UpperCase(Trim(GetFileName(FTmpFile))) then
        begin
          try
            FCurrCol := 1;
            Excel.CurrentSheet := InputSheet;
            Excel.CopySheet;
            Excel.ClearSheet('', SetStandardColWidth);
            Excel.Select(1, 1);
            OutputSheet := Excel.CurrentSheet;
            Excel.CurrentSheet := InputSheet;
            if not Excel.Active then Exit;
            PrepareSectionH('H_REPORTHEADER', 1);
            PrepareSectionH('H_GROUPHEADER', 2);
            PrepareSectionH('H_SUBGROUPHEADER', 3);
            PrepareSectionH('H_SUBSUBGROUPHEADER', 4);
            PrepareSectionH('H_SUPERSUBGROUPHEADER', 5);
            PrepareSectionH('H_BODY', 6);
            PrepareSectionH('H_SUPERSUBGROUPFOOTER', 7);
            PrepareSectionH('H_SUBSUBGROUPFOOTER', 8);
            PrepareSectionH('H_SUBGROUPFOOTER', 9);
            PrepareSectionH('H_GROUPFOOTER', 10);
            PrepareSectionH('H_REPORTFOOTER', 11);
            Excel.CurrentSheet := OutputSheet;
            if not Excel.Active then Exit;
            lWM_REP_SECTION_PREPARE := true;
            if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
            lWM_REP_SECTION_PREPARE := false;
            OutSectionH('H_REPORTHEADER', TRUE, 1);
            bCheck := True;
            if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
            while not bCheck do
            begin
              OutSectionH('H_GROUPHEADER', FALSE, 2);
              OutSectionH('H_SUBGROUPHEADER', FALSE, 3);
              OutSectionH('H_SUBSUBGROUPHEADER', FALSE, 4);
              OutSectionH('H_SUPERSUBGROUPHEADER', FALSE, 5);
              OutSectionH('H_BODY', TRUE, 6);
              OutSectionH('H_SUPERSUBGROUPFOOTER', FALSE, 7);
              OutSectionH('H_SUBSUBGROUPFOOTER', FALSE, 8);
              OutSectionH('H_SUBGROUPFOOTER', FALSE, 9);
              OutSectionH('H_GROUPFOOTER', FALSE, 10);
              if not Excel.Active then Exit;
              bCheck := True;
              if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
            end;
            OutSectionH('H_REPORTFOOTER', TRUE, 11);
          finally
            if Excel.Active then
            begin
              Excel.CurrentSheet := InputSheet;
              Excel.DeleteSheet;
              Excel.CurrentSheet := OutputSheet;
              Excel.SaveAs(OutputBlank);
              Excel.select(1, 1);
            end;
            CloseSectionsH;
            Excel.IgnoreErrors := OldIgnoreErrors;
          end;
        end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_FILE_NOT_EXISTS, 0, 'Open file not found');
  end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
  if Assigned(OnEndReport) then OnEndReport(self);
  FExcel.Active := false;
  DeleteFile(FTmpFile);
  FTypeDiraction := tdNotDefined;
end;

procedure TExcelDDEReport.CloseSectionsH;
var
  i, j, k: Integer;
  pp: pTExcelCellContent;
begin
	for i := 1 to MAX_SECTION_NUM do
	  if FSectionsH[i].SectionUsed then
    begin
		  FSectionsH[i].SectionUsed := FALSE;
      StrDispose(PChar(FSectionsH[i].Widths));
      k := FSectionsH[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTExcelCellContent(FSectionsH[i].Contents.Items[j]);
        pp^.ContentColRow := Unassigned;
        StrDispose(PChar(pp));
      end;
      FSectionsH[i].Contents.Destroy;
    end;
end;

procedure TExcelDDEReport.OutSectionH(SectionName: String; bOut: Boolean; SectionNum: LongWord);
var
  dwi: Integer;
  i: Integer;
  pp: pTExcelCellContent;
  ContentColRow: Variant;
  ContentColRowStr: String;
  bCheck: Boolean;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
begin
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
	if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
	if FSectionsH[SectionNum].SectionUsed then
  begin
    if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
		if bOut then
    begin
      Excel.Exec(Format('[COPY("''[%s]%s''!%s", "''[%s]%s''!%s1%s%d")]', [FCurrentWorkBook, InputSheet, SectionName, FCurrentWorkBook, OutputSheet, Excel.RowChar, Excel.ColChar, FCurrCol]));
      for i := 0 to FSectionsH[SectionNum].ColsCount-1 do
        Excel.SetColWidth((pInteger(PChar(FSectionsH[SectionNum].Widths) + i*SizeOf(Integer)))^, Format('%s%d', [Excel.ColChar, FCurrCol + i]));
			dwi := 0;
			while dwi < FSectionsH[SectionNum].Contents.Count do
      begin
        pp := pTExcelCellContent(FSectionsH[SectionNum].Contents.Items[dwi]);
        ContentColRow := pp^.ContentColRow;
        bCheck := false;
        if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, ContentColRow, bCheck, pp^.SectionRow, FCurrCol + pp^.SectionCol - 1, pp^.SectionRow, pp^.SectionCol);
        if bCheck then
          Excel.PutData(pp^.SectionRow, FCurrCol + pp^.SectionCol - 1, ContentColRow)
        else begin
          //
          if Assigned(OnRequestByNum) then
          begin
            pSt := 0;
            while true do
            begin
              ContentColRowStr := VarToStr(ContentColRow);
              pInt := LScan(ContentColRow, FCharByNum, pSt);
              if pInt = 0 then Break;
              pSt := pInt;
              if  ( ContentColRowStr[pInt + 1] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 2] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 3] in ['0'..'9'] ) and
                  ( ContentColRowStr[pInt + 4] in ['0'..'9'] ) then
              begin
                iP := StrToInt(Copy(ContentColRow, pInt + 1, 4));
                sStr := '';
                OnRequestByNum(self, SectionName, iP, sStr);
                if sStr <> '' then
                begin
                  ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt + 5, Length(ContentColRow));
                  Excel.PutData(pp^.SectionRow, FCurrCol + pp^.SectionCol - 1, ContentColRow);
                end;
              end;
            end;
          end;
          if Assigned(OnRequestByName) then
          begin
            pSt := 0;
            while true do
            begin
              pInt := LScan(ContentColRow, FCharByName, pSt);
              if pInt = 0 then Break;
              pInt1 := LScan(ContentColRow, FCharByName, pInt);
              if pInt1 = 0 then Break;
              pSt := pInt1;
              sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
              sStr := '';
              OnRequestByName(self, SectionName, sField, sStr);
              if sStr <> '' then
              begin
                ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
                Excel.PutData(pp^.SectionRow, FCurrCol + pp^.SectionCol - 1, ContentColRow);
              end;
            end;
          end;
          //
        end;
				Inc(dwi);
			end;
			FCurrCol := FCurrCol + FSectionsH[SectionNum].ColsCount;
      if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
		end;
	end;
  if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
end;

procedure TExcelDDEReport.PrepareSectionH(SectionName: String; SectionNum: LongWord);
var
	TopCell: TExcelCell;
	BottomCell: TExcelCell;
  SectionRow, SectionCol: LongWord;
  ContentColRow: String;
  CurrentContetns: String;
  CurrentChar: Char;
  i, j: Integer;
  pp: pTExcelCellContent;
  k: LongWord;
begin
	FSectionsH[SectionNum].SectionName := SectionName;
	FSectionsH[SectionNum].SectionNum := SectionNum;
	FSectionsH[SectionNum].SectionUsed := FALSE;
  Excel.Select(100, 100);
	Excel.Select(SectionName);
	Excel.GetRectSelection(TopCell, BottomCell);
	if not (( TopCell.Row = 100 ) and ( BottomCell.Row = 100 ) and ( TopCell.Col  = 100 ) and ( BottomCell.Col = 100 )) then
  begin
		FSectionsH[SectionNum].SectionUsed := TRUE;
		FSectionsH[SectionNum].ColsCount := BottomCell.Col - TopCell.Col + 1;
    FSectionsH[SectionNum].Contents := TList.Create;
    //
    SectionRow := 1;
    SectionCol := 1;
    ContentColRow := '';
    CurrentContetns := Excel.GetData(SectionName);
    j := Length(CurrentContetns);
    for i := 1 to j do
    begin
      CurrentChar := CurrentContetns[i];
      if  CurrentChar = #9 then
      begin
        //
        if Length(ContentColRow) > 0 then
        begin
          pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
          pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
          pp^.ContentLen := Length(ContentColRow);
          //StrPCopy(pp^.ContentColRow, ContentColRow);
          pp^.SectionRow := SectionRow;
          pp^.SectionCol := SectionCol;
          FSectionsH[SectionNum].Contents.Add(pp);
        end;
        //
        Inc(SectionCol);
        ContentColRow := '';
      end;
      if  CurrentChar = #$D then
      begin
        //
        if Length(ContentColRow) > 0 then
        begin
          pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
          pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
          pp^.ContentLen := Length(ContentColRow);
          //StrPCopy(pp^.ContentColRow, ContentColRow);
          pp^.SectionRow := SectionRow;
          pp^.SectionCol := SectionCol;
          FSectionsH[SectionNum].Contents.Add(pp);
        end;
        //
        SectionCol := 1;
        Inc(SectionRow);
        ContentColRow := '';
      end;
      if  CurrentChar = #$A then ContentColRow := '';
      if not (CurrentChar in [#9, #$A, #$D]) then ContentColRow := ContentColRow + CurrentChar;
    end;
    if ContentColRow <> '' then
    begin
      //
      if Length(ContentColRow) > 0 then
      begin
        pp := pTExcelCellContent(StrAlloc(SizeOf(TExcelCellContent)));
        pp^.ContentColRow := GetContent(ContentColRow); //StrAlloc(Length(ContentColRow)+1);
        pp^.ContentLen := Length(ContentColRow);
        //StrPCopy(pp^.ContentColRow, ContentColRow);
        pp^.SectionRow := SectionRow;
        pp^.SectionCol := SectionCol;
        FSectionsH[SectionNum].Contents.Add(pp);
      end;
      //
    end;
    //
    FSectionsH[SectionNum].Widths := pInteger(StrAlloc(SizeOf(Integer)*FSectionsH[SectionNum].ColsCount));
    for k := 0 to FSectionsH[SectionNum].ColsCount-1 do
    begin
      Excel.select(Format('%s%d', [Excel.ColChar, TopCell.Col + k]));
      (pInteger(PChar(FSectionsH[SectionNum].Widths) + k*SizeOf(Integer)))^ := Excel.GetCellWidth;
    end;
	end;
	Excel.Select(1, 1);
end;

procedure TExcelDDEReport.SetStandardColWidth(var t, b: TExcelCell);
begin
  Excel.SetColWidthToStandard(t, b);
end;

procedure TExcelDDEReport.BiDirectionExecute;
begin
  StartBiDirectionReport;
end;

procedure TExcelDDEReport.BiDirectionResume;
begin
  StartBiDirectionReport;
end;

procedure TExcelDDEReport.StartBiDirectionReport;
var
  OldIgnoreErrors: Boolean;
  bCheck: Boolean;
  sInputSheet: String;
begin
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  FTypeDiraction := tdHorizontaly;
  if Assigned(OnStartReport) then OnStartReport(self);
  FExcel.DDEService := 'Excel';
  FExcel.DDETopic := 'System';
  FExcel.Active := true;
  if Excel.Active then
  begin
    OldIgnoreErrors := Excel.IgnoreErrors;
    Excel.IgnoreErrors := true;
    if FileExists(InputBlank) then
    begin
      FTmpFile := GetTmpFile;
      if CopyFile(pChar(InputBlank), pChar(FTmpFile), False) then
      begin
        Excel.OpenWorkBook(FTmpFile);
        FCurrentWorkBook := Excel.CurrentWorkBook;
        if UpperCase(Trim(FCurrentWorkBook)) = UpperCase(Trim(GetFileName(FTmpFile))) then
        begin
          FCurrCol := 1;
          Excel.CurrentSheet := InputSheet;
          Excel.CopySheet;
          Excel.ClearSheet('', SetStandardColWidth);
          Excel.Select(1, 1);
          OutputSheet := Excel.CurrentSheet;
          Excel.CurrentSheet := InputSheet;
          PrepareSectionH('H_REPORTHEADER', 1);
          PrepareSectionH('H_GROUPHEADER', 2);
          PrepareSectionH('H_SUBGROUPHEADER', 3);
          PrepareSectionH('H_SUBSUBGROUPHEADER', 4);
          PrepareSectionH('H_SUPERSUBGROUPHEADER', 5);
          PrepareSectionH('H_BODY', 6);
          PrepareSectionH('H_SUPERSUBGROUPFOOTER', 7);
          PrepareSectionH('H_SUBSUBGROUPFOOTER', 8);
          PrepareSectionH('H_SUBGROUPFOOTER', 9);
          PrepareSectionH('H_GROUPFOOTER', 10);
          PrepareSectionH('H_REPORTFOOTER', 11);
          Excel.CurrentSheet := OutputSheet;
          lWM_REP_SECTION_PREPARE := true;
          if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
          lWM_REP_SECTION_PREPARE := false;
          OutSectionH('H_REPORTHEADER', TRUE, 1);
          bCheck := True;
          if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
          while not bCheck do
          begin
            OutSectionH('H_GROUPHEADER', FALSE, 2);
            OutSectionH('H_SUBGROUPHEADER', FALSE, 3);
            OutSectionH('H_SUBSUBGROUPHEADER', FALSE, 4);
            OutSectionH('H_SUPERSUBGROUPHEADER', FALSE, 5);
            OutSectionH('H_BODY', TRUE, 6);
            OutSectionH('H_SUPERSUBGROUPFOOTER', FALSE, 7);
            OutSectionH('H_SUBSUBGROUPFOOTER', FALSE, 8);
            OutSectionH('H_SUBGROUPFOOTER', FALSE, 9);
            OutSectionH('H_GROUPFOOTER', FALSE, 10);
            if not Excel.Active then Exit;
            bCheck := True;
            if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
          end;
          OutSectionH('H_REPORTFOOTER', TRUE, 11);
          if Excel.Active then
          begin
            Excel.CurrentSheet := InputSheet;
            Excel.DeleteSheet;
            Excel.CurrentSheet := OutputSheet;
            Excel.select(1, 1);
            //
            FTypeDiraction := tdVerticaly;
            if Assigned(OnStartReport) then OnStartReport(self);
            sInputSheet := InputSheet;
            try
              FCurrRow := 1;
              InputSheet := Excel.CurrentSheet;
              Excel.CopySheet;
              Excel.ClearSheet('', SetStandardRowHeight);
              Excel.Select(1, 1);
              OutputSheet := Excel.CurrentSheet;
              Excel.CurrentSheet := InputSheet;
              if not Excel.Active then Exit;
              PrepareSectionV('REPORTHEADER', 1);
              PrepareSectionV('GROUPHEADER', 2);
              PrepareSectionV('SUBGROUPHEADER', 3);
              PrepareSectionV('SUBSUBGROUPHEADER', 4);
              PrepareSectionV('SUPERSUBGROUPHEADER', 5);
              PrepareSectionV('BODY', 6);
              PrepareSectionV('SUPERSUBGROUPFOOTER', 7);
              PrepareSectionV('SUBSUBGROUPFOOTER', 8);
              PrepareSectionV('SUBGROUPFOOTER', 9);
              PrepareSectionV('GROUPFOOTER', 10);
              PrepareSectionV('REPORTFOOTER', 11);
              Excel.CurrentSheet := OutputSheet;
              if not Excel.Active then Exit;
              lWM_REP_SECTION_PREPARE := true;
              if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
              lWM_REP_SECTION_PREPARE := false;
              OutSectionV('REPORTHEADER', TRUE, 1);
              bCheck := True;
              if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
              while not bCheck do
              begin
                OutSectionV('GROUPHEADER', FALSE, 2);
                OutSectionV('SUBGROUPHEADER', FALSE, 3);
                OutSectionV('SUBSUBGROUPHEADER', FALSE, 4);
                OutSectionV('SUPERSUBGROUPHEADER', FALSE, 5);
                OutSectionV('BODY', TRUE, 6);
                OutSectionV('SUPERSUBGROUPFOOTER', FALSE, 7);
                OutSectionV('SUBSUBGROUPFOOTER', FALSE, 8);
                OutSectionV('SUBGROUPFOOTER', FALSE, 9);
                OutSectionV('GROUPFOOTER', FALSE, 10);
                if not Excel.Active then Exit;
                bCheck := True;
                if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
              end;
              OutSectionV('REPORTFOOTER', TRUE, 11);
            finally
              if Excel.Active then
              begin
                Excel.CurrentSheet := InputSheet;
                Excel.DeleteSheet;
                Excel.CurrentSheet := OutputSheet;
                Excel.SaveAs(OutputBlank);
                Excel.select(1, 1);
              end;
              InputSheet := sInputSheet;
              CloseSectionsV;
              CloseSectionsH;
              Excel.IgnoreErrors := OldIgnoreErrors;
            end;
            //
          end;
          CloseSectionsH;
          Excel.IgnoreErrors := OldIgnoreErrors;
        end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_FILE_NOT_EXISTS, 0, 'Open file not found');
  end else if Assigned(OnReportError) then OnReportError(self, WM_REP_NOT_EXCEL_CONNECTED, 0, 'Not Excel connected');
  if Assigned(OnEndReport) then OnEndReport(self);
  FExcel.Active := false;
  DeleteFile(FTmpFile);
  FTypeDiraction := tdNotDefined;
end;

procedure TExcelDDEReport.SetDdeMgrClient(const Value: TDdeMgrClient);
begin
  Excel.DdeMgrClient := Value;
end;

function TExcelDDEReport.GetDdeMgrClient: TDdeMgrClient;
begin
  Result := Excel.DdeMgrClient;
end;

procedure TExcelDDEReport.OutHorizontalySection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin

  if SectionName = 'H_REPORTHEADER'         then
  begin
    OutSectionH('H_REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPHEADER'          then
  begin
    OutSectionH('H_GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPHEADER'       then
  begin
    OutSectionH('H_SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPHEADER'    then
  begin
    OutSectionH('H_SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPHEADER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'H_BODY'                 then
  begin
    OutSectionH('H_BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPFOOTER'    then
  begin
    OutSectionH('H_SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPFOOTER'       then
  begin
    OutSectionH('H_SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPFOOTER'          then
  begin
    OutSectionH('H_GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'H_REPORTFOOTER'         then
  begin
    OutSectionH('H_REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsH[i].SectionName then
    begin
      OutSectionH(SectionName, true, i);
      Break;
    end;
lEnd:
end;

procedure TExcelDDEReport.OutSection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin
  if SectionName = 'REPORTHEADER'         then
  begin
    OutSectionV('REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'GROUPHEADER'          then
  begin
    OutSectionV('GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPHEADER'       then
  begin
    OutSectionV('SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPHEADER'    then
  begin
    OutSectionV('SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPHEADER'  then
  begin
    OutSectionV('SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'BODY'                 then
  begin
    OutSectionV('BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionV('SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPFOOTER'    then
  begin
    OutSectionV('SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPFOOTER'       then
  begin
    OutSectionV('SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'GROUPFOOTER'          then
  begin
    OutSectionV('GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'REPORTFOOTER'         then
  begin
    OutSectionV('REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsV[i].SectionName then
    begin
      OutSectionV(SectionName, true, i);
      Break;
    end;
lEnd:
end;

function TExcelDDEReport.PrepareHorizontalySection(
  const SectionName: ShortString): Integer;
begin
  if FLastDynSectionH > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSectionH;
  Inc(FLastDynSectionH);
  PrepareSectionH(SectionName, Result);
end;

function TExcelDDEReport.PrepareSection(const SectionName: ShortString): Integer;
begin
  if FLastDynSection > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSection;
  Inc(FLastDynSection);
  PrepareSectionV(SectionName, Result);
end;

procedure TExcelDDEReport.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  OutSectionH(FSectionsH[SectionNum].SectionName, true, SectionNum);
end;

procedure TExcelDDEReport.OutSectionByNum(SectionNum: Integer);
begin
  OutSectionV(FSectionsV[SectionNum].SectionName, true, SectionNum);
end;

function TExcelDDEReport.GetCurrentSection: String;
begin
  Result := FCurrentSection;
end;

function TExcelDDEReport.GetCurrentSectionNum: Integer;
begin
  Result := FCurrentSectionNum;
end;

function TExcelDDEReport.GetTypeDiraction: TTtypeDiraction;
begin
  Result := FTypeDiraction;
end;

function TExcelDDEReport.GetInputSheet: String;
var
  s: String;
  q: Integer;
begin
  if FInputSheet <> '' then
    Result := FInputSheet
  else begin
    s := UpperCase(Trim(FCurrentWorkBook));
    q := Pos('.', s);
    if q = 0 then
      Result := s
    else
      Result := Copy(s, 1, q - 1);
  end;
end;

end.
