unit GridToExcel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, DDEML, StdCtrls,
  DDEClient, dbgrids, DB;

type


  TGridToExcel = class(TComponent)
    private

      FExcel: TExcelDDEClient;
      FGrid: TDBGrid;
      FFast: Boolean;
      FStatus: Boolean;
    function GetDdeMgrClient: TDdeMgrClient;
    procedure SetFDdeMgrClient(const Value: TDdeMgrClient);

    protected

      procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    public

      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      procedure Start;

      property Excel: TExcelDDEClient read FExcel;
      property Status: boolean read FStatus write FStatus;

    published

      property DdeMgrClient: TDdeMgrClient read GetDdeMgrClient write SetFDdeMgrClient;
      property Grid: TDBGrid read FGrid write FGrid;
      property Fast: Boolean read FFast write FFast;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK DDE', [TGridToExcel]);
end;

{ TGridToExcell }

constructor TGridToExcel.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FExcel := TExcelDDEClient.Create(self);
  FFast := True;
  FStatus := false;
end;

destructor TGridToExcel.Destroy;
begin
  FExcel.Destroy;
  inherited Destroy;
end;

function TGridToExcel.GetDdeMgrClient: TDdeMgrClient;
begin
  Result := Excel.DdeMgrClient;
end;

procedure TGridToExcel.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent = FGrid then
      FGrid := nil;
end;

procedure TGridToExcel.SetFDdeMgrClient(const Value: TDdeMgrClient);
begin
  Excel.DdeMgrClient := Value;
end;

procedure TGridToExcel.Start;
var
  i, j, k: Integer;
  sS: String;
  oBookMark: TBookMark;
begin
  if Assigned(FGrid) then
  begin
    FExcel.DDETopic := 'System';
    FExcel.Active := true;
    if FExcel.Active then
    begin
      if FExcel.NewWorkBook then
      begin
        FStatus := true;
        j := FGrid.FieldCount;
        k := 1;
        for i := 1 to j do
        begin
          if not FExcel.Active then Break;
          if not FStatus then Break;
          FExcel.Select(k, i);
          FExcel.SetColWidth(FGrid.Columns.Items[i-1].Width/5);
      		FExcel.Alignment(haHORIZ_ALIGN_CENTER, wTRUE, vaVERT_ALIGN_CENTER);
      		FExcel.FontProperties('', 'Bold');
      		FExcel.Border(lsTHICK_LINE, lsTHICK_LINE, lsTHICK_LINE, lsTHICK_LINE, lsTHICK_LINE);
		      sS := FGrid.Columns.Items[i-1].Title.Caption;
      		FExcel.PutData(k, i,  sS);
          FExcel.select(1, 1);
        end;
        Inc(k);
        if FGrid.DataSource.DataSet.Active then
        begin
          oBookMark := FGrid.DataSource.DataSet.GetBookmark;
          FGrid.DataSource.DataSet.First;
          while not FGrid.DataSource.DataSet.Eof do
          begin
            for i := 1 to j do
            begin
              Application.ProcessMessages;
              if not FExcel.Active then Break;
              if not FStatus then Break;
              if not FFast then
              begin
                FExcel.SELECT(k, i);
                FExcel.Alignment(haHORIZ_ALIGN_CENTER, wTRUE, vaVERT_ALIGN_CENTER);
                FExcel.FontProperties('', '', 10);
                FExcel.Border(lsTHIN_LINE, lsTHIN_LINE, lsTHIN_LINE, lsTHIN_LINE, lsTHIN_LINE);
              end;
              sS := FGrid.Fields[i - 1].AsString;
              FExcel.PutData(k, i,  sS);
            end;
            if not FExcel.Active then Break;
            if not FStatus then Break;
            Inc(k);
            FGrid.DataSource.DataSet.Next;
          end;
          FGrid.DataSource.DataSet.GotoBookmark(oBookMark);
          FGrid.DataSource.DataSet.FreeBookmark(oBookMark);
        end;
        FStatus := false;
      end else raise Exception.Create('Not create workbook!');
      FExcel.Active := false;
    end else raise Exception.Create('Not Excel Started!');
  end else raise Exception.Create('Not assigned grid property!');
end;

end.
