unit ReportContainer;

interface

uses

  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, DDEML, StdCtrls,
  DsgnIntf, SEditProp, DDEReports, RThread, CommonType;

type

  TReportContainer = class(TComponent)
    private

      FReport: TReportAbstract;

      FOnStartReport: TNotifyEvent;
      FOnEndReport: TNotifyEvent;
      FOnReportError: TOnReportError;
      FOnMoveTop: TOnCheck;
      FOnMoveNext: TOnCheck;
      FOnSectionBegin: TOnBeginEndSectionEvent;
      FOnSectionEnd: TOnBeginEndSectionEvent;
      FOnSectionComplit: TOnBeginEndSectionEvent;
      FOnSectionCheck: TOnSectionCheck;
      FOnDataRequest: TOnDataRequest;
      FOnRequestByName: TOnSectionDataRequestByName;
      FOnRequestByNum: TOnSectionDataRequestByNum;
      FOnSectionPrepare: TNotifyEvent;

      FOldOnStartReport: TNotifyEvent;
      FOldOnEndReport: TNotifyEvent;
      FOldOnReportError: TOnReportError;
      FOldOnMoveTop: TOnCheck;
      FOldOnMoveNext: TOnCheck;
      FOldOnSectionBegin: TOnBeginEndSectionEvent;
      FOldOnSectionEnd: TOnBeginEndSectionEvent;
      FOldOnSectionComplit: TOnBeginEndSectionEvent;
      FOldOnSectionCheck: TOnSectionCheck;
      FOldOnDataRequest: TOnDataRequest;
      FOldOnRequestByName: TOnSectionDataRequestByName;
      FOldOnRequestByNum: TOnSectionDataRequestByNum;
      FOldOnSectionPrepare: TNotifyEvent;

      function GetTypeDiraction: TTtypeDiraction;
      procedure ReportPrepare;
      procedure ReportUnPrepare;
      procedure ReportUnPrepareNotify(Sender: TObject);
      function GetCurrentSection: String;
      function GetCurrentSectionNum: Integer;

    protected
    public

      constructor Create(AOwner: TComponent); override;

      procedure ReportResume; virtual;
      procedure ReportExecute; virtual;
      procedure ReportHorizontalyResume; virtual;
      procedure ReportHorizontalyExecute; virtual;
      procedure BiDirectionResume; virtual;
      procedure BiDirectionExecute; virtual;
      function PrepareSection(const SectionName: ShortString): Integer; virtual;
      function PrepareHorizontalySection(const SectionName: ShortString): Integer; virtual;

      procedure OutSection(SectionName: ShortString);
      procedure OutHorizontalySection(SectionName: ShortString);
      procedure OutSectionByNum(SectionNum: Integer);
      procedure OutHorizontalySectionByNum(SectionNum: Integer);

      property CurrentSection: String read GetCurrentSection;
      property CurrentSectionNum: Integer read GetCurrentSectionNum;

      property TypeDiraction: TTtypeDiraction read GetTypeDiraction;

      property OnOldStartReport: TNotifyEvent read FOldOnStartReport;
      property OnOldEndReport: TNotifyEvent read FOldOnEndReport;
      property OnOldReportError: TOnReportError read FOldOnReportError;
      property OnOldMoveTop: TOnCheck read FOldOnMoveTop;
      property OnOldMoveNext: TOnCheck read FOldOnMoveNext;
      property OnOldSectionBegin: TOnBeginEndSectionEvent read FOldOnSectionBegin;
      property OnOldSectionEnd: TOnBeginEndSectionEvent read FOldOnSectionEnd;
      property OnOldSectionComplit: TOnBeginEndSectionEvent read FOldOnSectionComplit;
      property OnOldSectionCheck: TOnSectionCheck read FOldOnSectionCheck;
      property OnOldDataRequest: TOnDataRequest read FOldOnDataRequest;
      property OnOldRequestByName: TOnSectionDataRequestByName read FOldOnRequestByName;
      property OnOldRequestByNum: TOnSectionDataRequestByNum read FOldOnRequestByNum;

    published

      property Report: TReportAbstract read FReport write FReport;

      property OnStartReport: TNotifyEvent read FOnStartReport write FOnStartReport;
      property OnEndReport: TNotifyEvent read FOnEndReport write FOnEndReport;
      property OnReportError: TOnReportError read FOnReportError write FOnReportError;
      property OnMoveTop: TOnCheck read FOnMoveTop write FOnMoveTop;
      property OnMoveNext: TOnCheck read FOnMoveNext write FOnMoveNext;
      property OnSectionBegin: TOnBeginEndSectionEvent read FOnSectionBegin write FOnSectionBegin;
      property OnSectionEnd: TOnBeginEndSectionEvent read FOnSectionEnd write FOnSectionEnd;
      property OnSectionComplit: TOnBeginEndSectionEvent read FOnSectionComplit write FOnSectionComplit;
      property OnSectionCheck: TOnSectionCheck read FOnSectionCheck write FOnSectionCheck;
      property OnDataRequest: TOnDataRequest read FOnDataRequest write FOnDataRequest;
      property OnRequestByName: TOnSectionDataRequestByName read FOnRequestByName write FOnRequestByName;
      property OnRequestByNum: TOnSectionDataRequestByNum read FOnRequestByNum write FOnRequestByNum;
      property OnSectionPrepare: TNotifyEvent read FOnSectionPrepare write FOnSectionPrepare;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TReportContainer]);
end;

{ TReportContainer }

constructor TReportContainer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReport := nil;
end;

procedure TReportContainer.ReportResume;
var
  FThread: TRThread;
begin
  ReportPrepare;
  FThread := TRThread.Create(true);
  FThread.FreeOnTerminate := true;
  if FReport <> nil then FThread.OnExecute := Freport.ReportExecute;
  FThread.OnTerminate := ReportUnPrepareNotify;
  FThread.Resume;
end;

procedure TReportContainer.ReportExecute;
begin
  ReportPrepare;
  if FReport <> nil then Freport.ReportExecute;
  ReportUnPrepare;
end;

procedure TReportContainer.ReportHorizontalyResume;
var
  FThread: TRThread;
begin
  ReportPrepare;
  FThread := TRThread.Create(true);
  FThread.FreeOnTerminate := true;
  if FReport <> nil then FThread.OnExecute := Freport.ReportHorizontalyExecute;
  FThread.OnTerminate := ReportUnPrepareNotify;
  FThread.Resume;
end;

procedure TReportContainer.ReportHorizontalyExecute;
begin
  ReportPrepare;
  if FReport <> nil then Freport.ReportHorizontalyExecute;
  ReportUnPrepare;
end;

procedure TReportContainer.BiDirectionResume;
var
  FThread: TRThread;
begin
  ReportPrepare;
  FThread := TRThread.Create(true);
  FThread.FreeOnTerminate := true;
  if FReport <> nil then FThread.OnExecute := Freport.BiDirectionExecute;
  FThread.OnTerminate := ReportUnPrepareNotify;
  FThread.Resume;
end;

procedure TReportContainer.BiDirectionExecute;
begin
  ReportPrepare;
  if FReport <> nil then Freport.BiDirectionExecute;
  ReportUnPrepare;
end;

procedure TReportContainer.ReportPrepare;
begin

  if Freport <> nil then
  begin

    FOldOnStartReport     :=  Freport.OnStartReport;
    FOldOnEndReport       :=  Freport.OnEndReport;
    FOldOnReportError     :=  Freport.OnReportError;
    FOldOnMoveTop         :=  Freport.OnMoveTop;
    FOldOnMoveNext        :=  Freport.OnMoveNext;
    FOldOnSectionBegin    :=  Freport.OnSectionBegin;
    FOldOnSectionEnd      :=  Freport.OnSectionEnd;
    FOldOnSectionComplit  :=  Freport.OnSectionComplit;
    FOldOnSectionCheck    :=  Freport.OnSectionCheck;
    FOldOnDataRequest     :=  Freport.OnDataRequest;
    FOldOnRequestByName   :=  Freport.OnRequestByName;
    FOldOnRequestByNum    :=  Freport.OnRequestByNum;
    FOldOnSectionPrepare  :=  Freport.OnSectionPrepare;

    if Assigned(FOnStartReport    ) then Freport.OnStartReport     := FOnStartReport;
    if Assigned(FOnEndReport      ) then Freport.OnEndReport       := FOnEndReport;
    if Assigned(FOnReportError    ) then Freport.OnReportError     := FOnReportError;
    if Assigned(FOnMoveTop        ) then Freport.OnMoveTop         := FOnMoveTop;
    if Assigned(FOnMoveNext       ) then Freport.OnMoveNext        := FOnMoveNext;
    if Assigned(FOnSectionBegin   ) then Freport.OnSectionBegin    := FOnSectionBegin;
    if Assigned(FOnSectionEnd     ) then Freport.OnSectionEnd      := FOnSectionEnd;
    if Assigned(FOnSectionComplit ) then Freport.OnSectionComplit  := FOnSectionComplit;
    if Assigned(FOnSectionCheck   ) then Freport.OnSectionCheck    := FOnSectionCheck;
    if Assigned(FOnDataRequest    ) then Freport.OnDataRequest     := FOnDataRequest;
    if Assigned(FOnRequestByName  ) then Freport.OnRequestByName   := FOnRequestByName;
    if Assigned(FOnRequestByNum   ) then Freport.OnRequestByNum    := FOnRequestByNum;
    if Assigned(FOnSectionPrepare ) then Freport.OnSectionPrepare  := FOnSectionPrepare;

  end;

end;

procedure TReportContainer.ReportUnPrepare;
begin

  if Freport <> nil then
  begin

    Freport.OnStartReport     := FOldOnStartReport;
    Freport.OnEndReport       := FOldOnEndReport;
    Freport.OnReportError     := FOldOnReportError;
    Freport.OnMoveTop         := FOldOnMoveTop;
    Freport.OnMoveNext        := FOldOnMoveNext;
    Freport.OnSectionBegin    := FOldOnSectionBegin;
    Freport.OnSectionEnd      := FOldOnSectionEnd;
    Freport.OnSectionComplit  := FOldOnSectionComplit;
    Freport.OnSectionCheck    := FOldOnSectionCheck;
    Freport.OnDataRequest     := FOldOnDataRequest;
    Freport.OnRequestByName   := FOldOnRequestByName;
    Freport.OnRequestByNum    := FOldOnRequestByNum;
    Freport.OnSectionPrepare  := FOldOnSectionPrepare;

  end;

end;

function TReportContainer.GetTypeDiraction: TTtypeDiraction;
begin
  Result := tdNotDefined;
  if FReport <> nil then Result := Freport.TypeDiraction;
end;

procedure TReportContainer.OutSection(SectionName: ShortString);
begin
  if FReport <> nil then Freport.OutSection(SectionName);
end;

function TReportContainer.PrepareHorizontalySection(
  const SectionName: ShortString): Integer;
begin
  Result := 0;
  if FReport <> nil then Result := Freport.PrepareHorizontalySection(SectionName);
end;

function TReportContainer.PrepareSection(
  const SectionName: ShortString): Integer;
begin
  Result := 0;
  if FReport <> nil then Result := Freport.PrepareSection(SectionName);
end;

procedure TReportContainer.ReportUnPrepareNotify(Sender: TObject);
begin
  ReportUnPrepare;
end;

procedure TReportContainer.OutHorizontalySection(SectionName: ShortString);
begin
  if FReport <> nil then Freport.OutHorizontalySection(SectionName);
end;

procedure TReportContainer.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  if FReport <> nil then Freport.OutHorizontalySectionByNum(SectionNum);
end;

procedure TReportContainer.OutSectionByNum(SectionNum: Integer);
begin
  if FReport <> nil then Freport.OutSectionByNum(SectionNum);
end;

function TReportContainer.GetCurrentSection: String;
begin
  Result := '';
  if FReport <> nil then Result := Freport.CurrentSection;
end;

function TReportContainer.GetCurrentSectionNum: Integer;
begin
  Result := -1;
  if FReport <> nil then Result := Freport.CurrentSectionNum;
end;

end.
