unit SSWordStream;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, SSmrtRep, SEditProp, DsgnIntf, CommonType;

type

  TSmartWordStream = class(TComponent)
  private

    { Private declarations }
		FInBlank : String;
		FOutBlank : String;
    FOutStream: TStream;
    FCharByNum: Char;
    FCharByName: Char;

		cS: String;
    cFieldName: String;
		bCheck: Boolean;

		_SRpt: TSmartRpt;

    FOnEndReport: TNotifyEvent;
    FOnStartReport: TNotifyEvent;
    FOnMoveNext: TOnCheck;
    FOnMoveTop: TOnCheck;
    FOnReportError: TOnReportError;

    FOnSectionEnd: TNotifyEvent;
    FOnSectionComplit: TNotifyEvent;
    FOnSectionBegin: TNotifyEvent;
    FOnSectionCheck: TOnCheck;
    FOnSectionDataRequestByName: TOnSectionDataRequestByName;
    FOnSectionDataRequestByNum: TOnSectionDataRequestByNum;
    FOnSectionPrepare: TNotifyEvent;

    procedure SetInBlank(const Value: String);
    procedure SetOutBlank(const Value: String);
    procedure SetOutStream(Value: TStream);

    procedure SectionHandler( var Msg: TMessage;
                              e0: TNotifyEvent;
                              e1: TOnCheck;
                              e2: TOnSectionDataRequestByNum;
                              e3: TOnSectionDataRequestByName;
                              e4: TNotifyEvent;
                              e5: TNotifyEvent;
                              ch: boolean);

    //Message handlers
		procedure MsgStartReport(var Msg: TMessage);
		procedure MsgEndReport(var Msg: TMessage);
		procedure MsgReportMove(var Msg: TMessage);
    procedure MsgReportError(var Msg: TMessage);
    //End message handlers

    procedure DispatchSmartMessages(Sender: TObject; const iMessage, iW, iL: Integer);
    function GetCurrentSection: String;
    function GetCurrentSectionNum: Integer;

  protected
    { Protected declarations }
  public

    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Execute;
		procedure Start;
    function PrepareSection(const SectionName: ShortString): Integer;
    procedure OutSection(SectionName: ShortString);
    procedure OutSectionByNum(SectionNum: Integer);

    property OutStream: TStream read FOutStream write SetOutStream;
    property CurrentSection: String read GetCurrentSection;
    property CurrentSectionNum: Integer read GetCurrentSectionNum;

  published

    { Published declarations }
		//Properties
		property InBlank : String read FInBlank write SetInBlank;
		property OutBlank : String read FOutBlank write SetOutBlank;
    property CharByNum: Char read FCharByNum write FCharByNum default #64;
    property CharByName: Char read FCharByName write FCharByName default #126;

		//Events
    property OnStartReport: TNotifyEvent read FOnStartReport write FOnStartReport;
    property OnEndReport: TNotifyEvent read FOnEndReport write FOnEndReport;

    property OnMoveTop: TOnCheck read FOnMoveTop write FOnMoveTop;
    property OnMoveNext: TOnCheck read FOnMoveNext write FOnMoveNext;

    property OnReportError: TOnReportError read FOnReportError write FOnReportError;

    property OnSectionBegin: TNotifyEvent read FOnSectionBegin write FOnSectionBegin;
    property OnSectionCheck: TOnCheck read FOnSectionCheck write FOnSectionCheck;
    property OnSectionDataRequestByNum: TOnSectionDataRequestByNum read FOnSectionDataRequestByNum write FOnSectionDataRequestByNum;
    property OnSectionDataRequestByName: TOnSectionDataRequestByName read FOnSectionDataRequestByName write FOnSectionDataRequestByName;
    property OnSectionEnd: TNotifyEvent read FOnSectionEnd write FOnSectionEnd;
    property OnSectionComplit: TNotifyEvent read FOnSectionComplit write FOnSectionComplit;

    property OnSectionPrepare: TNotifyEvent read FOnSectionPrepare write FOnSectionPrepare;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TSmartWordStream]);
  RegisterPropertyEditor(TypeInfo(String), TSmartWordStream, 'InBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TSmartWordStream, 'OutBlank', TEdSaveBlank);
end;

constructor TSmartWordStream.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	_SRpt := TSmartRpt.Create();
  CharByNum := #64;
  CharByName := #126;
  _SRpt.OnSmartReportMessage := DispatchSmartMessages;
end;

destructor TSmartWordStream.Destroy;
begin
  _SRpt.Free;
  inherited Destroy;
end;

procedure TSmartWordStream.Start;
begin
	_SRpt.Resume;
end;

procedure TSmartWordStream.Execute;
begin
  _SRpt.Build;
end;

procedure TSmartWordStream.SetOutStream(Value: TStream);
begin
  FOutStream := Value;
	_SRpt.OutStream := Value;
end;

procedure TSmartWordStream.SetInBlank(const Value: String);
begin
  FInBlank := Value;
	_SRpt.InBlank := Value;
end;

procedure TSmartWordStream.SetOutBlank(const Value: String);
begin
  FOutBlank := Value;
  _SRpt.OutBlank := Value;
end;

procedure TSmartWordStream.MsgEndReport(var Msg: TMessage);
begin
  if Assigned(OnEndReport) then OnEndReport(self);
end;

procedure TSmartWordStream.MsgReportError(var Msg: TMessage);
begin
  case Msg.Msg of
    WM_REP_THREAD_ERROR:    cS := 'Thread Error';
    WM_REP_OPEN_ERROR:      cS := 'Open Error';
    WM_REP_CREATE_ERROR:    cS := 'Create Error';
    WM_REP_MEM_ERROR:       cS := 'Memory Error';
    WM_REP_READ_ERROR:      cS := 'Read Error';
    WM_REP_SECTION_ERROR:   cS := 'Section Error';
    WM_REP_WRITE_ERROR:     cS := 'Write Error';
  end;
  if Assigned(OnReportError) then OnReportError(self, Msg, cS);
end;

procedure TSmartWordStream.MsgReportMove(var Msg: TMessage);
begin
  bCheck := true;
  if Msg.LParam = MOVE_TOP then
  begin
    if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
  end
  else
  begin
    if Msg.LParam = MOVE_NEXT then if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
  end;
  _SRpt.CheckRequest := bCheck;
end;

procedure TSmartWordStream.MsgStartReport(var Msg: TMessage);
begin
  _SRpt.SymbolByNumber := CharByNum;
  _SRpt.SymbolByName := CharByName;
  if Assigned(OnStartReport) then OnStartReport(self);
end;

procedure TSmartWordStream.SectionHandler(var Msg: TMessage; e0: TNotifyEvent;
  e1: TOnCheck; e2: TOnSectionDataRequestByNum;
  e3: TOnSectionDataRequestByName; e4: TNotifyEvent; e5: TNotifyEvent; ch: boolean);
begin
  cS := '';
  bCheck := ch;
  case Msg.LParam of
    SECTION_DATA_REQUEST_BY_NUM:
      begin
        if Assigned(e2) then e2(self, Msg.WParam, cS)
          else if Assigned(OnSectionDataRequestByNum) then OnSectionDataRequestByNum(self, Msg.WParam, cS);
        _SRpt.DataRequest := cS;
      end;
    SECTION_DATA_REQUEST_BY_NAME:
      begin
        cFieldName := String(pChar(Msg.WParam));
        if Assigned(e3) then e3(self, cFieldName, cS)
          else if Assigned(OnSectionDataRequestByName) then OnSectionDataRequestByName(self, cFieldName, cS);
        _SRpt.DataRequest := cS;
      end;
    SECTION_CHECK:
      begin
        if Assigned(e1) then e1(self, bCheck)
          else if Assigned(OnSectionCheck) then OnSectionCheck(self, bCheck);
        _SRpt.CheckRequest := bCheck;   //print section with this record or not
      end;
    SECTION_BEGIN: if Assigned(e0) then e0(self) else if Assigned(OnSectionBegin) then OnSectionBegin(self);
    SECTION_END: if Assigned(e4) then e4(self) else if Assigned(OnSectionEnd) then OnSectionEnd(self);
    SECTION_COMPLIT: if Assigned(e5) then e5(self) else if Assigned(OnSectionComplit) then OnSectionComplit(self);
  else
  end;
end;

procedure TSmartWordStream.DispatchSmartMessages(Sender: TObject;
  const iMessage, iW, iL: Integer);
var
  Msg: TMessage;
begin
  Msg.Msg := iMessage;
  Msg.Wparam := iW;
  Msg.Lparam := iL;
  case (iMessage) of
    WM_REP_START_REPORT:           MsgStartReport(Msg);
    WM_REP_END_REPORT:             MsgEndReport(Msg);

    WM_REP_MOVE:                   MsgReportMove(Msg);
    WM_REP_THREAD_ERROR:           MsgReportError(Msg);
    WM_REP_OPEN_ERROR:             MsgReportError(Msg);
    WM_REP_CREATE_ERROR:           MsgReportError(Msg);
    WM_REP_MEM_ERROR:              MsgReportError(Msg);
    WM_REP_READ_ERROR:             MsgReportError(Msg);
    WM_REP_SECTION_ERROR:          MsgReportError(Msg);
    WM_REP_WRITE_ERROR:            MsgReportError(Msg);
    WM_REP_SECTION_PREPARE:
      if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
  else
    if iMessage >= WM_REP_RTF_HEADER then
    begin
      SectionHandler(Msg, OnSectionBegin,
      OnSectionCheck, OnSectionDataRequestByNum,
      OnSectionDataRequestByName, OnSectionEnd, OnSectionComplit, _SRpt.CheckRequest);
    end;
  end;
end;

procedure TSmartWordStream.OutSection(SectionName: ShortString);
begin
  _SRpt.OutSection(SectionName);
end;

function TSmartWordStream.PrepareSection(const SectionName: ShortString): Integer;
begin
  Result := _SRpt.PrepareSection(SectionName);
end;

procedure TSmartWordStream.OutSectionByNum(SectionNum: Integer);
begin
  _SRpt.OutSection(SectionNum);
end;

function TSmartWordStream.GetCurrentSection: String;
begin
  Result := _SRpt.CurrentSection;
end;

function TSmartWordStream.GetCurrentSectionNum: Integer;
begin
  Result := _SRpt.CurrentSectionNum;
end;

end.

