unit SSmrtRep;

interface

uses
	Windows, Messages, Controls, Classes, SysUtils, CommonType;

type

  //
  TOnSmartReportMessage = procedure(Sender: TObject; const iMessage, iW, iL: Integer) of object;

  //
  TOnReportError = procedure(Sender: TObject; var Msg: TMessage; var sStr: String) of object;
  TOnSectionDataRequestByNum = procedure(Sender: TObject; var iP: Longint; var sStr: String) of object;
  TOnSectionDataRequestByName = procedure(Sender: TObject; var sFieldName: String; var sStr: String) of object;
  TOnCheck = procedure(Sender: TObject; var bCheck: Boolean) of object;

  //
	TSmartRpt = class(TThread)
	private

    ucpIn:              String;
	  ucpOut:             String;
    FOutStream:         TStream;
    fhInBlank:          Integer;
    fhOutBlank:         Integer;
    uiSizeIn:           Integer;
    uiOutP:             Integer;
    ucpInBuff:          pChar;            //array [0..FILE_BUFFER-1] of char;
    ucpOutBuff:         pChar;            //array [0..FILE_BUFFER-1] of char;
    cFieldName:         String;
    BY_NUM_SYMBOL:      char;
    BY_NAME_SYMBOL:     char;
    bCheck:             Boolean;
    RequestStr:         String;
    FCurrentSection:    String;
    FCurrentSectionNum: Integer;
    FLastDynSection:    Integer;
    FPrepareDynSection: Boolean;
    lWM_REP_SECTION_PREPARE: boolean;
    SectionNames:       array [0..12] of string;

    Sections:           array [0..MAX_SECTION_NUM] of Section;

    FOnSmartReportMessage: TOnSmartReportMessage;

    procedure PrepareSection(const SectionName: ShortString; const SectionIndex: Integer); overload;
    procedure SeekTextWithBrackets(const ucpW: String; var fppP: FoundPar);
    procedure OutSection(const bDefOut: Boolean; const SectionIndex: Integer); overload;
    function GetSectionName(iIndex: Integer): ShortString;

	protected

    procedure Execute(); override;

	public

    constructor Create;
    destructor Destroy; override;
    procedure Build;
    procedure OutSection(SectionName: ShortString); overload;
    procedure OutSection(SectionNum: Integer); overload;
    function PrepareSection(const SectionName: ShortString): Integer; overload;

    property InBlank: String read ucpIn write ucpIn;
    property OutBlank: String read ucpOut write ucpOut;
    property OutStream: TStream read FOutStream write FOutStream;

    property SymbolByName: char read BY_NAME_SYMBOL write BY_NAME_SYMBOL;
    property SymbolByNumber: char read BY_NUM_SYMBOL write BY_NUM_SYMBOL;

    property DataRequest: String write RequestStr;
    property CheckRequest: Boolean read bCheck write bCheck;

    property OnSmartReportMessage: TOnSmartReportMessage read FOnSmartReportMessage write FOnSmartReportMessage;

    property SectionName[iIndex: Integer]: ShortString read GetSectionName;
    property CurrentSection: String read FCurrentSection;
    property CurrentSectionNum: Integer read FCurrentSectionNum;

	end;

implementation

constructor TSmartRpt.Create();
begin
  inherited Create(true);
  Priority := tpNormal;
  FreeOnTerminate := false;
  FOnSmartReportMessage := nil;
  FOutStream := nil;
  FLastDynSection := 13;
  FPrepareDynSection := false;
  lWM_REP_SECTION_PREPARE := false;
  ucpInBuff := StrAlloc(FILE_BUFFER);
  ucpOutBuff := StrAlloc(FILE_BUFFER);
  SectionNames[FILE_HEADER  		    ] := 'FILEHEADER';
  SectionNames[REPORT_HEADER 			  ] := 'REPORTHEADER';
  SectionNames[GROUP_HEADER				  ] := 'GROUPHEADER';
  SectionNames[SUBGROUP_HEADER		  ] := 'SUBGROUPHEADER';
  SectionNames[SUBSUBGROUP_HEADER	  ] := 'SUBSUBGROUPHEADER';
  SectionNames[SUPERSUBGROUP_HEADER ] := 'SUPERSUBGROUPHEADER';
  SectionNames[BODY								  ] := 'BODY';
  SectionNames[SUPERSUBGROUP_FOOTER ] := 'SUPERSUBGROUPFOOTER';
  SectionNames[SUBSUBGROUP_FOOTER	  ] := 'SUBSUBGROUPFOOTER';
  SectionNames[SUBGROUP_FOOTER			] := 'SUBGROUPFOOTER';
  SectionNames[GROUP_FOOTER				  ] := 'GROUPFOOTER';
  SectionNames[REPORT_FOOTER				] := 'REPORTFOOTER';
  SectionNames[FILE_FOOTER  		  	] := 'FILEFOOTER';
end;

destructor TSmartRpt.Destroy;
begin
  StrDispose(ucpInBuff);
  StrDispose(ucpOutBuff);
  inherited Destroy;
end;

procedure TSmartRpt.Build;
begin
  Execute;
end;

procedure TSmartRpt.Execute;
var
  uiJ, uiRW: Integer;
  fppP, fppQ: FoundPar;
  CreateOK: boolean;
begin
  FLastDynSection := 13;
  if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_START_REPORT, 0, 0);
  fhInBlank := FileOpen(ucpIn, fmOpenRead or fmShareDenyWrite);
  if fhInBlank > 0 then
  begin
    CreateOK := true;
    if not Assigned(FOutStream) then
    begin
      fhOutBlank := FileCreate(ucpOut);
      CreateOK := (fhOutBlank > 0);
    end;
		if CreateOK then
    begin
      //ucpInBuff := StrAlloc(FILE_BUFFER);
      //ucpOutBuff := StrAlloc(FILE_BUFFER);
      //try
        FileSeek(fhInBlank, 0, 0);
        uiSizeIn := FileRead(fhInBlank, ucpInBuff^, FILE_BUFFER);
        if uiSizeIn <> 0 then
        begin
          uiOutP := 0;
          //try
            SeekTextWithBrackets('#BEGIN QREPORT', fppP);
            SeekTextWithBrackets('#END QREPORT', fppQ);
            if fppP.bF then
            begin
              if fppQ.bF then
              begin
                if fppP.bFL then
                begin
                  if fppP.bFR then
                  begin
                    if fppQ.bFL then
                    begin
                      if fppQ.bFR then
                      begin
                        if fppP.uiZ <= fppQ.uiP then
                        begin
                          Sections[0].Name := 'RTF HEADER';
                          Sections[0].StartPosition := 0;
                          Sections[0].Length := fppP.uiP;
                          PrepareSection('REPORT HEADER',       	1);
                          PrepareSection('GROUP HEADER',        	2);
                          PrepareSection('SUBGROUP HEADER',     	3);
                          PrepareSection('SUBSUBGROUP HEADER',  	4);
                          PrepareSection('SUPERSUBGROUP HEADER',	5);
                          PrepareSection('BODY',                	6);
                          PrepareSection('SUPERSUBGROUP FOOTER',	7);
                          PrepareSection('SUBSUBGROUP FOOTER',  	8);
                          PrepareSection('SUBGROUP FOOTER',     	9);
                          PrepareSection('GROUP FOOTER',        	10);
                          PrepareSection('REPORT FOOTER',       	11);
                          Sections[12].Name := 'RTF FOOTER';
                          Sections[12].StartPosition := fppQ.uiZ;
                          Sections[12].Length := uiSizeIn - fppQ.uiZ;
                          OutSection(true, 0);
                          lWM_REP_SECTION_PREPARE := true;
                          if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_PREPARE, 0, 0);
                          lWM_REP_SECTION_PREPARE := false;
                          OutSection(true, 1);
                          uiJ := 1;
                          bCheck := false;
                          if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_MOVE, uiJ, MOVE_TOP);
                          while not bCheck do
                          begin
                            OutSection(false, 2);
                            OutSection(false, 3);
                            OutSection(false, 4);
                            OutSection(false, 5);
                            OutSection(true,  6);
                            OutSection(false, 7);
                            OutSection(false, 8);
                            OutSection(false, 9);
                            OutSection(false, 10);
                            bCheck := false;
                            if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_MOVE, uiJ, MOVE_NEXT);
                            Inc(uiJ);
                          end;
                          OutSection(true,  11);
                          OutSection(true,  12);
                          if uiOutP <> 0 then
                          begin
                            if not Assigned(FOutStream) then
                              uiRW := FileWrite(fhOutBlank, ucpOutBuff^, uiOutP)
                            else
                              uiRW := FOutStream.Write(ucpOutBuff^, uiOutP);
                            if uiOutP <> uiRW then
                            begin
                              if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_WRITE_ERROR, 0, 0);
                              Raise Exception.Create('WRITE_EXCEPTION'); //Exception(WRITE_EXCEPTION, 0, NULL, NULL);
                            end;
                            uiOutP := 0;
                          end;
                        end else begin
                          if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_BEGIN_QREPORT_AFTER_END_QREPORT);
                          Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
                        end;
                      end else begin
                        if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_END_QREPORT_WITHOUT_RIGHT_RTF_TAG);
                        Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
                      end;
                    end else begin
                      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_END_QREPORT_WITHOUT_LEFT_RTF_TAG);
                      Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
                    end;
                  end else begin
                    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_BEGIN_QREPORT_WITHOUT_RIGHT_RTF_TAG);
                    Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
                  end;
                end else begin
                  if  Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_BEGIN_QREPORT_WITHOUT_LEFT_RTF_TAG);
                  Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
                end;
              end else begin
                if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_NOT_END_QREPORT);
                Raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
              end;
            end else begin
              if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, SECTION_ERROR_NOT_BEGIN_QREPORT);
              raise Exception.Create('SECTION_EXCEPTION'); //Exception(SECTION_EXCEPTION, 0, NULL, NULL);
            end;
          //except
          //end;
        end else if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_READ_ERROR, 0, READ_ERROR_ZERO);
      //finally
      //  StrDispose(ucpInBuff);
      //  StrDispose(ucpOutBuff);
      //end;
      if not Assigned(FOutStream) then FileClose(fhOutBlank);
    end else if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_CREATE_ERROR, 0, 0);
    FileClose(fhInBlank);
  end else if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_OPEN_ERROR, 0, 0);
  if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_END_REPORT, 0, 0);
end;

procedure TSmartRpt.OutSection(const bDefOut: Boolean; const SectionIndex: Integer);
var
  uiI, uiDataIndex, i, j, uiRW: Integer;
  cI, a0, a1, a2, a3: char;
  bJ: boolean;
  LenRequestStr: Integer;
  OldInd: Integer;
  OldName: String;
begin
  OldInd := FCurrentSectionNum;
  OldName := FCurrentSection;
  if SectionIndex < 13 then
    FCurrentSection := SectionNames[SectionIndex]
  else
    FCurrentSection := Sections[SectionIndex].Name;
  FCurrentSectionNum := SectionIndex;
  try
    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, 0, SECTION_BEGIN);
    if Sections[SectionIndex].Length <> 0 then
    begin
      bCheck := bDefOut;
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, 0, SECTION_CHECK);
      if ( (SectionIndex = 0) or (SectionIndex = 12) ) then bCheck := true;
      if bCheck then
      begin
        uiI := 0;
        while uiI < Sections[SectionIndex].Length do
        begin
          //Request by number
          if ucpInBuff[Sections[SectionIndex].StartPosition+uiI] = BY_NUM_SYMBOL then   //REQUEST_DATA_BY_NUM_SYMBOL /*@*/
          begin
            //Foword 4 sign
            a0 := ucpInBuff[Sections[SectionIndex].StartPosition + uiI + 1];
            a1 := ucpInBuff[Sections[SectionIndex].StartPosition + uiI + 2];
            a2 := ucpInBuff[Sections[SectionIndex].StartPosition + uiI + 3];
            a3 := ucpInBuff[Sections[SectionIndex].StartPosition + uiI + 4];
            if (  (($30 <= byte(a0)) and (byte(a0) <= $39)) and
                  (($30 <= byte(a1)) and (byte(a1) <= $39)) and
                  (($30 <= byte(a2)) and (byte(a2) <= $39)) and
                  (($30 <= byte(a3)) and (byte(a3) <= $39)) ) then
            begin
              uiDataIndex := (byte(a0) - $30) * 1000 + (byte(a1) - $30) * 100 + (byte(a2) - $30) * 10 + (byte(a3) - $30);
              if uiDataIndex <> 0 then  //Begin with 1  !!!!
              begin
                if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, uiDataIndex, SECTION_DATA_REQUEST_BY_NUM);
                i := 1;
                LenRequestStr := Length(RequestStr);
                while ( i <= LenRequestStr ) do
                begin
                  ucpOutBuff[uiOutP] := RequestStr[i];
                  Inc(uiOutP);
                  Inc(i);
                  if uiOutP = FILE_BUFFER  then
                  begin
                    if not Assigned(FOutStream) then
                      uiRW := FileWrite(fhOutBlank, ucpOutBuff^, FILE_BUFFER)
                    else
                      uiRW := FOutStream.Write(ucpOutBuff^, FILE_BUFFER);
                    uiOutP := 0;
                    if FILE_BUFFER <> uiRW then
                    begin
                      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_WRITE_ERROR, 0, 0);
                      raise Exception.Create('WRITE_EXCEPTION');
                    end;
                  end;
                end;
                Inc(uiI, 5);
              end;
            end;
          end;
          //Request by name
          if ucpInBuff[Sections[SectionIndex].StartPosition + uiI] = BY_NAME_SYMBOL then    //REQUEST_DATA_BY_NAME_SYMBOL /*~*/
          begin
            cFieldName := '';
            bJ := false;
            for j := 1 to MAX_LABEL_NAME_LEGTH do
            begin
              cI := ucpInBuff[Sections[SectionIndex].StartPosition + uiI + j];
              if cI = BY_NAME_SYMBOL then
              begin
                bJ := true;
                break;
              end;
              cFieldName := cFieldName + cI;
            end;
            if bJ then  //if found
            begin
              if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, Integer(Pchar(cFieldName)), SECTION_DATA_REQUEST_BY_NAME);
              i := 1;
              LenRequestStr := Length(RequestStr);
              while i <= LenRequestStr do
              begin
                ucpOutBuff[uiOutP] := RequestStr[i];
                Inc(uiOutP);
                Inc(i);
                if uiOutP = FILE_BUFFER then
                begin
                  if not Assigned(FOutStream) then
                    uiRW := FileWrite(fhOutBlank, ucpOutBuff^, FILE_BUFFER)
                  else
                    uiRW := FOutStream.Write(ucpOutBuff^, FILE_BUFFER);
                  uiOutP := 0;
                  if FILE_BUFFER <> uiRW then
                  begin
                    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_WRITE_ERROR, 0, 0);
                    raise Exception.Create('WRITE_EXCEPTION');
                  end;
                end;
              end;
              uiI := uiI + j + 1;
            end;
          end;
          //
          ucpOutBuff[uiOutP] := ucpInBuff[Sections[SectionIndex].StartPosition + uiI];
          Inc(uiOutP);
          if uiOutP = FILE_BUFFER then
          begin
            if not Assigned(FOutStream) then
              uiRW := FileWrite(fhOutBlank, ucpOutBuff^, FILE_BUFFER)
            else
              uiRW := FOutStream.Write(ucpOutBuff^, FILE_BUFFER);
            uiOutP := 0;
            if FILE_BUFFER <> uiRW then
            begin
              if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_WRITE_ERROR, 0, 0);
              raise Exception.Create('WRITE_EXCEPTION');
            end;
          end;
          Inc(uiI);
        end;
        if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, 0, SECTION_END);
      end;
    end;
  finally
    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, NEW_SECTION_START + SectionIndex, 0, SECTION_COMPLIT);
    FCurrentSectionNum := OldInd;
    FCurrentSection := OldName;
  end;
end;

procedure TSmartRpt.PrepareSection(const SectionName: ShortString; const SectionIndex: Integer);
var
  fppP, fppQ: FoundPar;
begin
  Sections[SectionIndex].Name := SectionName;
  Sections[SectionIndex].StartPosition := 0;
  Sections[SectionIndex].Length := 0;
  if FPrepareDynSection then
  begin
    SeekTextWithBrackets('#BEGIN SECTION \lquote ' + SectionName + '\rquote', fppP);
    SeekTextWithBrackets('#END SECTION \lquote ' + SectionName + '\rquote', fppQ);
    FPrepareDynSection := false;
  end else begin
    SeekTextWithBrackets('#BEGIN ' + SectionName, fppP);
    SeekTextWithBrackets('#END ' + SectionName, fppQ);
  end;
  if ( (fppP.bF and fppQ.bF) and (fppP.bFL and fppQ.bFL) and (fppP.bFR and fppQ.bFR) ) then
  begin
    if (fppQ.uiP >=  fppP.uiZ) then
    begin
      Sections[SectionIndex].StartPosition := fppP.uiZ;
      Sections[SectionIndex].Length := fppQ.uiP - fppP.uiZ;
    end else begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 8);  //*_END_*_BEFORE_BEGIN_*
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
  end else begin
    if ( fppQ.bF and (not fppQ.bFL) and (not fppQ.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 0);   //*_END_*_WITHOUT_LEFT_AND_RIGHT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppQ.bF and (not fppQ.bFL) and fppQ.bFR ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 1);  //*_END_*_WITHOUT_LEFT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppQ.bF and fppQ.bFL and (not fppQ.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 2);  //*_END_*_WITHOUT_RIGHT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppP.bF and (not fppP.bFL) and (not fppP.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 3);  //*_BEGIN_*_WITHOUT_LEFT_AND_RIGHT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppP.bF and (not fppP.bFL) and fppP.bFR ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 4);  //*_BEGIN_*_WITHOUT_LEFT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppP.bF and fppP.bFL and (not fppP.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 5);  //*_BEGIN_*_WITHOUT_RIGHT_RTF_TAG
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppP.bF and (not fppQ.bF) and (fppP.bFL and fppQ.bFL) and (fppP.bFR and fppQ.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 6);  //*_BEGIN_*_WITHOUT_END_*
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
    if ( fppP.bF and (not fppQ.bF) and (fppP.bFL and fppQ.bFL) and (fppP.bFR and fppQ.bFR) ) then
    begin
      if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_SECTION_ERROR, 0, (NEW_SECTION_ERROR_START + (SectionIndex - 1) * SECTION_ERROR_NUMS) + 7);  //*_END_*_WITHOUT_BEGIN_*
      Raise Exception.Create('SECTION_EXCEPTION');
    end;
  end;
end;

procedure TSmartRpt.SeekTextWithBrackets(const ucpW: String; var fppP: FoundPar);
var
  uiI, LenW: Integer;
begin
  fppP.bF := false;
  LenW := Length(ucpW);
  for uiI := 0 to uiSizeIn-1 do
  begin
    if StrLComp(Pchar(@ucpInBuff[uiI]), Pchar(ucpW), LenW) = 0 then
    begin
      fppP.bF := true;
      break;
    end;
  end;
  fppP.bFL := true;
  fppP.bFR := true;
  fppP.uiP := uiI;
  fppP.uiZ := uiI + Length(ucpW);
  if ( fppP.bF ) then
  begin
    while( true ) do
    begin
      if ( fppP.uiP <> 0 ) then
      begin
        Dec(fppP.uiP);
        if ( byte(ucpInBuff[fppP.uiP]) = RTF_SECTION_BEGIN_BREAKED ) then break;  //{
        if ( byte(ucpInBuff[fppP.uiP]) = RTF_SECTION_END_BREAKED ) then           //}
        begin
          fppP.bFL := false;
          break;
        end;
      end else begin
        fppP.bFL := false;
        break;
      end;
    end;
    while(true) do
    begin
      if (fppP.uiZ < uiSizeIn) then
      begin
        Inc(fppP.uiZ);
        if ( byte(ucpInBuff[fppP.uiZ]) = RTF_SECTION_END_BREAKED ) then    //}
        begin
          Inc(fppP.uiZ);
          break;
        end;
        if ( byte(ucpInBuff[fppP.uiZ]) = RTF_SECTION_BEGIN_BREAKED ) then  //{
        begin
          fppP.bFR := false;
          break;
        end;
      end else begin
        fppP.bFR := false;
        break;
      end;
    end;
  end;
end;

procedure TSmartRpt.OutSection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin
  if SectionName = 'REPORTHEADER' then
  begin
    OutSection(true, 1);
    goto lEnd;
  end;
  if SectionName = 'GROUPHEADER' then
  begin
    OutSection(true, 2);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPHEADER' then
  begin
    OutSection(true, 3);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPHEADER' then
  begin
    OutSection(true, 4);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPHEADER' then
  begin
    OutSection(true, 5);
    goto lEnd;
  end;
  if SectionName = 'BODY' then
  begin
    OutSection(true, 6);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPFOOTER' then
  begin
    OutSection(true, 7);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPFOOTER' then
  begin
    OutSection(true, 8);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPFOOTER' then
  begin
    OutSection(true, 9);
    goto lEnd;
  end;
  if SectionName = 'GROUPFOOTER' then
  begin
    OutSection(true, 10);
    goto lEnd;
  end;
  if SectionName = 'REPORTFOOTER' then
  begin
    OutSection(true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = Sections[i].Name then
    begin
      OutSection(true, i);
      Break;
    end;
lEnd:
end;

function TSmartRpt.PrepareSection(const SectionName: ShortString): Integer;
begin
  if FLastDynSection > MAX_SECTION_NUM then
  begin
    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 0);
    Raise Exception.Create('LIMIT_EXCEEDED');
  end;
  if not lWM_REP_SECTION_PREPARE then
  begin
    if Assigned(FOnSmartReportMessage) then FOnSmartReportMessage(self, WM_REP_ALLAGIAL_CALL, 0, 0);
    Raise Exception.Create('ALLAGIAL_CALL');
  end;
  Result := FLastDynSection;
  Inc(FLastDynSection);
  FPrepareDynSection := true;
  PrepareSection(SectionName, Result);
  FPrepareDynSection := false;
end;

function TSmartRpt.GetSectionName(iIndex: Integer): ShortString;
begin
  Result := Sections[iIndex].Name;
end;

procedure TSmartRpt.OutSection(SectionNum: Integer);
begin
  OutSection(true, SectionNum);
end;

end.
