unit SylkReports;

interface

uses
	Windows, Messages, Controls, Classes, SysUtils, DsgnIntf, SEditProp, ERep, DDEReports, CommonType;

type

  TSylkReport = class(TReportAbstract)
	private

    FSylkStrim: TSylkStream;
    FRqData: DataRq;

    function GetInputBlank: String;
    function GetOutputBlank: String;
    procedure SetInputBlank(const Value: String);
    procedure SetOutputBlank(const Value: String);
    function GetCharByName: Char;
    function GetCharByNum: Char;
    procedure SetCharByName(const Value: Char);
    procedure SetCharByNum(const Value: Char);

    procedure OnErrorH(Sender: TObject; var Msg: TMessage; var sStr: String);

    procedure lOnSectionBegin(Sender: TObject);
    procedure lOnSectionCheck(Sender: TObject; var bCheck: Boolean);
    procedure lOnSectionDataRequestByCell(Sender: TObject; var RqData: DataRq);
    procedure lOnSectionDataRequestByVal(Sender: TObject; var iP: Longint; var sStr: String);
    procedure lOnSectionRqByNum(Sender: TObject; var iP: Longint; var sStr: String);
    procedure lOnSectionRqByName(Sender: TObject; var sFieldName: String; var sStr: String);
    procedure lOnSectionEnd(Sender: TObject);
    procedure lOnSectionComplit(Sender: TObject);

    procedure EventsRedirect;

  protected

    function GetCurrentSection: String; override;
    function GetCurrentSectionNum: Integer; override;
    function GetTypeDiraction: TTtypeDiraction; override;

  public

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    procedure ReportResume; override;
    procedure ReportExecute; override;
    procedure ReportHorizontalyResume; override;
    procedure ReportHorizontalyExecute; override;
    procedure BiDirectionResume; override;
    procedure BiDirectionExecute; override;
    function PrepareSection(const SectionName: ShortString): Integer; override;
    function PrepareHorizontalySection(const SectionName: ShortString): Integer; override;

    procedure OutSection(SectionName: ShortString); override;
    procedure OutHorizontalySection(SectionName: ShortString); override;
    procedure OutSectionByNum(SectionNum: Integer); override;
    procedure OutHorizontalySectionByNum(SectionNum: Integer); override;

    property SylkStrim: TSylkStream read FSylkStrim;

  published

    property InputBlank: String read GetInputBlank write SetInputBlank;
    property OutputBlank: String read GetOutputBlank write SetOutputBlank;
    property CharByNum: Char read GetCharByNum write SetCharByNum;
    property CharByName: Char read GetCharByName write SetCharByName;

    property OnStartReport;
    property OnEndReport;
    property OnReportError;
    property OnMoveTop;
    property OnMoveNext;
    property OnSectionBegin;
    property OnSectionEnd;
    property OnSectionComplit;
    property OnSectionCheck;
    property OnDataRequest;
    property OnRequestByName;
    property OnRequestByNum;
    property OnSectionPrepare;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TSylkReport]);
  RegisterPropertyEditor(TypeInfo(String), TSylkStream, 'InputBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TSylkStream, 'OutputBlank', TEdSaveBlank);
end;

{ TSylkReport }

procedure TSylkReport.BiDirectionExecute;
begin
  EventsRedirect;
  FSylkStrim.BiDirectionExecute;
end;

procedure TSylkReport.BiDirectionResume;
begin
  EventsRedirect;
  FSylkStrim.BiDirectionResume;
end;

constructor TSylkReport.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FSylkStrim := TSylkStream.Create(self);
end;

destructor TSylkReport.Destroy;
begin
  FSylkStrim.Destroy;
  inherited Destroy;
end;

procedure TSylkReport.EventsRedirect;
begin

  FSylkStrim.OnStartReport := OnStartReport;
  FSylkStrim.OnEndReport := OnEndReport;

  FSylkStrim.OnMoveTop := OnMoveTop;
  FSylkStrim.OnMoveNext := OnMoveNext;

  FSylkStrim.OnReportError := OnErrorH;

  FSylkStrim.OnSectionBegin :=             lOnSectionBegin;
  FSylkStrim.OnSectionCheck :=             lOnSectionCheck;
  FSylkStrim.OnSectionDataRequestByCell := lOnSectionDataRequestByCell;
  FSylkStrim.OnSectionDataRequestByVal :=  lOnSectionDataRequestByVal;
  FSylkStrim.OnSectionDataRequestByNum :=  lOnSectionRqByNum;
  FSylkStrim.OnSectionDataRequestByName := lOnSectionRqByName;
  FSylkStrim.OnSectionEnd :=               lOnSectionEnd;
  FSylkStrim.OnSectionComplit :=           lOnSectionComplit;

  FSylkStrim.OnSectionPrepare := OnSectionPrepare;

end;

function TSylkReport.GetCharByName: Char;
begin
  Result := FSylkStrim.CharByName;
end;

function TSylkReport.GetCharByNum: Char;
begin
  Result := FSylkStrim.CharByNum;
end;

function TSylkReport.GetCurrentSection: String;
begin
  Result := FSylkStrim.CurrentSection;
end;

function TSylkReport.GetCurrentSectionNum: Integer;
begin
  Result := FSylkStrim.CurrentSectionNum;
end;

function TSylkReport.GetInputBlank: String;
begin
  Result := FSylkStrim.InputBlank;
end;

function TSylkReport.GetOutputBlank: String;
begin
  Result := FSylkStrim.OutputBlank;
end;

function TSylkReport.GetTypeDiraction: TTtypeDiraction;
begin
  Result := FSylkStrim.TypeDiraction;
end;

procedure TSylkReport.lOnSectionBegin(Sender: TObject);
begin
  if Assigned(OnSectionBegin) then OnSectionBegin(self, FSylkStrim.CurrentSection, FSylkStrim.CurrentSectionNum);
end;

procedure TSylkReport.lOnSectionCheck(Sender: TObject;
  var bCheck: Boolean);
begin
  if Assigned(OnSectionCheck) then OnSectionCheck(self, FSylkStrim.CurrentSection, FSylkStrim.CurrentSectionNum, bCheck);
end;

procedure TSylkReport.lOnSectionComplit(Sender: TObject);
begin
  if Assigned(OnSectionComplit) then OnSectionComplit(self, FSylkStrim.CurrentSection, FSylkStrim.CurrentSectionNum);
end;

procedure TSylkReport.lOnSectionDataRequestByCell(Sender: TObject;
  var RqData: DataRq);
begin
  FRqData := RqData;
end;

procedure TSylkReport.lOnSectionDataRequestByVal(Sender: TObject;
  var iP: Integer; var sStr: String);
var
  bCheck: Boolean;
  SVar: Variant;
begin
  SVar := iP;
  bCheck := false;
  if Assigned(OnDataRequest) then OnDataRequest(self,
                                                FSylkStrim.CurrentSection,
                                                FSylkStrim.CurrentSectionNum,
                                                SVar,
                                                bCheck,
                                                FRqData.CurrCell.iRow,
                                                FRqData.CurrCell.iCol,
                                                FRqData.BlankCell.iRow,
                                                FRqData.BlankCell.iCol);
  if not bCheck then sStr := '' else begin
    sStr := VarToStr(SVar);
    if (VarType(SVar) = varOleStr) or (VarType(SVar) = varString) then begin
      sStr := '"' + sStr + '"';
    end;
  end;
end;

procedure TSylkReport.lOnSectionEnd(Sender: TObject);
begin
  if Assigned(OnSectionEnd) then OnSectionEnd(self, FSylkStrim.CurrentSection, FSylkStrim.CurrentSectionNum);
end;

procedure TSylkReport.lOnSectionRqByName(Sender: TObject; var sFieldName,
  sStr: String);
begin
  if Assigned(OnRequestByName) then OnRequestByName(self, FSylkStrim.CurrentSection, sFieldName, sStr);
end;

procedure TSylkReport.lOnSectionRqByNum(Sender: TObject; var iP: Integer;
  var sStr: String);
begin
  if Assigned(OnRequestByNum) then OnRequestByNum(self, FSylkStrim.CurrentSection, iP, sStr);
end;

procedure TSylkReport.OnErrorH(Sender: TObject; var Msg: TMessage;
  var sStr: String);
begin
  if Assigned(OnReportError) then OnReportError(self, Msg.Msg, Msg.LParam, sStr);
end;

procedure TSylkReport.OutHorizontalySection(SectionName: ShortString);
begin
  FSylkStrim.OutHorizontalySection(SectionName);
end;

procedure TSylkReport.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  FSylkStrim.OutHorizontalySectionByNum(SectionNum);
end;

procedure TSylkReport.OutSection(SectionName: ShortString);
begin
  FSylkStrim.OutSection(SectionName);
end;

procedure TSylkReport.OutSectionByNum(SectionNum: Integer);
begin
  FSylkStrim.OutSectionByNum(SectionNum);
end;

function TSylkReport.PrepareHorizontalySection(
  const SectionName: ShortString): Integer;
begin
  Result := FSylkStrim.PrepareHorizontalySection(SectionName);
end;

function TSylkReport.PrepareSection(
  const SectionName: ShortString): Integer;
begin
  Result := FSylkStrim.PrepareSection(SectionName);
end;

procedure TSylkReport.ReportExecute;
begin
  EventsRedirect;
  FSylkStrim.ReportExecute;
end;

procedure TSylkReport.ReportHorizontalyExecute;
begin
  EventsRedirect;
  FSylkStrim.ReportHorizontalyExecute;
end;

procedure TSylkReport.ReportHorizontalyResume;
begin
  EventsRedirect;
  FSylkStrim.ReportHorizontalyResume;
end;

procedure TSylkReport.ReportResume;
begin
  EventsRedirect;
  FSylkStrim.ReportResume;
end;

procedure TSylkReport.SetCharByName(const Value: Char);
begin
  FSylkStrim.CharByName := Value;
end;

procedure TSylkReport.SetCharByNum(const Value: Char);
begin
  FSylkStrim.CharByNum := Value;
end;

procedure TSylkReport.SetInputBlank(const Value: String);
begin
  FSylkStrim.InputBlank := Value;
end;

procedure TSylkReport.SetOutputBlank(const Value: String);
begin
  FSylkStrim.OutputBlank := Value;
end;

end.
