unit WordDDEClient;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DDEClient;

type
  TWordDDEClient = class(TDDEClient)
  private
    { Private declarations }

		FFileExt: String;
		FExecutable: String;

    FSysItemsList: TStringList;
    FTopicsList: TStringList;
    FFormatsList: TStringList;

    procedure LocateMSWord; virtual;
    class function CountMSWordInstance: LongInt;
    function GetBookMarkCount: Integer;
    function GetBookMarkName(Index: Integer): String;

  protected
    { Protected declarations }

    procedure Loaded; override;
    procedure TerminateTransaction(Sender: TObject; sCmd: String); override;
    function GetStrings(var List: TStringList; const Topic: string): TStringList;
    procedure SetActive(const Value: Boolean); override;
    procedure SetDdeTopic(const Value: String); override;
    function IsLoadedServiceApp: Boolean; override;

  public
    { Public declarations }

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    procedure Connect;
    procedure Disconnect;

    function SwitchTopic(const NewTopic: String): Boolean;
    function SwitchToSystemTopic: Boolean;
    function GetSysItems: TStringList;
    function GetFormats: TStringList;
    function GetTopics: TStringList;

    function OpenDocument(sDoc: String; ConfirmConversions: Boolean = False; ReadOnly: Boolean = False): boolean;
    function CloseDocument: boolean;
    function AppMaximize: Boolean;
    function AppActivate: Boolean;

    property SysItems: TStringList read GetSysItems;
    property Formats: TStringList read GetFormats;
    property Topics: TStringList read GetTopics;

    property BookMarkCount: Integer read GetBookMarkCount;
    property BookMarkName[Index: Integer]: String read GetBookMarkName;

  published
    { Published declarations }

    property FileExt: String read FFileExt write FFileExt;
    property Executable: String read FExecutable write FExecutable;

  end;

procedure Register;
function EnumMSWordWindows(AHandle: THandle; PInstanceCount: LPARAM): Boolean; stdcall;

implementation

procedure Register;
begin
  RegisterComponents('VK DDE', [TWordDDEClient]);
end;

function EnumMSWordWindows(AHandle: THandle; PInstanceCount: LPARAM): Boolean; stdcall;
type
  PLongint = ^Longint;
var
  ClassName: array[0..8] of Char;
begin
  if GetClassName(AHandle, ClassName, SizeOf(ClassName)) = 0 then
    raise Exception.Create('WinApi Error');
  if StrComp(ClassName, 'OpusApp') = 0 then
    Inc(PLongint(PInstanceCount)^);
  Result := (PLongint(PInstanceCount)^ < 2);
end;

{ TWordDDEClient }

function TWordDDEClient.AppActivate: Boolean;
begin
  Result := Exec('[AppActivate "Microsoft Word", 1 ]');
end;

function TWordDDEClient.AppMaximize: Boolean;
begin
  Result := Exec('[AppMaximize "Microsoft Word", 1 ]');
end;

function TWordDDEClient.CloseDocument: boolean;
begin
  Result := Exec('[FileClose ]');
end;

procedure TWordDDEClient.Connect;
var
  oErr: Exception;
begin
  if not Active then
  begin
    CloseLink;
    if OpenLink then Exit;
    LocateMSWord;
    if not OpenLink then
      if not OpenLink then
      begin
  			oErr := Exception.Create('MSWord not started!');
    		if ErrorProcessed then raise oErr else if Assigned(OnError) then OnError(oErr);
      end;
  end;
end;

class function TWordDDEClient.CountMSWordInstance: LongInt;
var
  MSWordCounter: Longint;
begin
  MSWordCounter := 0;
  EnumWindows(@EnumMSWordWindows, LPARAM(@MSWordCounter));
  Result := MSWordCounter;
end;

constructor TWordDDEClient.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  Disconnect;
	FFileExt := '.DOC';
	FExecutable := 'WINWORD.EXE';
  ServiceApplication := 'WINWORD';
  FDDEService := 'WinWord';
  FDDETopic := 'System';
end;

destructor TWordDDEClient.Destroy;
begin
  Disconnect;
  if Assigned(FSysItemsList) then FSysItemsList.Destroy;
  if Assigned(FTopicsList) then FTopicsList.Destroy;
  if Assigned(FFormatsList) then FFormatsList.Destroy;
  inherited Destroy;
end;

procedure TWordDDEClient.Disconnect;
begin
  if Active then
    CloseLink;
end;

function TWordDDEClient.GetBookMarkCount: Integer;
var
  Commands: String;
begin
  Commands :=
    '[EndOfDocument]' +
    '[EditBookmark .Name = "TempBookMark", .Add]' +
    '[Insert RTrim$(LTrim$(Str$(CountBookmarks())))]' +
    '[EditBookmark .Name = "TempBookMark", .GoTo]' +
    '[EndOfDocument 1]' +
    '[EditBookmark .Name = "TempBookMark1", .Add]';
  Exec(Commands);
  try
    Result := StrToInt(Trim(Request('TempBookMark1'))) - 1;
  except
    Result := 0;
  end;
  Commands :=
    '[EditBookmark .Name = "TempBookMark1", .Delete]' +
    '[EditClear]' +
    '[EditBookmark .Name = "TempBookMark", .Delete]';
  Exec(Commands);
end;

function TWordDDEClient.GetBookMarkName(Index: Integer): String;
var
  Commands: String;
begin
  Commands :=
    '[EndOfDocument]' +
    '[EditBookmark .Name = "TempBookMark", .Add]' +
    '[Insert RTrim$(LTrim$(BookmarkName$(' + IntToStr(Index) + ')))]' +
    '[EditBookmark .Name = "TempBookMark", .GoTo]' +
    '[EndOfDocument 1]' +
    '[EditBookmark .Name = "TempBookMark1", .Add]';
  Exec(Commands);
  try
    Result := Trim(Request('TempBookMark1'));
  except
    Result := '';
  end;
  Commands :=
    '[EditBookmark .Name = "TempBookMark1", .Delete]' +
    '[EditClear]' +
    '[EditBookmark .Name = "TempBookMark", .Delete]';
  Exec(Commands);
end;

function TWordDDEClient.GetFormats: TStringList;
begin
  Result := GetStrings(FFormatsList, 'Formats');
end;

function TWordDDEClient.GetStrings(var List: TStringList;
  const Topic: string): TStringList;
var
  Reply: String;
  OldTopic: String;
begin
  if not Assigned(List) then
    List := TStringList.Create
  else
    List.Clear;
  OldTopic := DDETopic;
  SwitchToSystemTopic;
  Reply := Request(Topic);
  TABStringToStringList(PChar(Reply), List);
  if DDETopic <> OldTopic then
    DDETopic := OldTopic;
  Result := List;
end;

function TWordDDEClient.GetSysItems: TStringList;
begin
  Result := GetStrings(FTopicsList, 'SysItems');
end;

function TWordDDEClient.GetTopics: TStringList;
begin
  Result := GetStrings(FTopicsList, 'Topics');
end;

function TWordDDEClient.IsLoadedServiceApp: Boolean;
begin
  Result := ( CountMSWordInstance > 0 );
end;

procedure TWordDDEClient.Loaded;
begin
  inherited Loaded;
end;

procedure TWordDDEClient.LocateMSWord;
const
  BufferSize = 540;
var
  Buffer: PChar;
  StringPosition: PChar;
  ReturnedData: Longint;
  OldServiceApp: String;
begin
  OldServiceApp := ServiceApplication;
  Buffer := StrAlloc(BufferSize);
  try
    ReturnedData := BufferSize;
    StrPCopy(Buffer, FFileExt);
    RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData);
    if StrLen(Buffer) > 0 then
    begin
      StrCat(Buffer, '\shell\Open\command');
      ReturnedData := BufferSize;
      if RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData) = ERROR_SUCCESS then
      begin
        StringPosition := StrUpper(Buffer);
        while StrPos(StringPosition+1, PChar(FExecutable)) <> nil do
          StringPosition := StrPos(StringPosition+1, PChar(FExecutable));
        StringPosition[Length(FExecutable)] := #0;
        StringPosition := StrScan(Buffer, '"');
        if StringPosition <> nil then
        begin
          ReturnedData :=  Longint(StrLen(Buffer)) - (StringPosition-Buffer) - 1;
          StrMove(Buffer, Buffer + (StringPosition-Buffer) + 1, ReturnedData);
          Buffer[ReturnedData] := #0;
        end;
        ServiceApplication := StrPas(Buffer);
        if not FileExists(ServiceApplication) then ServiceApplication := OldServiceApp;
      end;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

function TWordDDEClient.OpenDocument(sDoc: String; ConfirmConversions: Boolean; ReadOnly: Boolean): boolean;
var
  AddititionPar: String;
begin
  AddititionPar := '';
  if ConfirmConversions then
    AddititionPar := ' .ConfirmConversions = 1 '
  else
    AddititionPar := ' .ConfirmConversions = 0 ';
  if ReadOnly then
    AddititionPar := AddititionPar + '.ReadOnly = 1 '
  else
    AddititionPar := AddititionPar + '.ReadOnly = 0 ';
  Result := Exec('[FileOpen .Name="'+sDoc+'" '+AddititionPar+']');
  //SwitchTopic(sDoc);
end;

procedure TWordDDEClient.SetActive(const Value: Boolean);
begin
  if ( csReading in ComponentState ) then FStreamedActive := Value
  else
  begin
    if Value then
      Connect
    else Disconnect;
  end;
end;

procedure TWordDDEClient.SetDdeTopic(const Value: String);
begin
  SwitchTopic(Value);
end;

function TWordDDEClient.SwitchTopic(const NewTopic: String): Boolean;
begin
  Result := True;
	if DdeTopic <> NewTopic then
		Result := SetLink('WinWord', NewTopic);
end;

function TWordDDEClient.SwitchToSystemTopic: Boolean;
begin
  Result := SwitchTopic('System');
end;

procedure TWordDDEClient.TerminateTransaction(Sender: TObject;
  sCmd: String);
begin
	if not Active then
		Active := True;
end;

end.
