unit WordReports;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, DDEML, StdCtrls,
  DDEClient, DDEReports, SSWordStream, DsgnIntf, SEditProp, CommonType;

type

  TWordReport = class(TReportAbstract)
  private

    FWordStrim: TSmartWordStream;

    function GetInputBlank: String;
    function GetOutputBlank: String;
    procedure SetInputBlank(const Value: String);
    procedure SetOutputBlank(const Value: String);
    function GetCharByName: Char;
    function GetCharByNum: Char;
    procedure SetCharByName(const Value: Char);
    procedure SetCharByNum(const Value: Char);

    procedure OnErrorH(Sender: TObject; var Msg: TMessage; var sStr: String);

    procedure lOnSectionBegin(Sender: TObject);
    procedure lOnSectionCheck(Sender: TObject; var bCheck: Boolean);
    procedure lOnSectionRqByNum(Sender: TObject; var iP: Longint; var sStr: String);
    procedure lOnSectionRqByName(Sender: TObject; var sFieldName: String; var sStr: String);
    procedure lOnSectionEnd(Sender: TObject);
    procedure lOnSectionComplit(Sender: TObject);

    procedure EventsRedirect;

  protected
  
    function GetCurrentSection: String; override;
    function GetCurrentSectionNum: Integer; override;
    function GetTypeDiraction: TTtypeDiraction; override;

  public

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    procedure ReportResume; override;
    procedure ReportExecute; override;
    procedure ReportHorizontalyResume; override;
    procedure ReportHorizontalyExecute; override;
    procedure BiDirectionResume; override;
    procedure BiDirectionExecute; override;
    function PrepareSection(const SectionName: ShortString): Integer; override;
    function PrepareHorizontalySection(const SectionName: ShortString): Integer; override;

    procedure OutSection(SectionName: ShortString); override;
    procedure OutHorizontalySection(SectionName: ShortString); override;
    procedure OutSectionByNum(SectionNum: Integer); override;
    procedure OutHorizontalySectionByNum(SectionNum: Integer); override;

    property WordStrim: TSmartWordStream read FWordStrim;

  published

    property InputBlank: String read GetInputBlank write SetInputBlank;
    property OutputBlank: String read GetOutputBlank write SetOutputBlank;
    property CharByNum: Char read GetCharByNum write SetCharByNum default #64;
    property CharByName: Char read GetCharByName write SetCharByName default #126;

    property OnStartReport;
    property OnEndReport;
    property OnReportError;
    property OnMoveTop;
    property OnMoveNext;
    property OnSectionBegin;
    property OnSectionEnd;
    property OnSectionComplit;
    property OnSectionCheck;
    property OnDataRequest;
    property OnRequestByName;
    property OnRequestByNum;
    property OnSectionPrepare;

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TWordReport]);
  RegisterPropertyEditor(TypeInfo(String), TWordReport, 'InputBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TWordReport, 'OutputBlank', TEdSaveBlank);
end;

{ TWordReport }

procedure TWordReport.BiDirectionExecute;
begin
end;

procedure TWordReport.BiDirectionResume;
begin
end;

constructor TWordReport.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FWordStrim := TSmartWordStream.Create(self);
end;

destructor TWordReport.Destroy;
begin
  FWordStrim.Free;
  inherited Destroy;
end;

function TWordReport.GetCharByName: Char;
begin
  Result := FWordStrim.CharByName;
end;

function TWordReport.GetCharByNum: Char;
begin
  Result := FWordStrim.CharByNum;
end;

function TWordReport.GetInputBlank: String;
begin
  Result := FWordStrim.InBlank
end;

function TWordReport.GetOutputBlank: String;
begin
  Result := FWordStrim.OutBlank;
end;

procedure TWordReport.OnErrorH(Sender: TObject; var Msg: TMessage;
  var sStr: String);
begin
  if Assigned(OnReportError) then OnReportError(self, Msg.Msg, Msg.LParam, sStr);
end;

procedure TWordReport.lOnSectionBegin(Sender: TObject);
begin
  if Assigned(OnSectionBegin) then OnSectionBegin(self, FWordStrim.CurrentSection, FWordStrim.CurrentSectionNum);
end;

procedure TWordReport.lOnSectionCheck(Sender: TObject; var bCheck: Boolean);
begin
  if Assigned(OnSectionCheck) then OnSectionCheck(self, FWordStrim.CurrentSection, FWordStrim.CurrentSectionNum, bCheck);
end;

procedure TWordReport.lOnSectionEnd(Sender: TObject);
begin
  if Assigned(OnSectionEnd) then OnSectionEnd(self, FWordStrim.CurrentSection, FWordStrim.CurrentSectionNum);
end;

procedure TWordReport.lOnSectionRqByName(Sender: TObject; var sFieldName, sStr: String);
begin
  if Assigned(OnRequestByName) then OnRequestByName(self, FWordStrim.CurrentSection, sFieldName, sStr);
end;

procedure TWordReport.lOnSectionRqByNum(Sender: TObject; var iP: Integer; var sStr: String);
begin
  if Assigned(OnRequestByNum) then OnRequestByNum(self, FWordStrim.CurrentSection, iP, sStr);
end;

procedure TWordReport.ReportExecute;
begin
  EventsRedirect;
  FWordStrim.Execute;
end;

procedure TWordReport.ReportHorizontalyExecute;
begin
end;

procedure TWordReport.ReportHorizontalyResume;
begin
end;

procedure TWordReport.ReportResume;
begin
  EventsRedirect;
  FWordStrim.Start;
end;

procedure TWordReport.SetCharByName(const Value: Char);
begin
  FWordStrim.CharByName := Value;
end;

procedure TWordReport.SetCharByNum(const Value: Char);
begin
  FWordStrim.CharByNum := Value;
end;

procedure TWordReport.SetInputBlank(const Value: String);
begin
  FWordStrim.InBlank := Value;
end;

procedure TWordReport.SetOutputBlank(const Value: String);
begin
  FWordStrim.OutBlank := Value;
end;

procedure TWordReport.EventsRedirect;
begin
  FWordStrim.OnStartReport := OnStartReport;
  FWordStrim.OnEndReport := OnEndReport;

  FWordStrim.OnMoveTop := OnMoveTop;
  FWordStrim.OnMoveNext := OnMoveNext;

  FWordStrim.OnReportError := OnErrorH;

  FWordStrim.OnSectionBegin :=             lOnSectionBegin;
  FWordStrim.OnSectionCheck :=             lOnSectionCheck;
  FWordStrim.OnSectionDataRequestByNum :=  lOnSectionRqByNum;
  FWordStrim.OnSectionDataRequestByName := lOnSectionRqByName;
  FWordStrim.OnSectionEnd :=               lOnSectionEnd;
  FWordStrim.OnSectionComplit :=           lOnSectionComplit;

  FWordStrim.OnSectionPrepare := OnSectionPrepare;

end;

procedure TWordReport.OutSection(SectionName: ShortString);
begin
  FWordStrim.OutSection(SectionName);
end;

function TWordReport.PrepareHorizontalySection(const SectionName: ShortString): Integer;
begin
  Result := FWordStrim.PrepareSection(SectionName);
end;

function TWordReport.PrepareSection(const SectionName: ShortString): Integer;
begin
  Result := FWordStrim.PrepareSection(SectionName);
end;

procedure TWordReport.OutHorizontalySection(SectionName: ShortString);
begin
  //dumy
end;

procedure TWordReport.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  //dumy
end;

procedure TWordReport.OutSectionByNum(SectionNum: Integer);
begin
  FWordStrim.OutSectionByNum(SectionNum);
end;

function TWordReport.GetCurrentSection: String;
begin
  Result := FWordStrim.CurrentSection;
end;

function TWordReport.GetCurrentSectionNum: Integer;
begin
  Result := FWordStrim.CurrentSectionNum;
end;

function TWordReport.GetTypeDiraction: TTtypeDiraction;
begin
  Result := tdVerticaly;
end;

procedure TWordReport.lOnSectionComplit(Sender: TObject);
begin
  if Assigned(OnSectionComplit) then OnSectionComplit(self, FWordStrim.CurrentSection, FWordStrim.CurrentSectionNum);
end;

end.
