unit XLS4Reports;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DDEReports, CommonType, DsgnIntf, SEditProp, RThread, Math, DDEClient, DB, ERep;

const
  XLSDimension: array [0..6] of Word = ($200, $0A, 0, $FFFF, 0, $FF, 0);
  XLSCOL: array[0..7] of Word = ($7D, $0C, 0, 0, 0, $F, 0, 0);
  XLSRow: array[0..9] of Word = ($208, 16, 0, 0, 0, 0, 0, 0, 0, 0);
  XLSBlank: array[0..4] of Word = ($201, 6, 0, 0, 0);
  XLSRK: array[0..4] of Word = ($27E, 10, 0, 0, 0);
  XLSNumber: array[0..4] of Word = ($203, 14, 0, 0, 0);
  XLSLabel: array[0..5] of Word = ($204, 0, 0, 0, 0, 0);
  XLSBool: array[0..4] of Word = ($206, 8, 0, 0, 0);

type

  TBiffRecHeader = packed record
    Opcode: Word;
    Length: Word;
  end;
  pTBiffRecHeader = ^TBiffRecHeader;

  TBiffRec = packed record
    head: TBiffRecHeader;
    data: Pointer;
    rb: Integer;
    cb: Integer;
    r: Integer;
    c: Integer;
    col_b: Integer;
    col_e: Integer;
    c1: Integer;
    c2: Integer;
    w: Integer;
    d: Variant;
  end;
  pTBiffRec = ^TBiffRec;

  TFileBufferedStream = class(TFileStream)
  private
    FBuffer: Pointer;
    FBufferCounter: Integer;
    FBufferMax: Integer;
    FBufferSize: Integer;
  public
    constructor Create(BuffSize: Integer; const FileName: string; Mode: Word);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure FlashBuffer;
  end;

  TXLS4Report = class(TReportAbstract)
  private

    FCharByNum: Char;
    FCharByName: Char;
    FInputBlank: String;
    FOutputBlank: String;

    FInputStream: TStream;
    FOutputStream: TStream;

    FStreamA: TMemoryStream;
    FStreamB: TMemoryStream;

    BiffBody: TList;

    FCurrRow: Integer;
    FCurrCol: Integer;

    FSectionsV: array [0..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;
    FSectionsH: array [0..MAX_SECTION_NUM] of TExcelDdeReportSectionInfo;

    FCurrentSection: String;
    FCurrentSectionNum: Integer;

    FTypeDiraction: TTtypeDiraction;

    FLastDynSection:    Integer;
    FLastDynSectionH:    Integer;
    lWM_REP_SECTION_PREPARE: boolean;

    LHandle: Integer;
    FFreeStreamIn: boolean;
    FFreeStreamOut: boolean;

    procedure CloseBody;

    procedure PrepareFileHeaderV;
    procedure PrepareFileFooterV;

    procedure StartReportV;
    procedure PrepareSectionV(SectionName: String; SectionNum: LongWord);
    procedure OutSectionV(SectionName: String; bOut: Boolean; SectionNum: LongWord);
    procedure CloseSectionsV;

    procedure PrepareFileHeaderH;
    procedure PrepareFileFooterH;

    procedure StartReportH;
    procedure PrepareSectionH(SectionName: String; SectionNum: LongWord);
    procedure OutSectionH(SectionName: String; bOut: Boolean; SectionNum: LongWord);
    procedure CloseSectionsH;
    procedure FlashHReport;

    procedure StartBiDirectionReport;

    function OpenInputBlank: boolean;
    function CreateOutputReport: boolean;
    procedure CloseInputBlank;
    procedure CloseOutputReport;
    procedure OutVariant(FOutputStream: TStream; V: Variant; var pp: TBiffRec);

  protected
    { Protected declarations }

    function GetCurrentSection: String; override;
    function GetCurrentSectionNum: Integer; override;
    function GetTypeDiraction: TTtypeDiraction; override;

  public
    { Public declarations }

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    procedure ReportResume; override;
    procedure ReportExecute; override;
    procedure ReportHorizontalyResume; override;
    procedure ReportHorizontalyExecute; override;
    procedure BiDirectionResume; override;
    procedure BiDirectionExecute; override;
    function PrepareSection(const SectionName: ShortString): Integer; override;
    function PrepareHorizontalySection(const SectionName: ShortString): Integer; override;

    procedure OutSection(SectionName: ShortString); override;
    procedure OutHorizontalySection(SectionName: ShortString); override;
    procedure OutSectionByNum(SectionNum: Integer); override;
    procedure OutHorizontalySectionByNum(SectionNum: Integer); override;

    property InputStream: TStream read FInputStream write FInputStream;
    property OutputStream: TStream read FOutputStream write FOutputStream;

    property FreeStreamIn: boolean read FFreeStreamIn write FFreeStreamIn default true;
    property FreeStreamOut: boolean read FFreeStreamOut write FFreeStreamOut default true;

  published
    { Published declarations }

    property InputBlank: String read FInputBlank write FInputBlank;
    property OutputBlank: String read FOutputBlank write FOutputBlank;
    property CharByNum: Char read FCharByNum write FCharByNum default #35; {#0001}
    property CharByName: Char read FCharByName write FCharByName default #126; {~name~}

    property OnStartReport;
    property OnEndReport;
    property OnReportError;
    property OnMoveTop;
    property OnMoveNext;
    property OnSectionBegin;
    property OnSectionEnd;
    property OnSectionComplit;
    property OnSectionCheck;
    property OnDataRequest;
    property OnRequestByName;
    property OnRequestByNum;
    property OnSectionPrepare;

  end;

function Ass2Number(j: LongWord): Variant;
function IsIntOrFloat(p: LongWord): TFieldType;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('VK Reports', [TXLS4Report]);
  RegisterPropertyEditor(TypeInfo(String), TXLS4Report, 'InputBlank', TEdOpenBlank);
  RegisterPropertyEditor(TypeInfo(String), TXLS4Report, 'OutputBlank', TEdSaveBlank);
end;

{ TXLS4Report }

procedure TXLS4Report.BiDirectionExecute;
begin
  StartBiDirectionReport;
end;

procedure TXLS4Report.BiDirectionResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartBiDirectionReport;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TXLS4Report.CloseBody;
var
  j, k: Integer;
  pp: pTBiffRec;
begin
  k := BiffBody.Count - 1;
  for j := 0 to k do
  begin
    pp := pTBiffRec(BiffBody.Items[j]);
    if pp.head.Length > 0 then
      FreeMem(pp.data);
    FreeMem(pp);
  end;
  BiffBody.Destroy;
  BiffBody := nil;
end;

procedure TXLS4Report.CloseInputBlank;
begin
  if FFreeStreamIn then begin
    FInputStream.Free;
    FInputStream := nil;
  end;
end;

procedure TXLS4Report.CloseOutputReport;
begin
  if FFreeStreamOut then begin
    FOutputStream.Free;
    FOutputStream := nil;
  end;
end;

procedure TXLS4Report.CloseSectionsH;
var
  i, j, k: Integer;
  pp: pTBiffRec;
begin
	for i := 0 to MAX_SECTION_NUM do
	  if FSectionsH[i].SectionUsed then
    begin
		  FSectionsH[i].SectionUsed := FALSE;
      k := FSectionsH[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTBiffRec(FSectionsH[i].Contents.Items[j]);
        if pp.head.Length > 0 then
          FreeMem(pp.data);
        FreeMem(pp);
      end;
      FSectionsH[i].Contents.Destroy;
      FSectionsH[i].Contents := nil;
    end;
end;

procedure TXLS4Report.CloseSectionsV;
var
  i, j, k: Integer;
  pp: pTBiffRec;
begin
	for i := 0 to MAX_SECTION_NUM do
	  if FSectionsV[i].SectionUsed then
    begin
		  FSectionsV[i].SectionUsed := FALSE;
      k := FSectionsV[i].Contents.Count - 1;
      for j := 0 to k do
      begin
        pp := pTBiffRec(FSectionsV[i].Contents.Items[j]);
        if pp.head.Length > 0 then
          FreeMem(pp.data);
        FreeMem(pp);
      end;
      FSectionsV[i].Contents.Destroy;
      FSectionsV[i].Contents := nil;
    end;
end;

constructor TXLS4Report.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FCharByNum := #35;
  FCharByName := #126;
  FLastDynSection := 13;
  FLastDynSectionH := 13;
  lWM_REP_SECTION_PREPARE := false;
  FInputStream := nil;
  FOutputStream := nil;
  BiffBody := nil;
  FFreeStreamIn := true;
  FFreeStreamOut := true;
end;

function TXLS4Report.CreateOutputReport: boolean;
begin
  if FOutputStream <> nil then begin
    FOutputStream.Seek(0, soFromBeginning);
    FOutputStream.Size := 0;
    Result := true;
  end else begin
    try
      FOutputStream := TFileBufferedStream.Create(FILE_BUFFER, FOutputBlank, fmCreate);
      Result := true;
    except
      FOutputStream := nil;
      Result := false;
    end;
  end;
end;

destructor TXLS4Report.Destroy;
begin
  CloseSectionsH;
  CloseSectionsV;
  inherited Destroy;
end;

procedure TXLS4Report.FlashHReport;
begin
  FStreamA.WriteBuffer(XLSDimension, SizeOf(XLSDimension));
  FStreamA.Seek(0, soFromBeginning);
  FOutputStream.CopyFrom(FStreamA, FStreamA.Size);
  FStreamB.Seek(0, soFromBeginning);
  FOutputStream.CopyFrom(FStreamB, FStreamB.Size);
end;

function TXLS4Report.GetCurrentSection: String;
begin
  Result := FCurrentSection;
end;

function TXLS4Report.GetCurrentSectionNum: Integer;
begin
  Result := FCurrentSectionNum;
end;

function TXLS4Report.GetTypeDiraction: TTtypeDiraction;
begin
  Result := FTypeDiraction;
end;

function TXLS4Report.OpenInputBlank: boolean;
begin
  if FInputStream <> nil then begin
    FInputStream.Seek(0, soFromBeginning);
    Result := true;
  end else begin
    try
      FInputStream := TFileBufferedStream.Create(FILE_BUFFER, FInputBlank, fmOpenRead or fmShareDenyWrite);
      Result := true;
    except
      FInputStream := nil;
      Result := false;
    end;
  end;
end;

procedure TXLS4Report.OutHorizontalySection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin

  if SectionName = 'H_REPORTHEADER'         then
  begin
    OutSectionH('H_REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPHEADER'          then
  begin
    OutSectionH('H_GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPHEADER'       then
  begin
    OutSectionH('H_SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPHEADER'    then
  begin
    OutSectionH('H_SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPHEADER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'H_BODY'                 then
  begin
    OutSectionH('H_BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'H_SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionH('H_SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'H_SUBSUBGROUPFOOTER'    then
  begin
    OutSectionH('H_SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'H_SUBGROUPFOOTER'       then
  begin
    OutSectionH('H_SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'H_GROUPFOOTER'          then
  begin
    OutSectionH('H_GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'H_REPORTFOOTER'         then
  begin
    OutSectionH('H_REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsH[i].SectionName then
    begin
      OutSectionH(SectionName, true, i);
      Break;
    end;
lEnd:
end;

procedure TXLS4Report.OutHorizontalySectionByNum(SectionNum: Integer);
begin
  OutSectionH(FSectionsH[SectionNum].SectionName, true, SectionNum);
end;

procedure TXLS4Report.OutSection(SectionName: ShortString);
var
  i: Integer;
label
  lEnd;
begin
  if SectionName = 'REPORTHEADER'         then
  begin
    OutSectionV('REPORTHEADER', true, 1);
    goto lEnd;
  end;
  if SectionName = 'GROUPHEADER'          then
  begin
    OutSectionV('GROUPHEADER', true, 2);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPHEADER'       then
  begin
    OutSectionV('SUBGROUPHEADER', true, 3);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPHEADER'    then
  begin
    OutSectionV('SUBSUBGROUPHEADER', true, 4);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPHEADER'  then
  begin
    OutSectionV('SUPERSUBGROUPHEADER', true, 5);
    goto lEnd;
  end;
  if SectionName = 'BODY'                 then
  begin
    OutSectionV('BODY', true, 6);
    goto lEnd;
  end;
  if SectionName = 'SUPERSUBGROUPFOOTER'  then
  begin
    OutSectionV('SUPERSUBGROUPFOOTER', true, 7);
    goto lEnd;
  end;
  if SectionName = 'SUBSUBGROUPFOOTER'    then
  begin
    OutSectionV('SUBSUBGROUPFOOTER', true, 8);
    goto lEnd;
  end;
  if SectionName = 'SUBGROUPFOOTER'       then
  begin
    OutSectionV('SUBGROUPFOOTER', true, 9);
    goto lEnd;
  end;
  if SectionName = 'GROUPFOOTER'          then
  begin
    OutSectionV('GROUPFOOTER', true, 10);
    goto lEnd;
  end;
  if SectionName = 'REPORTFOOTER'         then
  begin
    OutSectionV('REPORTFOOTER', true, 11);
    goto lEnd;
  end;
  // if Section is a Dynamic.
  for i := 13 to MAX_SECTION_NUM do
    if SectionName = FSectionsV[i].SectionName then
    begin
      OutSectionV(SectionName, true, i);
      Break;
    end;
lEnd:
end;

procedure TXLS4Report.OutSectionByNum(SectionNum: Integer);
begin
  OutSectionV(FSectionsV[SectionNum].SectionName, true, SectionNum);
end;

procedure TXLS4Report.OutSectionH(SectionName: String; bOut: Boolean;
  SectionNum: LongWord);
var
  i: Integer;
  List: TList;
  pp: TBiffRec;
  bCheck: boolean;
  V: Variant;
  H: Boolean;
  ContentColRow: String;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
  WStream: TStream;
  OldInd: Integer;
  OldName: String;
begin
  OldInd := FCurrentSectionNum;
  OldName := FCurrentSection;
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
  try
    if FCurrentSectionNum = 0 then
      WStream := FStreamA
    else
      WStream := FStreamB;
    if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
    if FSectionsH[SectionNum].SectionUsed then
    begin
      if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
      if bOut then
      begin
        List := FSectionsH[SectionNum].Contents;
        for i := 0 to List.Count - 1 do begin
          pp := pTBiffRec(List.Items[i])^;
          case pp.head.Opcode of
            513, 515, 516, 517, 638, 1030:
            begin
              bCheck := false;
              V := Unassigned;
              V := pp.d;
              if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, V, bCheck, pp.r + 1, FCurrCol + pp.cb + 1, pp.r + 1, pp.c + 1);
              if bCheck then
                OutVariant(WStream, V, pp)
              else begin
                H := false;
                if pp.head.Opcode = 516 then begin //Label
                  ContentColRow := V;
                  if Assigned(OnRequestByNum) then begin
                    pSt := 0;
                    while true do
                    begin
                      pInt := LScan(ContentColRow, FCharByNum, pSt);
                      if pInt = 0 then Break else H := true;
                      pSt := pInt;
                      if  ( ContentColRow[pInt + 1] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 2] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 3] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 4] in ['0'..'9'] ) then
                      begin
                        iP := StrToInt(Copy(ContentColRow, pInt + 1, 4));
                        sStr := '';
                        OnRequestByNum(self, SectionName, iP, sStr);
                        ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt + 5, Length(ContentColRow));
                      end;
                    end;
                  end;
                  if Assigned(OnRequestByName) then begin
                    pSt := 0;
                    while true do
                    begin
                      pInt := LScan(ContentColRow, FCharByName, pSt);
                      if pInt = 0 then Break;
                      pInt1 := LScan(ContentColRow, FCharByName, pInt);
                      if pInt1 = 0 then Break else H := true;
                      pSt := pInt1;
                      sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
                      sStr := '';
                      OnRequestByName(self, SectionName, sField, sStr);
                      ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
                    end;
                  end;
                end;
                if H then
                  OutVariant(WStream, ContentColRow, pp)
                else begin //if not handled with OnRequestByNum or OnRequestByName
                  //Copy with correct Row and Col
                  pWord(pp.data)^ := pp.r;
                  pWord(pChar(pp.data) + 2)^ := FCurrCol + pp.cb;
                  WStream.WriteBuffer(pp.head, SizeOf(pp.head));
                  WStream.WriteBuffer(pp.data^, pp.head.Length);
                end;
              end;
            end;
            125:      //Col
            begin
              //Copy with correct Col
              pWord(pp.data)^ := FCurrCol + pp.col_b - pp.cb;
              pWord(pChar(pp.data) + 2)^ := FCurrCol + pp.col_e - pp.cb;
              FStreamA.WriteBuffer(pp.head, SizeOf(pp.head));
              FStreamA.WriteBuffer(pp.data^, pp.head.Length);
            end;
          else
            begin
              //Only copy
              WStream.WriteBuffer(pp.head, SizeOf(pp.head));
              WStream.WriteBuffer(pp.data^, pp.head.Length);
            end;
          end;
        end;
        FCurrCol := FCurrCol + FSectionsH[SectionNum].ColsCount;
        if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
      end;
    end;
  finally
    if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
    FCurrentSectionNum := OldInd;
    FCurrentSection := OldName;
  end;
end;

procedure TXLS4Report.OutSectionV(SectionName: String; bOut: Boolean;
  SectionNum: LongWord);
var
  i: Integer;
  List: TList;
  pp: TBiffRec;
  bCheck: boolean;
  V: Variant;
  H: Boolean;
  ContentColRow: String;
  pInt, pInt1, pSt, iP: Integer;
  sStr, sField: String;
  OldInd: Integer;
  OldName: String;
begin
  OldInd := FCurrentSectionNum;
  OldName := FCurrentSection;
  FCurrentSection := SectionName;
  FCurrentSectionNum := SectionNum;
  try
    if Assigned(OnSectionBegin) then OnSectionBegin(self, SectionName, SectionNum);
    if FSectionsV[SectionNum].SectionUsed then
    begin
      if Assigned(OnSectionCheck) then OnSectionCheck(self, SectionName, SectionNum, bOut);
      if bOut then
      begin
        List := FSectionsV[SectionNum].Contents;
        for i := 0 to List.Count - 1 do begin
          pp := pTBiffRec(List.Items[i])^;
          case pp.head.Opcode of
            513, 515, 516, 517, 638, 1030:
            begin
              bCheck := false;
              V := Unassigned;
              V := pp.d;
              if Assigned(OnDataRequest) then OnDataRequest(self, SectionName, SectionNum, V, bCheck, FCurrRow + pp.rb + 1, pp.c + 1, pp.r + 1, pp.c + 1);
              if bCheck then
                OutVariant(FOutputStream, V, pp)
              else begin
                H := false;
                if pp.head.Opcode = 516 then begin //Label
                  ContentColRow := V;
                  if Assigned(OnRequestByNum) then begin
                    pSt := 0;
                    while true do
                    begin
                      pInt := LScan(ContentColRow, FCharByNum, pSt);
                      if pInt = 0 then Break else H := true;
                      pSt := pInt;
                      if  ( ContentColRow[pInt + 1] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 2] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 3] in ['0'..'9'] ) and
                          ( ContentColRow[pInt + 4] in ['0'..'9'] ) then
                      begin
                        iP := StrToInt(Copy(ContentColRow, pInt + 1, 4));
                        sStr := '';
                        OnRequestByNum(self, SectionName, iP, sStr);
                        ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt + 5, Length(ContentColRow));
                      end;
                    end;
                  end;
                  if Assigned(OnRequestByName) then begin
                    pSt := 0;
                    while true do
                    begin
                      pInt := LScan(ContentColRow, FCharByName, pSt);
                      if pInt = 0 then Break;
                      pInt1 := LScan(ContentColRow, FCharByName, pInt);
                      if pInt1 = 0 then Break else H := true;
                      pSt := pInt1;
                      sField := Copy(ContentColRow, pInt+1, pInt1-pInt-1);
                      sStr := '';
                      OnRequestByName(self, SectionName, sField, sStr);
                      ContentColRow := Copy(ContentColRow, 1, pInt - 1) + sStr + Copy(ContentColRow, pInt1 + 1, Length(ContentColRow));
                    end;
                  end;
                end;
                if H then
                  OutVariant(FOutputStream, ContentColRow, pp)
                else begin //if not handled with OnRequestByNum or OnRequestByName
                  //Copy with correct Row and Col
                  pWord(pp.data)^ := FCurrRow + pp.rb;
                  pWord(pChar(pp.data) + 2)^ := pp.c;;
                  FOutputStream.WriteBuffer(pp.head, SizeOf(pp.head));
                  FOutputStream.WriteBuffer(pp.data^, pp.head.Length);
                end;
              end;
            end;
            {
            520:      //Row
            begin
              //Copy with correct Row      !! not Col !!
              pWord(pp.data)^ := FCurrRow + pp.rb;
              FOutputStream.WriteBuffer(pp.head, SizeOf(pp.head));
              FOutputStream.WriteBuffer(pp.data^, pp.head.Length);

              //XLSRow[2] := FCurrRow + pp.rb;
              //XLSRow[3] := pp.c1;
              //XLSRow[4] := pp.c2;
              //XLSRow[5] := pp.w;
              //FOutputStream.WriteBuffer(XLSRow, SizeOf(XLSRow));

            end;
            }
          else
            begin
              //Only copy
              FOutputStream.WriteBuffer(pp.head, SizeOf(pp.head));
              FOutputStream.WriteBuffer(pp.data^, pp.head.Length);
            end;
          end;
        end;
        FCurrRow := FCurrRow + FSectionsV[SectionNum].RowsCount;
        if Assigned(OnSectionEnd) then OnSectionEnd(self, SectionName, SectionNum);
      end;
    end;
  finally
    if Assigned(OnSectionComplit) then OnSectionComplit(self, SectionName, SectionNum);
    FCurrentSectionNum := OldInd;
    FCurrentSection := OldName;
  end;
end;

procedure TXLS4Report.OutVariant(FOutputStream: TStream; V: Variant; var pp: TBiffRec);

  procedure OutNumber(F: Double);
  begin
    case FTypeDiraction of
    tdVerticaly:
      begin
        XLSNumber[2] := FCurrRow + pp.rb;
        XLSNumber[3] := pp.c;
      end;
    tdHorizontaly:
      begin
        XLSNumber[2] := pp.r;
        XLSNumber[3] := FCurrCol + pp.cb;
      end;
    end;
    XLSNumber[4] := pWord(pChar(pp.data) + 4)^;
    FOutputStream.WriteBuffer(XLSNumber, SizeOf(XLSNumber));
    FOutputStream.WriteBuffer(F, SizeOf(F));
  end;

var
  C: Integer;
  S: String;
  B: Word;
begin
  case VarType(V) of
    varEmpty, varNull:
      begin
        case FTypeDiraction of
        tdVerticaly:
          begin
            XLSBlank[2] := FCurrRow + pp.rb;
            XLSBlank[3] := pp.c;
          end;
        tdHorizontaly:
          begin
            XLSBlank[2] := pp.r;
            XLSBlank[3] := FCurrCol + pp.cb;
          end;
        end;
        XLSBlank[4] := pWord(pChar(pp.data) + 4)^;
        FOutputStream.WriteBuffer(XLSBlank, SizeOf(XLSBlank));
      end;
    varSmallint, varInteger:
      begin
        C := V;
        if (-536870912 <= C) and (C <= 536870911) then begin
          C := C shl 2 + 2;
          case FTypeDiraction of
          tdVerticaly:
            begin
              XLSRK[2] := FCurrRow + pp.rb;
              XLSRK[3] := pp.c;
            end;
          tdHorizontaly:
            begin
              XLSRK[2] := pp.r;
              XLSRK[3] := FCurrCol + pp.cb;
            end;
          end;
          XLSRK[4] := pWord(pChar(pp.data) + 4)^;
          FOutputStream.WriteBuffer(XLSRK, SizeOf(XLSRK));
          FOutputStream.WriteBuffer(C, SizeOf(C));
        end else
          OutNumber(C);
      end;
    varSingle, varDouble, varCurrency, varDate: OutNumber(V);
    varOleStr, varString:
      begin
        S := V;
        XLSLAbel[1] := Length(S) + 8;
        case FTypeDiraction of
        tdVerticaly:
          begin
            XLSLabel[2] := FCurrRow + pp.rb;
            XLSLabel[3] := pp.c;
          end;
        tdHorizontaly:
          begin
            XLSLabel[2] := pp.r;
            XLSLabel[3] := FCurrCol + pp.cb;
          end;
        end;
        XLSLabel[4] := pWord(pChar(pp.data) + 4)^;
        XLSLabel[5] := Length(S);
        FOutputStream.WriteBuffer(XLSLabel, SizeOf(XLSLabel));
        FOutputStream.WriteBuffer(Pointer(S)^, Length(S));
      end;
    varBoolean:
      begin
        if V then
          B := 1
        else
          B := 0;
        case FTypeDiraction of
        tdVerticaly:
          begin
            XLSBool[2] := FCurrRow + pp.rb;
            XLSBool[3] := pp.c;
          end;
        tdHorizontaly:
          begin
            XLSBool[2] := pp.r;
            XLSBool[3] := FCurrCol + pp.cb;
          end;
        end;
        XLSBool[4] := pWord(pChar(pp.data) + 4)^;
        FOutputStream.WriteBuffer(XLSBool, SizeOf(XLSBool));
        FOutputStream.WriteBuffer(B, SizeOf(B));
      end;
  end;
end;

procedure TXLS4Report.PrepareFileFooterH;
var
  i: Integer;
  pp: TBiffRec;
  qq: pTBiffRec;
begin
	FSectionsH[12].SectionName := 'FILEFOOTER';
	FSectionsH[12].SectionNum := 12;
	FSectionsH[12].SectionUsed := true;
  FSectionsH[12].Contents := TList.Create;
  for i := LHandle + 1 to BiffBody.Count - 1 do begin
    pp := pTBiffRec(BiffBody.Items[i])^;
    if not (pp.head.Opcode in [171, 153]) then begin
      GetMem(qq, SizeOf(TBiffRec));
      FillChar(qq^, SizeOf(TBiffRec), 0);
      qq^.head.Opcode := pp.head.Opcode;
      qq^.head.Length := pp.head.Length;
      qq^.data := nil;
      qq^.r := 0;
      qq^.c := 0;
      qq^.d := Unassigned;
      qq^.d := NULL;
      if pp.head.Length > 0 then begin
        GetMem(qq^.data, pp.head.Length);
        Move(pp.data^, qq^.data^, pp.head.Length);
      end;
      FSectionsH[12].Contents.Add(qq);
    end;
  end;
end;

procedure TXLS4Report.PrepareFileFooterV;
var
  i: Integer;
  pp: TBiffRec;
  qq: pTBiffRec;
begin
	FSectionsV[12].SectionName := 'FILEFOOTER';
	FSectionsV[12].SectionNum := 12;
	FSectionsV[12].SectionUsed := true;
  FSectionsV[12].Contents := TList.Create;
  for i := LHandle + 1 to BiffBody.Count - 1 do begin
    pp := pTBiffRec(BiffBody.Items[i])^;
    GetMem(qq, SizeOf(TBiffRec));
    FillChar(qq^, SizeOf(TBiffRec), 0);
    qq^.head.Opcode := pp.head.Opcode;
    qq^.head.Length := pp.head.Length;
    qq^.data := nil;
    qq^.r := 0;
    qq^.c := 0;
    qq^.d := Unassigned;
    qq^.d := NULL;
    if pp.head.Length > 0 then begin
      GetMem(qq^.data, pp.head.Length);
      Move(pp.data^, qq^.data^, pp.head.Length);
    end;
    FSectionsV[12].Contents.Add(qq);
  end;
end;

procedure TXLS4Report.PrepareFileHeaderH;
var
  r_h: TBiffRecHeader;
  pp: pTBiffRec;
begin
	FSectionsH[0].SectionName := 'FILEHEADER';
	FSectionsH[0].SectionNum := 0;
	FSectionsH[0].SectionUsed := true;
  FSectionsH[0].Contents := TList.Create;
  while true do
  begin
    FInputStream.Read(r_h, 4);
    GetMem(pp, SizeOf(TBiffRec));
    FillChar(pp^, SizeOf(TBiffRec), 0);
    pp^.head.Opcode := r_h.Opcode;
    pp^.head.Length := r_h.Length;
    pp^.data := nil;
    pp^.r := 0;
    pp^.c := 0;
    pp^.d := Unassigned;
    pp^.d := NULL;
    if r_h.Length > 0 then begin
      GetMem(pp^.data, r_h.Length);
      FInputStream.Read(pp^.data^, r_h.Length);
    end;
    if r_h.Opcode = $7D then begin //Col
      BiffBody.Add(pp);
      Break;
    end;
    if r_h.Opcode = $200 then Break; //Dimention
    FSectionsH[0].Contents.Add(pp);
  end;
  while true do
  begin
    FInputStream.Read(r_h, 4);
    GetMem(pp, SizeOf(TBiffRec));
    FillChar(pp^, SizeOf(TBiffRec), 0);
    pp^.head.Opcode := r_h.Opcode;
    pp^.head.Length := r_h.Length;
    pp^.data := nil;
    pp^.r := 0;
    pp^.c := 0;
    pp^.d := Unassigned;
    pp^.d := NULL;
    if r_h.Length > 0 then begin
      GetMem(pp^.data, r_h.Length);
      FInputStream.Read(pp^.data^, r_h.Length);
    end;
    BiffBody.Add(pp);
    if r_h.Opcode = $0A then Break; //EOF
  end;
end;

procedure TXLS4Report.PrepareFileHeaderV;
var
  r_h: TBiffRecHeader;
  pp: pTBiffRec;
begin
	FSectionsV[0].SectionName := 'FILEHEADER';
	FSectionsV[0].SectionNum := 0;
	FSectionsV[0].SectionUsed := true;
  FSectionsV[0].Contents := TList.Create;
  while true do
  begin
    FInputStream.Read(r_h, 4);
    GetMem(pp, SizeOf(TBiffRec));
    FillChar(pp^, SizeOf(TBiffRec), 0);
    pp^.head.Opcode := r_h.Opcode;
    pp^.head.Length := r_h.Length;
    pp^.data := nil;
    pp^.r := 0;
    pp^.c := 0;
    pp^.d := Unassigned;
    pp^.d := NULL;
    if r_h.Length > 0 then begin
      GetMem(pp^.data, r_h.Length);
      FInputStream.Read(pp^.data^, r_h.Length);
    end;
    if r_h.Opcode = $200 then begin //Dimentions
      pWord(pChar(pp^.data) + 0)^ := 0;
      pWord(pChar(pp^.data) + 2)^ := $FFFF;
      pWord(pChar(pp^.data) + 4)^ := 0;
      pWord(pChar(pp^.data) + 6)^ := $FF;
    end;
    FSectionsV[0].Contents.Add(pp);
    if r_h.Opcode = $200 then Break; //Dimentions
  end;
  while true do
  begin
    FInputStream.Read(r_h, 4);
    GetMem(pp, SizeOf(TBiffRec));
    FillChar(pp^, SizeOf(TBiffRec), 0);
    pp^.head.Opcode := r_h.Opcode;
    pp^.head.Length := r_h.Length;
    pp^.data := nil;
    pp^.r := 0;
    pp^.c := 0;
    pp^.d := Unassigned;
    pp^.d := NULL;
    if r_h.Length > 0 then begin
      GetMem(pp^.data, r_h.Length);
      FInputStream.Read(pp^.data^, r_h.Length);
    end;
    BiffBody.Add(pp);
    if r_h.Opcode = $0A then Break; //EOF
  end;
end;

function TXLS4Report.PrepareHorizontalySection(
  const SectionName: ShortString): Integer;
begin
  if FLastDynSectionH > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSectionH;
  Inc(FLastDynSectionH);
  PrepareSectionH(SectionName, Result);
end;

function TXLS4Report.PrepareSection(
  const SectionName: ShortString): Integer;
begin
  if FLastDynSection > MAX_SECTION_NUM then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_MAX_DYN_SECTION_EXCEEDED, 0, 'LIMIT EXCEEDED');
  if not lWM_REP_SECTION_PREPARE then
    if Assigned(OnReportError) then OnReportError(self, WM_REP_ALLAGIAL_CALL, 0, 'ALLAGIAL CALL');
  Result := FLastDynSection;
  Inc(FLastDynSection);
  PrepareSectionV(SectionName, Result);
end;

procedure TXLS4Report.PrepareSectionH(SectionName: String;
  SectionNum: LongWord);
var
  i: Integer;
  pp: TBiffRec;
  qq: pTBiffRec;
  sLen: Byte;
  sName: String;
  r, r1, r2: Word;
  col_b, col_e: Word;
  c, c1, c2: Byte;
  Found: Boolean;
  ss: String;
  ff: Double;
  bb: Word;
  lw: LongWord;
begin
  r1 := 0; r2 := 0; c1 := 0; c2 := 0;
  Found := false;
  for i := 0 to FSectionsH[0].Contents.Count - 1 do begin
    pp := pTBiffRec(FSectionsH[0].Contents.Items[i])^;
    if pp.head.Opcode = 536 then begin    //Names definition
      sLen := pByte(pChar(pp.data) + 3)^;
      SetString(sName, pChar(pp.data) + 6, sLen);
      if sName = SectionName then begin
        r1 := pWord(pChar(pp.data) + pp.head.Length - 6)^;
        r2 := pWord(pChar(pp.data) + pp.head.Length - 4)^;
        c1 := pByte(pChar(pp.data) + pp.head.Length - 2)^;
        c2 := pByte(pChar(pp.data) + pp.head.Length - 1)^;
        Found := true;
        Break;
      end;
    end;
  end;
  if Found then begin
    FSectionsH[SectionNum].SectionName := SectionName;
    FSectionsH[SectionNum].SectionNum := SectionNum;
    FSectionsH[SectionNum].SectionUsed := true;
    FSectionsH[SectionNum].Contents := TList.Create;
    FSectionsH[SectionNum].RowsCount := r2 - r1 + 1;
    FSectionsH[SectionNum].ColsCount := c2 - c1 + 1;
    for i := 0 to BiffBody.Count - 1 do begin
      pp := pTBiffRec(BiffBody.Items[i])^;
      if pp.head.Opcode = 125 then begin  //Col definition
        col_b := pWord(pp.data)^;
        col_e := pWord(pChar(pp.data)+ 2)^;
        if ( (c1 <= col_b) and (col_b <= c2) ) or ( (c1 <= col_e) and (col_e <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.col_b := col_b;
          qq^.col_e := col_e;
          qq^.cb := c1;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 513 then begin  //Blank cell
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 515 then begin  //8 byte float
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          ff := pDouble(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          qq^.d := ff;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 516 then begin  //Label
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          SetString(ss, pChar(pp.data) + 8, pWord(pChar(pp.data) + 6)^);
          qq^.d := Unassigned;
          qq^.d := ss;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 517 then begin  //Bool
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          bb := pWord(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          if bb = 1 then
            qq^.d := true
          else
            qq^.d := false;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 638 then begin  //Cool cell - clear ass
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          lw := pLongWord(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          qq^.d := Ass2Number(lw);
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 1030 then begin  //Formula cell
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.cb := c - c1;
          qq^.r := r;
          qq^.c := c;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsH[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
    end;
  end;
end;

procedure TXLS4Report.PrepareSectionV(SectionName: String;
  SectionNum: LongWord);
var
  i: Integer;
  pp: TBiffRec;
  qq: pTBiffRec;
  sLen: Byte;
  sName: String;
  r, r1, r2: Word;
  c, c1, c2: Byte;
  Found: Boolean;
  ss: String;
  ff: Double;
  bb: Word;
  lw: LongWord;
begin
  r1 := 0; r2 := 0; c1 := 0; c2 := 0;
  Found := false;
  for i := 0 to FSectionsV[0].Contents.Count - 1 do begin
    pp := pTBiffRec(FSectionsV[0].Contents.Items[i])^;
    if pp.head.Opcode = 536 then begin    //Names definition
      sLen := pByte(pChar(pp.data) + 3)^;
      SetString(sName, pChar(pp.data) + 6, sLen);
      if sName = SectionName then begin
        r1 := pWord(pChar(pp.data) + pp.head.Length - 6)^;
        r2 := pWord(pChar(pp.data) + pp.head.Length - 4)^;
        c1 := pByte(pChar(pp.data) + pp.head.Length - 2)^;
        c2 := pByte(pChar(pp.data) + pp.head.Length - 1)^;
        Found := true;
        Break;
      end;
    end;
  end;
  if Found then begin
    FSectionsV[SectionNum].SectionName := SectionName;
    FSectionsV[SectionNum].SectionNum := SectionNum;
    FSectionsV[SectionNum].SectionUsed := true;
    FSectionsV[SectionNum].Contents := TList.Create;
    FSectionsV[SectionNum].RowsCount := r2 - r1 + 1;
    for i := 0 to BiffBody.Count - 1 do begin
      pp := pTBiffRec(BiffBody.Items[i])^;
      if pp.head.Opcode = 513 then begin  //Blank cell
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 515 then begin  //8 byte float
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          ff := pDouble(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          qq^.d := ff;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 516 then begin  //Label
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          SetString(ss, pChar(pp.data) + 8, pWord(pChar(pp.data) + 6)^);
          qq^.d := Unassigned;
          qq^.d := ss;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 517 then begin  //Bool
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          bb := pWord(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          if bb = 1 then
            qq^.d := true
          else
            qq^.d := false;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      {
      if pp.head.Opcode = 520 then begin  //Row
        r := pWord(pp.data)^;
        if ((r1 <= r) and (r <= r2)) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := 0;
          qq^.c1 := pWord(pChar(pp.data)+ 2)^;
          qq^.c2 := pWord(pChar(pp.data)+ 4)^;
          qq^.w  := pWord(pChar(pp.data)+ 6)^;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      }
      if pp.head.Opcode = 638 then begin  //Cool cell - clear ass
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          lw := pLongWord(pChar(pp.data) + 6)^;
          qq^.d := Unassigned;
          qq^.d := Ass2Number(lw);
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
      if pp.head.Opcode = 1030 then begin  //Formula cell
        r := pWord(pp.data)^;
        c := pWord(pChar(pp.data)+ 2)^;
        if ((r1 <= r) and (r <= r2)) and ( (c1 <= c) and (c <= c2) ) then begin
          GetMem(qq, SizeOf(TBiffRec));
          FillChar(qq^, SizeOf(TBiffRec), 0);
          qq^.head.Opcode := pp.head.Opcode;
          qq^.head.Length := pp.head.Length;
          qq^.data := nil;
          qq^.rb := r - r1;
          qq^.r := r;
          qq^.c := c;
          qq^.d := Unassigned;
          qq^.d := NULL;
          if pp.head.Length > 0 then begin
            GetMem(qq^.data, pp.head.Length);
            Move(pp.data^, qq^.data^, pp.head.Length);
          end;
          FSectionsV[SectionNum].Contents.Add(qq);
          if i > LHandle then LHandle := i;
        end;
      end;
    end;
  end;
end;

procedure TXLS4Report.ReportExecute;
begin
  StartReportV;
end;

procedure TXLS4Report.ReportHorizontalyExecute;
begin
  StartReportH;
end;

procedure TXLS4Report.ReportHorizontalyResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartReportH;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TXLS4Report.ReportResume;
var
  FThread: TRThread;
begin
  FThread := TRThread.Create(true);
  FThread.OnExecute := StartReportV;
  FThread.FreeOnTerminate := true;
  FThread.Resume;
end;

procedure TXLS4Report.StartBiDirectionReport;
var
  cpTempFile: String;
  cpOldOut: String;
begin
  try
    cpOldOut := OutputBlank;
    cpTempFile := GetTmpFileName;
    OutputBlank := cpTempFile;
    StartReportH;
    InputBlank := cpTempFile;
    OutputBlank := cpOldOut;
    StartReportV;
  finally
    DeleteFile(cpTempFile);
  end;
end;

procedure TXLS4Report.StartReportH;
var
  bCheck: Boolean;
begin
  try
    LHandle := 0;
    BiffBody := TList.Create;
    FStreamA := TMemoryStream.Create;
    FStreamB := TMemoryStream.Create;
    FLastDynSection := 13;
    FLastDynSectionH := 13;
    FTypeDiraction := tdHorizontaly;
    if Assigned(OnStartReport) then OnStartReport(self);
    try
      if OpenInputBlank then
      begin
        if CreateOutputReport then
        begin
          try
            PrepareFileHeaderH;
            PrepareSectionH('H_REPORTHEADER', 1);
            PrepareSectionH('H_GROUPHEADER', 2);
            PrepareSectionH('H_SUBGROUPHEADER', 3);
            PrepareSectionH('H_SUBSUBGROUPHEADER', 4);
            PrepareSectionH('H_SUPERSUBGROUPHEADER', 5);
            PrepareSectionH('H_BODY', 6);
            PrepareSectionH('H_SUPERSUBGROUPFOOTER', 7);
            PrepareSectionH('H_SUBSUBGROUPFOOTER', 8);
            PrepareSectionH('H_SUBGROUPFOOTER', 9);
            PrepareSectionH('H_GROUPFOOTER', 10);
            PrepareSectionH('H_REPORTFOOTER', 11);
            PrepareFileFooterH;
            OutSectionH('FILEHEADER', TRUE, 0);
            lWM_REP_SECTION_PREPARE := true;
            if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
            lWM_REP_SECTION_PREPARE := false;
            FCurrCol := 0;
            OutSectionH('H_REPORTHEADER', TRUE, 1);
            bCheck := True;
            if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
            while not bCheck do
            begin
              OutSectionH('H_GROUPHEADER', FALSE, 2);
              OutSectionH('H_SUBGROUPHEADER', FALSE, 3);
              OutSectionH('H_SUBSUBGROUPHEADER', FALSE, 4);
              OutSectionH('H_SUPERSUBGROUPHEADER', FALSE, 5);
              OutSectionH('H_BODY', TRUE, 6);
              OutSectionH('H_SUPERSUBGROUPFOOTER', FALSE, 7);
              OutSectionH('H_SUBSUBGROUPFOOTER', FALSE, 8);
              OutSectionH('H_SUBGROUPFOOTER', FALSE, 9);
              OutSectionH('H_GROUPFOOTER', FALSE, 10);
              bCheck := True;
              if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
            end;
            OutSectionH('H_REPORTFOOTER', TRUE, 11);
            OutSectionH('FILEFOOTER', TRUE, 12);
            FlashHReport;
          finally
            CloseSectionsH;
          end;
        end else if Assigned(OnReportError) then OnReportError(self, WM_REP_CREATE_ERROR, 0, 'Create Error');
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    finally
      CloseInputBlank;
      CloseOutputReport;
    end;
    if Assigned(OnEndReport) then OnEndReport(self);
  finally
    FTypeDiraction := tdNotDefined;
    FStreamA.Free;
    FStreamB.Free;
    FStreamA := nil;
    FStreamB := nil;
    CloseBody;
  end;
end;

procedure TXLS4Report.StartReportV;
var
  bCheck: Boolean;
begin
  try
    LHandle := 0;
    BiffBody := TList.Create;
    FLastDynSection := 13;
    FLastDynSectionH := 13;
    FTypeDiraction := tdVerticaly;
    if Assigned(OnStartReport) then OnStartReport(self);
    try
      if OpenInputBlank then
      begin
        if CreateOutputReport then
        begin
          try
            PrepareFileHeaderV;
            PrepareSectionV('REPORTHEADER', 1);
            PrepareSectionV('GROUPHEADER', 2);
            PrepareSectionV('SUBGROUPHEADER', 3);
            PrepareSectionV('SUBSUBGROUPHEADER', 4);
            PrepareSectionV('SUPERSUBGROUPHEADER', 5);
            PrepareSectionV('BODY', 6);
            PrepareSectionV('SUPERSUBGROUPFOOTER', 7);
            PrepareSectionV('SUBSUBGROUPFOOTER', 8);
            PrepareSectionV('SUBGROUPFOOTER', 9);
            PrepareSectionV('GROUPFOOTER', 10);
            PrepareSectionV('REPORTFOOTER', 11);
            PrepareFileFooterV;
            OutSectionV('FILEHEADER', TRUE, 0);
            lWM_REP_SECTION_PREPARE := true;
            if Assigned(OnSectionPrepare) then OnSectionPrepare(self);
            lWM_REP_SECTION_PREPARE := false;
            FCurrRow := 0;
            OutSectionV('REPORTHEADER', TRUE, 1);
            bCheck := True;
            if Assigned(OnMoveTop) then OnMoveTop(self, bCheck);
            while not bCheck do
            begin
              OutSectionV('GROUPHEADER', FALSE, 2);
              OutSectionV('SUBGROUPHEADER', FALSE, 3);
              OutSectionV('SUBSUBGROUPHEADER', FALSE, 4);
              OutSectionV('SUPERSUBGROUPHEADER', FALSE, 5);
              OutSectionV('BODY', TRUE, 6);
              OutSectionV('SUPERSUBGROUPFOOTER', FALSE, 7);
              OutSectionV('SUBSUBGROUPFOOTER', FALSE, 8);
              OutSectionV('SUBGROUPFOOTER', FALSE, 9);
              OutSectionV('GROUPFOOTER', FALSE, 10);
              bCheck := True;
              if Assigned(OnMoveNext) then OnMoveNext(self, bCheck);
            end;
            OutSectionV('REPORTFOOTER', TRUE, 11);
            OutSectionV('FILEFOOTER', TRUE, 12);
          finally
            CloseSectionsV;
          end;
        end else if Assigned(OnReportError) then OnReportError(self, WM_REP_CREATE_ERROR, 0, 'Create Error');
      end else if Assigned(OnReportError) then OnReportError(self, WM_REP_OPEN_ERROR, 0, 'Open Error');
    finally
      CloseInputBlank;
      CloseOutputReport;
    end;
    if Assigned(OnEndReport) then OnEndReport(self);
  finally
    FTypeDiraction := tdNotDefined;
    CloseBody;
  end;
end;

{ TFileBufferedStream }

constructor TFileBufferedStream.Create(BuffSize: Integer;
  const FileName: string; Mode: Word);
begin
  inherited Create(FileName, Mode);
  FBufferMax := 0;
  FBufferCounter := 0;
  FBufferSize := BuffSize;
  FBuffer := nil;
  GetMem(FBuffer, FBufferSize);
end;

destructor TFileBufferedStream.Destroy;
begin
  FlashBuffer;
  if FBuffer <> nil then FreeMem(FBuffer);
  inherited Destroy;
end;

procedure TFileBufferedStream.FlashBuffer;
begin
  //
end;

function TFileBufferedStream.Read(var Buffer; Count: Integer): Longint;
begin
  Result := inherited Read(Buffer, Count);
end;

function TFileBufferedStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  Result := inherited Seek(Offset, Origin);
end;

function TFileBufferedStream.Write(const Buffer; Count: Integer): Longint;
begin
  Result := inherited Write(Buffer, Count);
end;

//------------------------------------------------------------------------------

function Ass2Number(j: LongWord): Variant;
var
  x: Integer;
  m: LongWord;
  p: Integer;
  F: Extended;
  q: Boolean;
  q1: Boolean;
  sign: Boolean;
begin
  if j in [0, 1, 2, 3] then
    Result := 0
  else begin
    q1 := ((j and 1) = 1);
    if ((j and 2) = 2) then begin
      p := j;
      q := (LongWord(p) = (LongWord(p) or $80000000));
      p := p shr 2;
      if q then
        p := LongWord(p) or $C0000000;
      Result := p;
    end else begin
      sign := (j = (j or $80000000));
      p := j;
      q := (p = (p or $40000000));
      p := p and $7FFFFFFF;
      p := p shr 20;
      p := p xor $00000400;
      if not q then
        p := LongWord(p) or $fffff800;
      x := p - 17;
      p := j and $000FFFFF;
      p := p or $00100000;
      m := (p shr 2);
      F :=  Ldexp(m, x);   //m*exp2(x);
      if sign then
        F := -F;
      if IsIntOrFloat(j) = ftInteger then
        Result := Integer(Trunc(F))
      else
        Result := F;
    end;
    if q1 then
      Result := Result * 0.01;
  end;
end;

function IsIntOrFloat(p: LongWord): TFieldType;
var
  b: Boolean;
  q: LongWord;
  mpos: Integer;
begin
  q := p;
  b := (p = (p or $40000000));  //  Second left bit is set
  p := p shr 20;
  p := p xor $00000400;
  if not b then p := p or $fffff800;
  Inc(p);
  q := q and $000FFFFF;
  q := q shr 1;
  mpos := 0;
  if q > 0 then begin
    mpos := 19;
    repeat
      Dec(mpos);
      q := q shr 1;
    until Odd(q);
  end;
  if mpos > Integer(p)
    then Result := ftFloat  //ShowMessage('Value must not be integer')
    else Result := ftInteger; //ShowMessage('Value is integer');
end;

end.

